
#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "sol-flow/oic.h"

#include "sol-coap.h"
#include "sol-mainloop.h"
#include "sol-oic.h"
#include "sol-oic-client.h"
#include "sol-oic-server.h"
#include "sol-str-slice.h"
#include "sol-str-table.h"
#include "sol-util.h"

#define DEFAULT_UDP_PORT 5683
#define MULTICAST_ADDRESS_IPv4 "224.0.1.187"
#define MULTICAST_ADDRESS_IPv6_LOCAL "ff02::fd"
#define MULTICAST_ADDRESS_IPv6_SITE "ff05::fd"
#define FIND_PERIOD_MS 5000
#define UPDATE_TIMEOUT_MS 50

#define DEVICE_ID_LEN (16)

#define streq(a, b) (strcmp((a), (b)) == 0)
#define likely(x)   __builtin_expect(!!(x), 1)

struct client_resource;
struct server_resource;

struct client_resource_funcs {
    bool (*to_repr_vec)(void *data, struct sol_oic_map_writer *repr_map);
    int (*from_repr_vec)(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec);
    void (*inform_flow)(struct client_resource *resource);
    int found_port;
    int device_id_port;
};

struct server_resource_funcs {
    bool (*to_repr_vec)(void *data, struct sol_oic_map_writer *repr_map);
    int (*from_repr_vec)(struct server_resource *resource, const struct sol_oic_map_reader *repr);
    void (*inform_flow)(struct server_resource *resource);
};

struct client_resource {
    struct sol_flow_node *node;
    const struct client_resource_funcs *funcs;

    struct sol_oic_resource *resource;

    struct pending {
        struct sol_oic_pending *ipv4, *ipv6_local, *ipv6_site;
    } discover_pending, scan_pending;
    /* discover registered device IDs or scan for unknown IDs */

    struct sol_timeout *find_timeout;
    struct sol_timeout *update_schedule_timeout;

    struct sol_oic_client *client;

    const char *rt;
    char device_id[DEVICE_ID_LEN];
    struct sol_ptr_vector scanned_ids;
};

struct server_resource {
    struct sol_flow_node *node;
    const struct server_resource_funcs *funcs;

    struct sol_oic_server_resource *resource;
    struct sol_timeout *update_schedule_timeout;

    struct sol_oic_resource_type type;
};

static struct sol_network_link_addr multicast_ipv4, multicast_ipv6_local, multicast_ipv6_site;

static bool
initialize_multicast_addresses_once(void)
{
    static bool multicast_addresses_initialized = false;

    if (multicast_addresses_initialized)
        return true;

    multicast_ipv4 = (struct sol_network_link_addr) { .family = SOL_NETWORK_FAMILY_INET, .port = DEFAULT_UDP_PORT };
    if (!sol_network_link_addr_from_str(&multicast_ipv4, MULTICAST_ADDRESS_IPv4)) {
        SOL_WRN("Could not parse multicast IP address");
        multicast_ipv4.family = SOL_NETWORK_FAMILY_UNSPEC;
    }
    multicast_ipv6_local = (struct sol_network_link_addr) { .family = SOL_NETWORK_FAMILY_INET6, .port = DEFAULT_UDP_PORT };
    if (!sol_network_link_addr_from_str(&multicast_ipv6_local, MULTICAST_ADDRESS_IPv6_LOCAL)) {
        SOL_WRN("Could not parse multicast IP address");
        return false;
    }
    multicast_ipv6_site = (struct sol_network_link_addr) { .family = SOL_NETWORK_FAMILY_INET6, .port = DEFAULT_UDP_PORT };
    if (!sol_network_link_addr_from_str(&multicast_ipv6_site, MULTICAST_ADDRESS_IPv6_SITE)) {
        SOL_WRN("Could not parse multicast IP address");
        return false;
    }

    multicast_addresses_initialized = true;
    return true;
}

static bool
client_resource_implements_type(struct sol_oic_resource *oic_res, const char *resource_type)
{
    struct sol_str_slice rt = SOL_STR_SLICE_STR(resource_type, strlen(resource_type));
    struct sol_str_slice *type;
    uint16_t idx;

    SOL_VECTOR_FOREACH_IDX(&oic_res->types, type, idx) {
        if (sol_str_slice_eq(*type, rt))
            return true;
    }

    return false;
}

static void
state_changed(void *data, enum sol_coap_response_code response_code, struct sol_oic_client *oic_cli, const struct sol_network_link_addr *cliaddr,
    const struct sol_oic_map_reader *repr_vec)
{
    struct client_resource *resource = data;

    if (!cliaddr || !repr_vec)
        return;

    if (!sol_network_link_addr_eq(cliaddr, &resource->resource->addr)) {
        SOL_BUFFER_DECLARE_STATIC(resaddr, SOL_NETWORK_INET_ADDR_STR_LEN);
        SOL_BUFFER_DECLARE_STATIC(respaddr, SOL_NETWORK_INET_ADDR_STR_LEN);

        if (!sol_network_link_addr_to_str(&resource->resource->addr, &resaddr)) {
            SOL_WRN("Could not convert network address to string");
            return;
        }
        if (!sol_network_link_addr_to_str(cliaddr, &respaddr)) {
            SOL_WRN("Could not convert network address to string");
            return;
        }

        SOL_WRN("Expecting response from %.*s, got from %.*s, ignoring",
            SOL_STR_SLICE_PRINT(sol_buffer_get_slice(&resaddr)),
            SOL_STR_SLICE_PRINT(sol_buffer_get_slice(&respaddr)));
        return;
    }

    if (resource->funcs->from_repr_vec(resource, repr_vec) > 0)
        resource->funcs->inform_flow(resource);
}

static bool
found_resource(void *data, struct sol_oic_client *oic_cli, struct sol_oic_resource *oic_res)
{
    struct client_resource *resource = data;
    int r;

    if (!oic_res) {
        SOL_WRN("resource discovery timeout");
        goto cancel;
    }

    /* Some OIC device sent this node a discovery response packet but node's already set up. */
    if (resource->resource) {
        SOL_DBG("Received discovery packet when resource already set up, ignoring");
        goto cancel;
    }

    if (memcmp(oic_res->device_id.data, resource->device_id, 16) != 0) {
        /* Not the droid we're looking for. */
        SOL_DBG("Received resource with an unknown device_id, ignoring");
        return true;
    }

    /* FIXME: Should this check move to sol-oic-client? Does it actually make sense? */
    if (resource->rt && !client_resource_implements_type(oic_res, resource->rt)) {
        SOL_DBG("Received resource that does not implement rt=%s, ignoring", resource->rt);
        return true;
    }

    SOL_INF("Found resource matching device_id");
    resource->resource = sol_oic_resource_ref(oic_res);

    if (sol_oic_client_resource_set_observable(oic_cli, resource->resource,
        state_changed, resource, true) < 0) {
        SOL_WRN("Could not observe resource as requested, will try again");
    }

    r = sol_flow_send_bool_packet(resource->node,
        resource->funcs->found_port, true);
    if (r < 0)
        SOL_WRN("Could not send flow packet, will try again");

cancel:
    resource->find_timeout = NULL;
    memset(&resource->discover_pending, 0, sizeof(resource->discover_pending));
    return false;
}

static void
find_resources(struct client_resource *resource,
    bool (*resource_found_cb)(void *data,
    struct sol_oic_client *cli,
    struct sol_oic_resource *res),
    bool scan)
{
    struct pending *pending;

    if (scan)
        pending = &resource->scan_pending;
    else
        pending = &resource->discover_pending;

    if (multicast_ipv4.family != SOL_NETWORK_FAMILY_UNSPEC) {
        if (pending->ipv4)
            sol_oic_pending_cancel(pending->ipv4);
        pending->ipv4 = sol_oic_client_find_resources(resource->client,
            &multicast_ipv4, resource->rt, NULL, resource_found_cb, resource);
    }

    if (pending->ipv6_local)
        sol_oic_pending_cancel(pending->ipv6_local);
    pending->ipv6_local = sol_oic_client_find_resources(resource->client,
        &multicast_ipv6_local, resource->rt, NULL, resource_found_cb, resource);

    if (pending->ipv6_site)
        sol_oic_pending_cancel(pending->ipv6_site);
    pending->ipv6_site = sol_oic_client_find_resources(resource->client,
        &multicast_ipv6_site, resource->rt, NULL, resource_found_cb, resource);
}

static void
send_discovery_packets(struct client_resource *resource)
{
    if (resource->resource)
        return;

    sol_flow_send_bool_packet(resource->node, resource->funcs->found_port, false);

    find_resources(resource, found_resource, false);
}

static bool
find_timer(void *data)
{
    struct client_resource *resource = data;

    if (resource->resource) {
        SOL_INF("Timer expired when node already configured; disabling");
        resource->find_timeout = NULL;
        return false;
    }

    send_discovery_packets(resource);
    return true;
}

static inline char
base16_encode_digit(const uint8_t nibble, const char a)
{
    if (likely(nibble < 10))
        return '0' + nibble;
    return a + (nibble - 10);
}

static void
binary_to_hex_ascii(const char *binary, char *ascii)
{
    const uint8_t *input = (const uint8_t *)binary;
    size_t i, o = 0;

    for (i = 0; i < DEVICE_ID_LEN; i++) {
        const uint8_t b = input[i];
        uint8_t n;

        const uint8_t nibble[2] = {
            (b & 0xf0) >> 4,
            (b & 0x0f)
        };

        for (n = 0; n < 2; n++)
            ascii[o++] = base16_encode_digit(nibble[n], 'a');
    }

    ascii[o] = 0;
}

static bool
scan_callback(void *data, struct sol_oic_client *oic_cli, struct sol_oic_resource *oic_res)
{
    struct client_resource *resource = data;
    char ascii[DEVICE_ID_LEN * 2 + 1];
    char *id;
    uint16_t i;
    int r;

    if (!oic_res) {
        SOL_WRN("Scanning timeout");
        goto cancel;
    }

    /* FIXME: Should this check move to sol-oic-client? Does it actually make sense? */
    if (resource->rt && !client_resource_implements_type(oic_res, resource->rt)) {
        SOL_DBG("Received resource that does not implement rt=%s, ignoring", resource->rt);
        return true;
    }

    SOL_PTR_VECTOR_FOREACH_IDX(&resource->scanned_ids, id, i)
        if (memcmp(id, oic_res->device_id.data, DEVICE_ID_LEN) == 0)
            return true;

    id = malloc(DEVICE_ID_LEN);
    SOL_NULL_CHECK(id, true);
    memcpy(id, oic_res->device_id.data, DEVICE_ID_LEN);
    r = sol_ptr_vector_append(&resource->scanned_ids, id);
    SOL_INT_CHECK_GOTO(r, < 0, error);

    binary_to_hex_ascii(oic_res->device_id.data, ascii);

    r = sol_flow_send_string_packet(resource->node,
        resource->funcs->device_id_port, ascii);
    if (r < 0)
        SOL_WRN("Could not send server id.");

    return true;

error:
    SOL_WRN("Failed to process id.");
    free(id);
    return true;

cancel:
    memset(&resource->discover_pending, 0, sizeof(resource->discover_pending));
    return false;
}

static void
clear_scanned_ids(struct sol_ptr_vector *scanned_ids)
{
    char *id;
    uint16_t i;

    SOL_PTR_VECTOR_FOREACH_IDX(scanned_ids, id, i)
        free(id);
    sol_ptr_vector_clear(scanned_ids);
}

static void
send_scan_packets(struct client_resource *resource)
{
    clear_scanned_ids(&resource->scanned_ids);
    find_resources(resource, scan_callback, true);
}

static bool
server_resource_perform_update(void *data)
{
    struct server_resource *resource = data;
    struct sol_oic_response *notification;

    SOL_NULL_CHECK(resource->funcs->to_repr_vec, false);

    notification = sol_oic_server_notification_new(resource->resource);
    SOL_NULL_CHECK(notification, false);
    if (!resource->funcs->to_repr_vec(resource,
        sol_oic_server_response_get_writer(notification))) {
        sol_oic_server_response_free(notification);
        return false;
    }

    if (sol_oic_server_notify(notification) < 0) {
        SOL_WRN("Error while serializing update message");
    } else {
        resource->funcs->inform_flow(resource);
    }

    resource->update_schedule_timeout = NULL;
    return false;
}

static void
server_resource_schedule_update(struct server_resource *resource)
{
    if (resource->update_schedule_timeout)
        return;

    resource->update_schedule_timeout = sol_timeout_add(UPDATE_TIMEOUT_MS,
        server_resource_perform_update, resource);
}

static int
server_handle_update(void *data, struct sol_oic_request *request)
{
    enum sol_coap_response_code code;
    struct server_resource *resource = (struct server_resource *)data;
    struct sol_oic_map_reader *input;
    int r;

    if (!resource->funcs->from_repr_vec) {
        code = SOL_COAP_RESPONSE_CODE_NOT_IMPLEMENTED;
        goto end;
    }

    input = sol_oic_server_request_get_reader(request);
    r = resource->funcs->from_repr_vec(resource, input);
    if (r > 0) {
        server_resource_schedule_update(resource);
        code = SOL_COAP_RESPONSE_CODE_CHANGED;
    } else if (r == 0)
        code = SOL_COAP_RESPONSE_CODE_OK;
    else
        code = SOL_COAP_RESPONSE_CODE_PRECONDITION_FAILED;

end:
    return sol_oic_server_send_response(request, NULL, code);
}

static int
server_handle_get(void *data, struct sol_oic_request *request)
{
    const struct server_resource *resource = data;
    struct sol_oic_map_writer *output;
    struct sol_oic_response *response;

    if (!resource->funcs->to_repr_vec)
        return sol_oic_server_send_response(request, NULL,
            SOL_COAP_RESPONSE_CODE_NOT_IMPLEMENTED);

    response = sol_oic_server_response_new(request);
    SOL_NULL_CHECK_GOTO(response, error);
    output = sol_oic_server_response_get_writer(response);
    SOL_NULL_CHECK_GOTO(output, error);
    if (!resource->funcs->to_repr_vec((void *)resource, output))
        goto error;

    return sol_oic_server_send_response(request, response,
        SOL_COAP_RESPONSE_CODE_CONTENT);

error:
    sol_oic_server_response_free(response);
    return sol_oic_server_send_response(request, NULL,
        SOL_COAP_RESPONSE_CODE_INTERNAL_ERROR);
}

// log_init() implementation happens within oic-gen.c
static void log_init(void);

static int
server_resource_init(struct server_resource *resource, struct sol_flow_node *node,
    struct sol_str_slice resource_type, const struct server_resource_funcs *funcs)
{
    log_init();

    resource->node = node;
    resource->update_schedule_timeout = NULL;
    resource->funcs = funcs;

    resource->type = (struct sol_oic_resource_type) {
        SOL_SET_API_VERSION(.api_version = SOL_OIC_RESOURCE_TYPE_API_VERSION, )
        .resource_type = resource_type,
        .interface = SOL_STR_SLICE_LITERAL("oc.mi.def"),
        .get = { .handle = server_handle_get },
        .put = { .handle = server_handle_update },
        .post = { .handle = server_handle_update },
    };

    resource->resource = sol_oic_server_register_resource(&resource->type,
        resource, SOL_OIC_FLAG_DISCOVERABLE | SOL_OIC_FLAG_OBSERVABLE | SOL_OIC_FLAG_ACTIVE);
    if (resource->resource)
        return 0;

    return -EINVAL;
}

static void
server_resource_close(struct server_resource *resource)
{
    if (resource->update_schedule_timeout)
        sol_timeout_del(resource->update_schedule_timeout);
    sol_oic_server_unregister_resource(resource->resource);
}

static unsigned int
as_nibble(const char c)
{
    if (c >= '0' && c <= '9')
        return c - '0';
    if (c >= 'a' && c <= 'f')
        return c - 'a' + 10;
    if (c >= 'A' && c <= 'F')
        return c - 'A' + 10;

    SOL_WRN("Invalid hex character: %d", c);
    return 0;
}

static void
hex_ascii_to_binary(const char *ascii, char *binary)
{
    const char *p;
    size_t i;

    for (p = ascii, i = 0; i < DEVICE_ID_LEN; i++, p += 2)
        binary[i] = as_nibble(*p) << 4 | as_nibble(*(p + 1));
}

static int
client_connect(struct client_resource *resource, const char *device_id)
{
    if (!device_id || strlen(device_id) != 32) {
        SOL_DBG("Invalid or empty device_id. Not trying to connect.");
        return 0;
    }

    hex_ascii_to_binary(device_id, resource->device_id);
    SOL_NULL_CHECK(resource->device_id, -ENOMEM);

    if (resource->find_timeout)
        sol_timeout_del(resource->find_timeout);

    if (resource->resource) {
        if (sol_oic_client_resource_set_observable(resource->client,
            resource->resource, NULL, NULL, false) < 0) {
            SOL_WRN("Could not unobserve resource");
        }

        sol_oic_resource_unref(resource->resource);
        resource->resource = NULL;
    }

    SOL_INF("Sending multicast packets to find resource with device_id %s (rt=%s)",
        device_id, resource->rt);
    resource->find_timeout = sol_timeout_add(FIND_PERIOD_MS, find_timer, resource);
    if (resource->find_timeout) {
        /* Perform a find now instead of waiting FIND_PERIOD_MS the first time.  If the
         * resource is found in the mean time, the timeout will be automatically disabled. */
        send_discovery_packets(resource);
        return 0;
    }

    SOL_ERR("Could not create timeout to find resource");
    return -ENOMEM;
}

static int
client_resource_init(struct sol_flow_node *node, struct client_resource *resource, const char *resource_type,
    const struct client_resource_funcs *funcs)
{
    log_init();

    if (!initialize_multicast_addresses_once()) {
        SOL_ERR("Could not initialize multicast addresses");
        return -ENOTCONN;
    }

    assert(resource_type);

    resource->client = sol_oic_client_new();
    SOL_NULL_CHECK(resource->client, -ENOMEM);

    sol_ptr_vector_init(&resource->scanned_ids);
    resource->node = node;
    resource->find_timeout = NULL;
    resource->update_schedule_timeout = NULL;
    resource->resource = NULL;
    resource->funcs = funcs;
    resource->rt = resource_type;

    return 0;
}

static void
pending_free(struct pending *pending)
{
    if (pending->ipv4)
        sol_oic_pending_cancel(pending->ipv4);
    if (pending->ipv6_local)
        sol_oic_pending_cancel(pending->ipv6_local);
    if (pending->ipv6_site)
        sol_oic_pending_cancel(pending->ipv6_site);
}

static void
client_resource_close(struct client_resource *resource)
{
    if (resource->find_timeout)
        sol_timeout_del(resource->find_timeout);
    if (resource->update_schedule_timeout)
        sol_timeout_del(resource->update_schedule_timeout);

    pending_free(&resource->scan_pending);
    pending_free(&resource->discover_pending);

    if (resource->resource) {
        if (sol_oic_client_resource_set_observable(resource->client,
            resource->resource, NULL, NULL, false) < 0)
            SOL_WRN("Could not unobserve resource");

        sol_oic_resource_unref(resource->resource);
    }

    clear_scanned_ids(&resource->scanned_ids);
    sol_oic_client_del(resource->client);
}

static void
client_resource_update_ack(void *data, enum sol_coap_response_code response_code, struct sol_oic_client *cli, const struct sol_network_link_addr *addr,
    const struct sol_oic_map_reader *repr_vec)
{
    struct client_resource *resource = data;

    resource->funcs->inform_flow(resource);
}

static bool
client_resource_perform_update(void *data)
{
    struct client_resource *resource = data;
    struct sol_oic_request *request;
    struct sol_oic_pending *pending;

    SOL_NULL_CHECK_GOTO(resource->resource, disable_timeout);
    SOL_NULL_CHECK_GOTO(resource->funcs->to_repr_vec, disable_timeout);

    request = sol_oic_client_request_new(SOL_COAP_METHOD_PUT, resource->resource);
    if (!request ||
        !resource->funcs->to_repr_vec(resource,
        sol_oic_client_request_get_writer(request))) {
        SOL_WRN("Failed to create request. Will try again");
        return true;
    }

    pending = sol_oic_client_request(resource->client, request,
        client_resource_update_ack, data);
    if (!pending) {
        SOL_WRN("Could not send update request to resource, will try again");
        return true;
    }

disable_timeout:
    resource->update_schedule_timeout = NULL;
    return false;
}

static void
client_resource_schedule_update(struct client_resource *resource)
{
    if (resource->update_schedule_timeout)
        return;

    resource->update_schedule_timeout = sol_timeout_add(UPDATE_TIMEOUT_MS,
        client_resource_perform_update, resource);
}

static int
scan(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct client_resource *resource = data;

    send_scan_packets(resource);

    return 0;
}

static int
device_id_process(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct client_resource *resource = data;
    const char *device_id;
    int r;

    r = sol_flow_packet_get_string(packet, &device_id);
    SOL_INT_CHECK(r, < 0, r);

    return client_connect(resource, device_id);
}

static inline bool
check_updated_string(const char *a, const char *b)
{
    if (a && b)
        return strcmp(a, b) != 0;
    else if ((a && !b) || (!a && b))
       return true;
    else
       return false;
}

static inline bool
check_updated_int32(const int32_t a, const int32_t b)
{
    return a != b;
}

static inline bool
check_updated_boolean(const bool a, const bool b)
{
    return a != b;
}

static inline bool
check_updated_number(const double a, const double b)
{
    return !sol_util_double_eq(a, b);
}

static inline int
send_string_packet(struct sol_flow_node *src, uint16_t src_port, const char *value)
{
    return sol_flow_send_string_packet(src, src_port, value ? value : "");
}

#define RETURN_ERROR(errcode) do { ret = errcode; goto out; } while(0)
struct oic_state_light {
    bool state; /* True if on, False if off */
int32_t power; /* Current sensed or set value for Brightness */
char * name; /* Friendly device name (e.g. Kitchen Light) */
};
static int
oic_state_light_from_repr_vec(struct oic_state_light *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_light fields = {
        .state = state->state,
        .power = state->power,
        .name = state->name ? strdup(state->name) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "state")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.state = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "power")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.power = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.power = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.power = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "name")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.name, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->state, fields.state)) {
        state->state = fields.state;
        updated = true;
    }

    if (check_updated_int32(state->power, fields.power)) {
        state->power = fields.power;
        updated = true;
    }

    if (check_updated_string(state->name, fields.name)) {
        free(state->name);
        state->name = fields.name;
        fields.name = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.name);
    return ret;
}
/* User’s current activity type */
enum oic_state_activity_Type { OIC_STATE_ACTIVITY_TYPE_SLEEP, OIC_STATE_ACTIVITY_TYPE_SIT, OIC_STATE_ACTIVITY_TYPE_STAND, OIC_STATE_ACTIVITY_TYPE_WALK, OIC_STATE_ACTIVITY_TYPE_RUN };
static const struct sol_str_table oic_state_activity_Type_tbl[] = {
    SOL_STR_TABLE_ITEM("sleep", OIC_STATE_ACTIVITY_TYPE_SLEEP),
SOL_STR_TABLE_ITEM("sit", OIC_STATE_ACTIVITY_TYPE_SIT),
SOL_STR_TABLE_ITEM("stand", OIC_STATE_ACTIVITY_TYPE_STAND),
SOL_STR_TABLE_ITEM("walk", OIC_STATE_ACTIVITY_TYPE_WALK),
SOL_STR_TABLE_ITEM("run", OIC_STATE_ACTIVITY_TYPE_RUN),
    { }
};
struct oic_state_activity {
    int32_t Acalory; /* Total used calories per day that calculated by user’s activity */
int32_t Atime; /* Accumulated active time per day within moving condition (stand, walk, run) */
enum oic_state_activity_Type Type; /* User’s current activity type */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_activity_from_repr_vec(struct oic_state_activity *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_activity fields = {
        .Acalory = state->Acalory,
        .Atime = state->Atime,
        .Type = state->Type,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "Acalory")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.Acalory = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.Acalory = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.Acalory = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "Atime")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.Atime = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.Atime = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.Atime = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "Type")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_activity_Type_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.Type = (enum oic_state_activity_Type)val;
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<3);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->Acalory, fields.Acalory)) {
        state->Acalory = fields.Acalory;
        updated = true;
    }

    if (check_updated_int32(state->Atime, fields.Atime)) {
        state->Atime = fields.Atime;
        updated = true;
    }

    if (check_updated_int32(state->Type, fields.Type)) {
        state->Type = fields.Type;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_altimeter {
    double alt; /* ReadOnly. The current height of the position (metres) */
};
static int
oic_state_altimeter_from_repr_vec(struct oic_state_altimeter *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_altimeter fields = {
        .alt = state->alt,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "alt")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.alt = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.alt = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->alt, fields.alt)) {
        state->alt = fields.alt;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_audio {
    bool mute; /* Mute setting of an audio rendering device */
int32_t volume; /* Volume setting of an audio rendering device. */
};
static int
oic_state_audio_from_repr_vec(struct oic_state_audio *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_audio fields = {
        .mute = state->mute,
        .volume = state->volume,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "mute")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.mute = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "volume")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.volume = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.volume = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.volume = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->mute, fields.mute)) {
        state->mute = fields.mute;
        updated = true;
    }

    if (check_updated_int32(state->volume, fields.volume)) {
        state->volume = fields.volume;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_autofocus {
    bool autoFocus; /* Status of the Auto Focus */
};
static int
oic_state_autofocus_from_repr_vec(struct oic_state_autofocus *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_autofocus fields = {
        .autoFocus = state->autoFocus,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "autoFocus")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.autoFocus = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->autoFocus, fields.autoFocus)) {
        state->autoFocus = fields.autoFocus;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_automaticdocumentfeeder {
    char * adfStates; /* Comma separated list of the possible adf states. */
char * currentAdfState; /* Current adf state. */
};
static int
oic_state_automaticdocumentfeeder_from_repr_vec(struct oic_state_automaticdocumentfeeder *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_automaticdocumentfeeder fields = {
        .adfStates = state->adfStates ? strdup(state->adfStates) : NULL,
        .currentAdfState = state->currentAdfState ? strdup(state->currentAdfState) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "adfStates")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.adfStates, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "currentAdfState")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.currentAdfState, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_string(state->adfStates, fields.adfStates)) {
        free(state->adfStates);
        state->adfStates = fields.adfStates;
        fields.adfStates = NULL;
        updated = true;
    }

    if (check_updated_string(state->currentAdfState, fields.currentAdfState)) {
        free(state->currentAdfState);
        state->currentAdfState = fields.currentAdfState;
        fields.currentAdfState = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.adfStates);
    free(fields.currentAdfState);
    return ret;
}
/* Units used to measure blood sugar levels */
enum oic_state_blood_glucose_units { OIC_STATE_BLOOD_GLUCOSE_UNITS_MG_DL, OIC_STATE_BLOOD_GLUCOSE_UNITS_MMOL_L };
static const struct sol_str_table oic_state_blood_glucose_units_tbl[] = {
    SOL_STR_TABLE_ITEM("mg_dL", OIC_STATE_BLOOD_GLUCOSE_UNITS_MG_DL),
SOL_STR_TABLE_ITEM("mmol_L", OIC_STATE_BLOOD_GLUCOSE_UNITS_MMOL_L),
    { }
};
struct oic_state_blood_glucose {
    double bloodsugar; /* The concentration of glucose in the blood */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
char * site; /* Lot number of strip */
enum oic_state_blood_glucose_units units; /* Units used to measure blood sugar levels */
};
static int
oic_state_blood_glucose_from_repr_vec(struct oic_state_blood_glucose *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_blood_glucose fields = {
        .bloodsugar = state->bloodsugar,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
        .site = state->site ? strdup(state->site) : NULL,
        .units = state->units,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "bloodsugar")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.bloodsugar = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.bloodsugar = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "site")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.site, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_blood_glucose_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_blood_glucose_units)val;
            decode_mask &= ~(1<<3);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->bloodsugar, fields.bloodsugar)) {
        state->bloodsugar = fields.bloodsugar;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }

    if (check_updated_string(state->site, fields.site)) {
        free(state->site);
        state->site = fields.site;
        fields.site = NULL;
        updated = true;
    }

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    free(fields.site);
    return ret;
}

struct oic_state_blood_o2_saturation {
    double perfusion; /* Ratio of AC over DC of PPG */
double spo2; /* Oxygen saturation level in percentage */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_blood_o2_saturation_from_repr_vec(struct oic_state_blood_o2_saturation *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_blood_o2_saturation fields = {
        .perfusion = state->perfusion,
        .spo2 = state->spo2,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "perfusion")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.perfusion = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.perfusion = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "spo2")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.spo2 = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.spo2 = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->perfusion, fields.perfusion)) {
        state->perfusion = fields.perfusion;
        updated = true;
    }

    if (check_updated_number(state->spo2, fields.spo2)) {
        state->spo2 = fields.spo2;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_blood_pressure {
    double map; /* measured blood pressure in mmHG. MAP(Mean Arterial Pressure) is a value of the integral of one cycle of the blood pressure curve divided by the period between successive heart beats. */
char * bloodpressure; /* measured blood pressure in mmHG. It uses CSV (data type) containing systolic, diastolic values. */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_blood_pressure_from_repr_vec(struct oic_state_blood_pressure *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_blood_pressure fields = {
        .map = state->map,
        .bloodpressure = state->bloodpressure ? strdup(state->bloodpressure) : NULL,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "map")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.map = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.map = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "bloodpressure")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.bloodpressure, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->map, fields.map)) {
        state->map = fields.map;
        updated = true;
    }

    if (check_updated_string(state->bloodpressure, fields.bloodpressure)) {
        free(state->bloodpressure);
        state->bloodpressure = fields.bloodpressure;
        fields.bloodpressure = NULL;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.bloodpressure);
    free(fields.observedtime);
    return ret;
}

struct oic_state_bmi {
    double bmi; /* Body Mass Index of a person in kg/m^2 */
double bmip; /* BMI prime is a normalized BMI by 25 */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_bmi_from_repr_vec(struct oic_state_bmi *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_bmi fields = {
        .bmi = state->bmi,
        .bmip = state->bmip,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "bmi")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.bmi = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.bmi = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "bmip")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.bmip = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.bmip = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->bmi, fields.bmi)) {
        state->bmi = fields.bmi;
        updated = true;
    }

    if (check_updated_number(state->bmip, fields.bmip)) {
        state->bmip = fields.bmip;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* Unit of a measure of the obesity of a person */
enum oic_state_bodyfat_units { OIC_STATE_BODYFAT_UNITS_KG, OIC_STATE_BODYFAT_UNITS_LB, OIC_STATE_BODYFAT_UNITS__ };
static const struct sol_str_table oic_state_bodyfat_units_tbl[] = {
    SOL_STR_TABLE_ITEM("kg", OIC_STATE_BODYFAT_UNITS_KG),
SOL_STR_TABLE_ITEM("lb", OIC_STATE_BODYFAT_UNITS_LB),
SOL_STR_TABLE_ITEM("_", OIC_STATE_BODYFAT_UNITS__),
    { }
};
struct oic_state_bodyfat {
    enum oic_state_bodyfat_units units; /* Unit of a measure of the obesity of a person */
double bodyfat; /* The difference between the body weight and the fat free mass */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_bodyfat_from_repr_vec(struct oic_state_bodyfat *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_bodyfat fields = {
        .units = state->units,
        .bodyfat = state->bodyfat,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_bodyfat_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_bodyfat_units)val;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "bodyfat")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.bodyfat = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.bodyfat = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }

    if (check_updated_number(state->bodyfat, fields.bodyfat)) {
        state->bodyfat = fields.bodyfat;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* Unit of a person's height */
enum oic_state_bodyheight_units { OIC_STATE_BODYHEIGHT_UNITS_CM, OIC_STATE_BODYHEIGHT_UNITS_INCH };
static const struct sol_str_table oic_state_bodyheight_units_tbl[] = {
    SOL_STR_TABLE_ITEM("cm", OIC_STATE_BODYHEIGHT_UNITS_CM),
SOL_STR_TABLE_ITEM("inch", OIC_STATE_BODYHEIGHT_UNITS_INCH),
    { }
};
struct oic_state_bodyheight {
    enum oic_state_bodyheight_units units; /* Unit of a person's height */
double height; /* Height of an object */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_bodyheight_from_repr_vec(struct oic_state_bodyheight *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_bodyheight fields = {
        .units = state->units,
        .height = state->height,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_bodyheight_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_bodyheight_units)val;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "height")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.height = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.height = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }

    if (check_updated_number(state->height, fields.height)) {
        state->height = fields.height;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* Current body position type */
enum oic_state_body_position_bpos { OIC_STATE_BODY_POSITION_BPOS_STANDING, OIC_STATE_BODY_POSITION_BPOS_SITTING, OIC_STATE_BODY_POSITION_BPOS_SUPINE, OIC_STATE_BODY_POSITION_BPOS_PRONE, OIC_STATE_BODY_POSITION_BPOS_LEFT, OIC_STATE_BODY_POSITION_BPOS_RIGHT, OIC_STATE_BODY_POSITION_BPOS_FALL };
static const struct sol_str_table oic_state_body_position_bpos_tbl[] = {
    SOL_STR_TABLE_ITEM("standing", OIC_STATE_BODY_POSITION_BPOS_STANDING),
SOL_STR_TABLE_ITEM("sitting", OIC_STATE_BODY_POSITION_BPOS_SITTING),
SOL_STR_TABLE_ITEM("supine", OIC_STATE_BODY_POSITION_BPOS_SUPINE),
SOL_STR_TABLE_ITEM("prone", OIC_STATE_BODY_POSITION_BPOS_PRONE),
SOL_STR_TABLE_ITEM("left", OIC_STATE_BODY_POSITION_BPOS_LEFT),
SOL_STR_TABLE_ITEM("right", OIC_STATE_BODY_POSITION_BPOS_RIGHT),
SOL_STR_TABLE_ITEM("fall", OIC_STATE_BODY_POSITION_BPOS_FALL),
    { }
};
struct oic_state_body_position {
    enum oic_state_body_position_bpos bpos; /* Current body position type */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_body_position_from_repr_vec(struct oic_state_body_position *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_body_position fields = {
        .bpos = state->bpos,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "bpos")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_body_position_bpos_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.bpos = (enum oic_state_body_position_bpos)val;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->bpos, fields.bpos)) {
        state->bpos = fields.bpos;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_bodysite {
    int32_t bodysite; /* Nomenclature code referring a specific muscle group engaged. Refer to Annex C of IEEE Std 11073-10442-2008. */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_bodysite_from_repr_vec(struct oic_state_bodysite *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_bodysite fields = {
        .bodysite = state->bodysite,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "bodysite")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.bodysite = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.bodysite = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.bodysite = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->bodysite, fields.bodysite)) {
        state->bodysite = fields.bodysite;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* Unit of a body water of a person */
enum oic_state_bodywater_units { OIC_STATE_BODYWATER_UNITS_KG, OIC_STATE_BODYWATER_UNITS_LB, OIC_STATE_BODYWATER_UNITS__ };
static const struct sol_str_table oic_state_bodywater_units_tbl[] = {
    SOL_STR_TABLE_ITEM("kg", OIC_STATE_BODYWATER_UNITS_KG),
SOL_STR_TABLE_ITEM("lb", OIC_STATE_BODYWATER_UNITS_LB),
SOL_STR_TABLE_ITEM("_", OIC_STATE_BODYWATER_UNITS__),
    { }
};
struct oic_state_bodywater {
    enum oic_state_bodywater_units units; /* Unit of a body water of a person */
double bodywater; /* The total water of the human body */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_bodywater_from_repr_vec(struct oic_state_bodywater *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_bodywater fields = {
        .units = state->units,
        .bodywater = state->bodywater,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_bodywater_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_bodywater_units)val;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "bodywater")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.bodywater = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.bodywater = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }

    if (check_updated_number(state->bodywater, fields.bodywater)) {
        state->bodywater = fields.bodywater;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* Unit of a person's weight */
enum oic_state_bodyweight_units { OIC_STATE_BODYWEIGHT_UNITS_KG, OIC_STATE_BODYWEIGHT_UNITS_LB };
static const struct sol_str_table oic_state_bodyweight_units_tbl[] = {
    SOL_STR_TABLE_ITEM("kg", OIC_STATE_BODYWEIGHT_UNITS_KG),
SOL_STR_TABLE_ITEM("lb", OIC_STATE_BODYWEIGHT_UNITS_LB),
    { }
};
struct oic_state_bodyweight {
    double weight; /* ReadOnly. Weight of an object */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
enum oic_state_bodyweight_units units; /* Unit of a person's weight */
};
static int
oic_state_bodyweight_from_repr_vec(struct oic_state_bodyweight *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_bodyweight fields = {
        .weight = state->weight,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
        .units = state->units,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "weight")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.weight = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.weight = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_bodyweight_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_bodyweight_units)val;
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->weight, fields.weight)) {
        state->weight = fields.weight;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_clock {
    double countdown; /* Desired total seconds for countdown */
char * datetime; /* Using ISO 8601 datetime format (e.g: 2007-04-05T14:30Z, 2007-04-05T14:30+09:00) */
};
static int
oic_state_clock_from_repr_vec(struct oic_state_clock *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_clock fields = {
        .countdown = state->countdown,
        .datetime = state->datetime ? strdup(state->datetime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "countdown")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.countdown = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.countdown = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "datetime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.datetime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->countdown, fields.countdown)) {
        state->countdown = fields.countdown;
        updated = true;
    }

    if (check_updated_string(state->datetime, fields.datetime)) {
        free(state->datetime);
        state->datetime = fields.datetime;
        fields.datetime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.datetime);
    return ret;
}

struct oic_state_colour_autowhitebalance {
    bool autoWhiteBalance; /* Status of the Auto White balance */
};
static int
oic_state_colour_autowhitebalance_from_repr_vec(struct oic_state_colour_autowhitebalance *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_colour_autowhitebalance fields = {
        .autoWhiteBalance = state->autoWhiteBalance,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "autoWhiteBalance")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.autoWhiteBalance = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->autoWhiteBalance, fields.autoWhiteBalance)) {
        state->autoWhiteBalance = fields.autoWhiteBalance;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_colour_saturation {
    int32_t colourSaturation; /* The colour saturation value */
};
static int
oic_state_colour_saturation_from_repr_vec(struct oic_state_colour_saturation *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_colour_saturation fields = {
        .colourSaturation = state->colourSaturation,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "colourSaturation")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.colourSaturation = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.colourSaturation = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.colourSaturation = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->colourSaturation, fields.colourSaturation)) {
        state->colourSaturation = fields.colourSaturation;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_cycling_power {
    bool leftonly; /* if power meters only measure power from the left side */
double al_power; /* Left side power */
double ar_power; /* Right side power */
double ftp; /* Functional Threshold Power (FTP). Calculate from 20 min Time Trial average wattage * 0.95 */
double power; /* It’s measured in watts(w). It means how much work you are outputting while on the bike. */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_cycling_power_from_repr_vec(struct oic_state_cycling_power *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_cycling_power fields = {
        .leftonly = state->leftonly,
        .al_power = state->al_power,
        .ar_power = state->ar_power,
        .ftp = state->ftp,
        .power = state->power,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "leftonly")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.leftonly = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "al_power")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.al_power = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.al_power = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "ar_power")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.ar_power = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.ar_power = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "ftp")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.ftp = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.ftp = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<3);
            continue;
        }
        if (decode_mask & (1<<4) && streq(field.key, "power")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.power = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.power = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<4);
            continue;
        }
        if (decode_mask & (1<<5) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<5);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->leftonly, fields.leftonly)) {
        state->leftonly = fields.leftonly;
        updated = true;
    }

    if (check_updated_number(state->al_power, fields.al_power)) {
        state->al_power = fields.al_power;
        updated = true;
    }

    if (check_updated_number(state->ar_power, fields.ar_power)) {
        state->ar_power = fields.ar_power;
        updated = true;
    }

    if (check_updated_number(state->ftp, fields.ftp)) {
        state->ftp = fields.ftp;
        updated = true;
    }

    if (check_updated_number(state->power, fields.power)) {
        state->power = fields.power;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* Unit of current speed */
enum oic_state_cycling_speedcadence_units { OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_MPH, OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_KM_H, OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_MACH, OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_KNOT, OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_M_MIN, OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_M_S, OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_FT_S, OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_CM_S, OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_YARD_S };
static const struct sol_str_table oic_state_cycling_speedcadence_units_tbl[] = {
    SOL_STR_TABLE_ITEM("mph", OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_MPH),
SOL_STR_TABLE_ITEM("km_h", OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_KM_H),
SOL_STR_TABLE_ITEM("Mach", OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_MACH),
SOL_STR_TABLE_ITEM("knot", OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_KNOT),
SOL_STR_TABLE_ITEM("m_min", OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_M_MIN),
SOL_STR_TABLE_ITEM("m_s", OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_M_S),
SOL_STR_TABLE_ITEM("ft_s", OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_FT_S),
SOL_STR_TABLE_ITEM("cm_s", OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_CM_S),
SOL_STR_TABLE_ITEM("yard_s", OIC_STATE_CYCLING_SPEEDCADENCE_UNITS_YARD_S),
    { }
};
struct oic_state_cycling_speedcadence {
    int32_t speed; /* current speed */
double cadence; /* Pedals per minutes */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
enum oic_state_cycling_speedcadence_units units; /* Unit of current speed */
};
static int
oic_state_cycling_speedcadence_from_repr_vec(struct oic_state_cycling_speedcadence *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_cycling_speedcadence fields = {
        .speed = state->speed,
        .cadence = state->cadence,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
        .units = state->units,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "speed")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.speed = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.speed = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.speed = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "cadence")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.cadence = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.cadence = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_cycling_speedcadence_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_cycling_speedcadence_units)val;
            decode_mask &= ~(1<<3);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->speed, fields.speed)) {
        state->speed = fields.speed;
        updated = true;
    }

    if (check_updated_number(state->cadence, fields.cadence)) {
        state->cadence = fields.cadence;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* Unit of current distance or length */
enum oic_state_distance_units { OIC_STATE_DISTANCE_UNITS_KILOMETER, OIC_STATE_DISTANCE_UNITS_METER, OIC_STATE_DISTANCE_UNITS_DECIMETER, OIC_STATE_DISTANCE_UNITS_CENTIMETER, OIC_STATE_DISTANCE_UNITS_MILLIMETER, OIC_STATE_DISTANCE_UNITS_MICROMETER, OIC_STATE_DISTANCE_UNITS_NANOMETER, OIC_STATE_DISTANCE_UNITS_ANGSTROM, OIC_STATE_DISTANCE_UNITS_MILE, OIC_STATE_DISTANCE_UNITS_YARD, OIC_STATE_DISTANCE_UNITS_FEET };
static const struct sol_str_table oic_state_distance_units_tbl[] = {
    SOL_STR_TABLE_ITEM("kilometer", OIC_STATE_DISTANCE_UNITS_KILOMETER),
SOL_STR_TABLE_ITEM("meter", OIC_STATE_DISTANCE_UNITS_METER),
SOL_STR_TABLE_ITEM("decimeter", OIC_STATE_DISTANCE_UNITS_DECIMETER),
SOL_STR_TABLE_ITEM("centimeter", OIC_STATE_DISTANCE_UNITS_CENTIMETER),
SOL_STR_TABLE_ITEM("millimeter", OIC_STATE_DISTANCE_UNITS_MILLIMETER),
SOL_STR_TABLE_ITEM("micrometer", OIC_STATE_DISTANCE_UNITS_MICROMETER),
SOL_STR_TABLE_ITEM("nanometer", OIC_STATE_DISTANCE_UNITS_NANOMETER),
SOL_STR_TABLE_ITEM("angstrom", OIC_STATE_DISTANCE_UNITS_ANGSTROM),
SOL_STR_TABLE_ITEM("mile", OIC_STATE_DISTANCE_UNITS_MILE),
SOL_STR_TABLE_ITEM("yard", OIC_STATE_DISTANCE_UNITS_YARD),
SOL_STR_TABLE_ITEM("feet", OIC_STATE_DISTANCE_UNITS_FEET),
    { }
};
struct oic_state_distance {
    double distance; /* Current distance or length */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
enum oic_state_distance_units units; /* Unit of current distance or length */
};
static int
oic_state_distance_from_repr_vec(struct oic_state_distance *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_distance fields = {
        .distance = state->distance,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
        .units = state->units,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "distance")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.distance = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.distance = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_distance_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_distance_units)val;
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->distance, fields.distance)) {
        state->distance = fields.distance;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* The state of the door (open or closed) */
enum oic_state_door_openState { OIC_STATE_DOOR_OPENSTATE_OPEN, OIC_STATE_DOOR_OPENSTATE_CLOSED };
static const struct sol_str_table oic_state_door_openState_tbl[] = {
    SOL_STR_TABLE_ITEM("Open", OIC_STATE_DOOR_OPENSTATE_OPEN),
SOL_STR_TABLE_ITEM("Closed", OIC_STATE_DOOR_OPENSTATE_CLOSED),
    { }
};
struct oic_state_door {
    bool openAlarm; /* The state of the door open alarm */
enum oic_state_door_openState openState; /* The state of the door (open or closed) */
char * openDuration; /* The time duration the door has been open */
};
static int
oic_state_door_from_repr_vec(struct oic_state_door *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_door fields = {
        .openAlarm = state->openAlarm,
        .openState = state->openState,
        .openDuration = state->openDuration ? strdup(state->openDuration) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "openAlarm")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.openAlarm = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "openState")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_door_openState_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.openState = (enum oic_state_door_openState)val;
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "openDuration")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.openDuration, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->openAlarm, fields.openAlarm)) {
        state->openAlarm = fields.openAlarm;
        updated = true;
    }

    if (check_updated_int32(state->openState, fields.openState)) {
        state->openState = fields.openState;
        updated = true;
    }

    if (check_updated_string(state->openDuration, fields.openDuration)) {
        free(state->openDuration);
        state->openDuration = fields.openDuration;
        fields.openDuration = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.openDuration);
    return ret;
}

struct oic_state_emg {
    double mi; /* The value of Muscle intensity that generated by skeletal muscles. */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_emg_from_repr_vec(struct oic_state_emg *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_emg fields = {
        .mi = state->mi,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "mi")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.mi = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.mi = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->mi, fields.mi)) {
        state->mi = fields.mi;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_energy_battery {
    int32_t charge; /* The current charge percentage. */
};
static int
oic_state_energy_battery_from_repr_vec(struct oic_state_energy_battery *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_energy_battery fields = {
        .charge = state->charge,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "charge")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.charge = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.charge = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.charge = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->charge, fields.charge)) {
        state->charge = fields.charge;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_energy_consumption {
    double energy; /* Energy consumed */
double power; /* Instantaneous Power */
};
static int
oic_state_energy_consumption_from_repr_vec(struct oic_state_energy_consumption *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_energy_consumption fields = {
        .energy = state->energy,
        .power = state->power,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "energy")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.energy = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.energy = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "power")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.power = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.power = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->energy, fields.energy)) {
        state->energy = fields.energy;
        updated = true;
    }

    if (check_updated_number(state->power, fields.power)) {
        state->power = fields.power;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_energy_drlc {
    bool override; /* Whether the consumer has overriden the application of DR */
int32_t DRType; /* The to be applied demand-response type */
int32_t duration; /* The duration of the to be applied DR type */
char * start; /* The start time for the application of DR */
};
static int
oic_state_energy_drlc_from_repr_vec(struct oic_state_energy_drlc *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_energy_drlc fields = {
        .override = state->override,
        .DRType = state->DRType,
        .duration = state->duration,
        .start = state->start ? strdup(state->start) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "override")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.override = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "DRType")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.DRType = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.DRType = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.DRType = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "duration")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.duration = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.duration = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.duration = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "start")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.start, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<3);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->override, fields.override)) {
        state->override = fields.override;
        updated = true;
    }

    if (check_updated_int32(state->DRType, fields.DRType)) {
        state->DRType = fields.DRType;
        updated = true;
    }

    if (check_updated_int32(state->duration, fields.duration)) {
        state->duration = fields.duration;
        updated = true;
    }

    if (check_updated_string(state->start, fields.start)) {
        free(state->start);
        state->start = fields.start;
        fields.start = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.start);
    return ret;
}

struct oic_state_energy_overload {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_energy_overload_from_repr_vec(struct oic_state_energy_overload *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_energy_overload fields = {
        .value = state->value,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "value")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.value = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->value, fields.value)) {
        state->value = fields.value;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_fev1 {
    double fev1; /* The expiratory volume of a subject under forced conditions at 1 second, measured from time zero. FEV1 is measured in liters */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_fev1_from_repr_vec(struct oic_state_fev1 *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_fev1 fields = {
        .fev1 = state->fev1,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "fev1")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.fev1 = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.fev1 = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->fev1, fields.fev1)) {
        state->fev1 = fields.fev1;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* Unit of a fat free mass of a person */
enum oic_state_ffm_units { OIC_STATE_FFM_UNITS_KG, OIC_STATE_FFM_UNITS_LB };
static const struct sol_str_table oic_state_ffm_units_tbl[] = {
    SOL_STR_TABLE_ITEM("kg", OIC_STATE_FFM_UNITS_KG),
SOL_STR_TABLE_ITEM("lb", OIC_STATE_FFM_UNITS_LB),
    { }
};
struct oic_state_ffm {
    double ffm; /* The sum of the soft lean mass and mineral mass */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
enum oic_state_ffm_units units; /* Unit of a fat free mass of a person */
};
static int
oic_state_ffm_from_repr_vec(struct oic_state_ffm *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_ffm fields = {
        .ffm = state->ffm,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
        .units = state->units,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "ffm")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.ffm = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.ffm = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_ffm_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_ffm_units)val;
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->ffm, fields.ffm)) {
        state->ffm = fields.ffm;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_gsr {
    double skincon; /* the value of skin conductance. unit is μS (micro siemens) */
double skincv; /* the value of skin conductance in voltage. unit is volt */
double skinres; /* the value of skin resistance. unit is Ω (ohm) */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_gsr_from_repr_vec(struct oic_state_gsr *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_gsr fields = {
        .skincon = state->skincon,
        .skincv = state->skincv,
        .skinres = state->skinres,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "skincon")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.skincon = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.skincon = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "skincv")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.skincv = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.skincv = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "skinres")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.skinres = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.skinres = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<3);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->skincon, fields.skincon)) {
        state->skincon = fields.skincon;
        updated = true;
    }

    if (check_updated_number(state->skincv, fields.skincv)) {
        state->skincv = fields.skincv;
        updated = true;
    }

    if (check_updated_number(state->skinres, fields.skinres)) {
        state->skinres = fields.skinres;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_heartrate {
    int32_t heartrate; /* measured heart rate in bpm. */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_heartrate_from_repr_vec(struct oic_state_heartrate *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_heartrate fields = {
        .heartrate = state->heartrate,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "heartrate")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.heartrate = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.heartrate = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.heartrate = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->heartrate, fields.heartrate)) {
        state->heartrate = fields.heartrate;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_height {
    double height; /* Height of an object */
};
static int
oic_state_height_from_repr_vec(struct oic_state_height *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_height fields = {
        .height = state->height,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "height")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.height = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.height = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->height, fields.height)) {
        state->height = fields.height;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_humidity {
    int32_t desiredHumidity; /* Desired value for Humidity */
int32_t humidity; /* Current sensed value for Humidity */
};
static int
oic_state_humidity_from_repr_vec(struct oic_state_humidity *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_humidity fields = {
        .desiredHumidity = state->desiredHumidity,
        .humidity = state->humidity,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "desiredHumidity")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.desiredHumidity = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.desiredHumidity = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.desiredHumidity = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "humidity")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.humidity = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.humidity = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.humidity = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->desiredHumidity, fields.desiredHumidity)) {
        state->desiredHumidity = fields.desiredHumidity;
        updated = true;
    }

    if (check_updated_int32(state->humidity, fields.humidity)) {
        state->humidity = fields.humidity;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}
/* Status of the Ice Maker */
enum oic_state_icemaker_status { OIC_STATE_ICEMAKER_STATUS_ON, OIC_STATE_ICEMAKER_STATUS_OFF, OIC_STATE_ICEMAKER_STATUS_FULL };
static const struct sol_str_table oic_state_icemaker_status_tbl[] = {
    SOL_STR_TABLE_ITEM("on", OIC_STATE_ICEMAKER_STATUS_ON),
SOL_STR_TABLE_ITEM("off", OIC_STATE_ICEMAKER_STATUS_OFF),
SOL_STR_TABLE_ITEM("full", OIC_STATE_ICEMAKER_STATUS_FULL),
    { }
};
struct oic_state_icemaker {
    enum oic_state_icemaker_status status; /* Status of the Ice Maker */
};
static int
oic_state_icemaker_from_repr_vec(struct oic_state_icemaker *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_icemaker fields = {
        .status = state->status,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "status")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_icemaker_status_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.status = (enum oic_state_icemaker_status)val;
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->status, fields.status)) {
        state->status = fields.status;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_light_brightness {
    int32_t brightness; /* Current sensed or set value for Brightness */
};
static int
oic_state_light_brightness_from_repr_vec(struct oic_state_light_brightness *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_light_brightness fields = {
        .brightness = state->brightness,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "brightness")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.brightness = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.brightness = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.brightness = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->brightness, fields.brightness)) {
        state->brightness = fields.brightness;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}
/* State of the lock. */
enum oic_state_lock_status_lockState { OIC_STATE_LOCK_STATUS_LOCKSTATE_LOCKED, OIC_STATE_LOCK_STATUS_LOCKSTATE_UNLOCKED };
static const struct sol_str_table oic_state_lock_status_lockState_tbl[] = {
    SOL_STR_TABLE_ITEM("Locked", OIC_STATE_LOCK_STATUS_LOCKSTATE_LOCKED),
SOL_STR_TABLE_ITEM("Unlocked", OIC_STATE_LOCK_STATUS_LOCKSTATE_UNLOCKED),
    { }
};
struct oic_state_lock_status {
    enum oic_state_lock_status_lockState lockState; /* State of the lock. */
};
static int
oic_state_lock_status_from_repr_vec(struct oic_state_lock_status *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_lock_status fields = {
        .lockState = state->lockState,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "lockState")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_lock_status_lockState_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.lockState = (enum oic_state_lock_status_lockState)val;
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->lockState, fields.lockState)) {
        state->lockState = fields.lockState;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_mode {
    char * modes; /* Comma separated list of the currently active mode(s) */
char * supportedModes; /* Comma separated list of possible modes the device supports. */
};
static int
oic_state_mode_from_repr_vec(struct oic_state_mode *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_mode fields = {
        .modes = state->modes ? strdup(state->modes) : NULL,
        .supportedModes = state->supportedModes ? strdup(state->supportedModes) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "modes")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.modes, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "supportedModes")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.supportedModes, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_string(state->modes, fields.modes)) {
        free(state->modes);
        state->modes = fields.modes;
        fields.modes = NULL;
        updated = true;
    }

    if (check_updated_string(state->supportedModes, fields.supportedModes)) {
        free(state->supportedModes);
        state->supportedModes = fields.supportedModes;
        fields.supportedModes = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.modes);
    free(fields.supportedModes);
    return ret;
}

struct oic_state_muscle_o2_saturation {
    double smo2; /* SMO2(Muscle Oxygen Saturation). It is the percentage of hemoglobin that is carrying oxygen in muscle tissue. */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_muscle_o2_saturation_from_repr_vec(struct oic_state_muscle_o2_saturation *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_muscle_o2_saturation fields = {
        .smo2 = state->smo2,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "smo2")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.smo2 = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.smo2 = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->smo2, fields.smo2)) {
        state->smo2 = fields.smo2;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_nightmode {
    bool nightMode; /* Status of the Night Mode */
};
static int
oic_state_nightmode_from_repr_vec(struct oic_state_nightmode *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_nightmode fields = {
        .nightMode = state->nightMode,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "nightMode")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.nightMode = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->nightMode, fields.nightMode)) {
        state->nightMode = fields.nightMode;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_operational_state {
    int32_t progressPercentage; /* Percentage completeness of the current jobState */
char * currentJobState; /* Currently active jobState */
char * currentMachineState; /* Current state of operation of the device. */
char * jobStates; /* Comma separate list of the possible job states. */
char * machineStates; /* Comma separated list of the possible operational states. */
char * remainingTime; /* Time till completion of the current operational state */
char * runningTime; /* Elapsed time in the current operational state */
};
static int
oic_state_operational_state_from_repr_vec(struct oic_state_operational_state *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_operational_state fields = {
        .progressPercentage = state->progressPercentage,
        .currentJobState = state->currentJobState ? strdup(state->currentJobState) : NULL,
        .currentMachineState = state->currentMachineState ? strdup(state->currentMachineState) : NULL,
        .jobStates = state->jobStates ? strdup(state->jobStates) : NULL,
        .machineStates = state->machineStates ? strdup(state->machineStates) : NULL,
        .remainingTime = state->remainingTime ? strdup(state->remainingTime) : NULL,
        .runningTime = state->runningTime ? strdup(state->runningTime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "progressPercentage")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.progressPercentage = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.progressPercentage = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.progressPercentage = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "currentJobState")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.currentJobState, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "currentMachineState")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.currentMachineState, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "jobStates")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.jobStates, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<3);
            continue;
        }
        if (decode_mask & (1<<4) && streq(field.key, "machineStates")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.machineStates, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<4);
            continue;
        }
        if (decode_mask & (1<<5) && streq(field.key, "remainingTime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.remainingTime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<5);
            continue;
        }
        if (decode_mask & (1<<6) && streq(field.key, "runningTime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.runningTime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<6);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->progressPercentage, fields.progressPercentage)) {
        state->progressPercentage = fields.progressPercentage;
        updated = true;
    }

    if (check_updated_string(state->currentJobState, fields.currentJobState)) {
        free(state->currentJobState);
        state->currentJobState = fields.currentJobState;
        fields.currentJobState = NULL;
        updated = true;
    }

    if (check_updated_string(state->currentMachineState, fields.currentMachineState)) {
        free(state->currentMachineState);
        state->currentMachineState = fields.currentMachineState;
        fields.currentMachineState = NULL;
        updated = true;
    }

    if (check_updated_string(state->jobStates, fields.jobStates)) {
        free(state->jobStates);
        state->jobStates = fields.jobStates;
        fields.jobStates = NULL;
        updated = true;
    }

    if (check_updated_string(state->machineStates, fields.machineStates)) {
        free(state->machineStates);
        state->machineStates = fields.machineStates;
        fields.machineStates = NULL;
        updated = true;
    }

    if (check_updated_string(state->remainingTime, fields.remainingTime)) {
        free(state->remainingTime);
        state->remainingTime = fields.remainingTime;
        fields.remainingTime = NULL;
        updated = true;
    }

    if (check_updated_string(state->runningTime, fields.runningTime)) {
        free(state->runningTime);
        state->runningTime = fields.runningTime;
        fields.runningTime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.currentJobState);
    free(fields.currentMachineState);
    free(fields.jobStates);
    free(fields.machineStates);
    free(fields.remainingTime);
    free(fields.runningTime);
    return ret;
}

struct oic_state_pef {
    double pef; /* Maximum flow measured at the mouth during an expiration delivered with maximal force starting immediately after achieving maximum lung inflation. PEF is measured in liters per minute. */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_pef_from_repr_vec(struct oic_state_pef *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_pef fields = {
        .pef = state->pef,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "pef")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.pef = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.pef = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->pef, fields.pef)) {
        state->pef = fields.pef;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_refrigeration {
    bool defrost; /* Indicates whether a defrost cycle is currently active */
bool rapidCool; /* Indicates whether the unit has a rapid cool capability active */
bool rapidFreeze; /* Indicates whether the unit has a rapid freeze capability active. */
int32_t filter; /* Percentage life time remaining for the water filter */
};
static int
oic_state_refrigeration_from_repr_vec(struct oic_state_refrigeration *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_refrigeration fields = {
        .defrost = state->defrost,
        .rapidCool = state->rapidCool,
        .rapidFreeze = state->rapidFreeze,
        .filter = state->filter,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "defrost")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.defrost = field.v_boolean;
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "rapidCool")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.rapidCool = field.v_boolean;
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "rapidFreeze")) {
            if (field.type != SOL_OIC_REPR_TYPE_BOOL)
                RETURN_ERROR(-EINVAL);
            fields.rapidFreeze = field.v_boolean;
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "filter")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.filter = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.filter = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.filter = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<3);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_boolean(state->defrost, fields.defrost)) {
        state->defrost = fields.defrost;
        updated = true;
    }

    if (check_updated_boolean(state->rapidCool, fields.rapidCool)) {
        state->rapidCool = fields.rapidCool;
        updated = true;
    }

    if (check_updated_boolean(state->rapidFreeze, fields.rapidFreeze)) {
        state->rapidFreeze = fields.rapidFreeze;
        updated = true;
    }

    if (check_updated_int32(state->filter, fields.filter)) {
        state->filter = fields.filter;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_repetition {
    int32_t repetition; /* An instance to measure the number of a certain repetitive workout. */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_repetition_from_repr_vec(struct oic_state_repetition *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_repetition fields = {
        .repetition = state->repetition,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "repetition")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.repetition = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.repetition = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.repetition = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->repetition, fields.repetition)) {
        state->repetition = fields.repetition;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_respirationrate {
    int32_t respirate; /* Respiration rate per minute */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_respirationrate_from_repr_vec(struct oic_state_respirationrate *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_respirationrate fields = {
        .respirate = state->respirate,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "respirate")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.respirate = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.respirate = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.respirate = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->respirate, fields.respirate)) {
        state->respirate = fields.respirate;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_sensor_carbondioxide {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_carbondioxide_from_repr_vec(struct oic_state_sensor_carbondioxide *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_carbondioxide */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor_carbonmonoxide {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_carbonmonoxide_from_repr_vec(struct oic_state_sensor_carbonmonoxide *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_carbonmonoxide */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor_contact {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_contact_from_repr_vec(struct oic_state_sensor_contact *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_contact */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor_geolocation {
    double accuracy; /* ReadOnly. The accuracy level of the latitude and longitude coordinates (metres) */
double alt; /* ReadOnly. The current height of the position (metres) */
double altitudeAccuracy; /* ReadOnly. The accuracy level of the altitude coordinates (metres) */
double heading; /* ReadOnly. Direction of travel of device (degree) */
double latitude; /* ReadOnly. Device's Current Latitude coordinate (degrees) */
double longitude; /* ReadOnly. Device's Current Longitude coordinate (degrees) */
double speed; /* ReadOnly. Device's current velocity (metres per second) */
};
static int
oic_state_sensor_geolocation_from_repr_vec(struct oic_state_sensor_geolocation *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_sensor_geolocation fields = {
        .accuracy = state->accuracy,
        .alt = state->alt,
        .altitudeAccuracy = state->altitudeAccuracy,
        .heading = state->heading,
        .latitude = state->latitude,
        .longitude = state->longitude,
        .speed = state->speed,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "accuracy")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.accuracy = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.accuracy = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "alt")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.alt = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.alt = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "altitudeAccuracy")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.altitudeAccuracy = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.altitudeAccuracy = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "heading")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.heading = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.heading = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<3);
            continue;
        }
        if (decode_mask & (1<<4) && streq(field.key, "latitude")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.latitude = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.latitude = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<4);
            continue;
        }
        if (decode_mask & (1<<5) && streq(field.key, "longitude")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.longitude = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.longitude = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<5);
            continue;
        }
        if (decode_mask & (1<<6) && streq(field.key, "speed")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.speed = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.speed = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<6);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->accuracy, fields.accuracy)) {
        state->accuracy = fields.accuracy;
        updated = true;
    }

    if (check_updated_number(state->alt, fields.alt)) {
        state->alt = fields.alt;
        updated = true;
    }

    if (check_updated_number(state->altitudeAccuracy, fields.altitudeAccuracy)) {
        state->altitudeAccuracy = fields.altitudeAccuracy;
        updated = true;
    }

    if (check_updated_number(state->heading, fields.heading)) {
        state->heading = fields.heading;
        updated = true;
    }

    if (check_updated_number(state->latitude, fields.latitude)) {
        state->latitude = fields.latitude;
        updated = true;
    }

    if (check_updated_number(state->longitude, fields.longitude)) {
        state->longitude = fields.longitude;
        updated = true;
    }

    if (check_updated_number(state->speed, fields.speed)) {
        state->speed = fields.speed;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_sensor_glassbreak {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_glassbreak_from_repr_vec(struct oic_state_sensor_glassbreak *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_glassbreak */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_from_repr_vec(struct oic_state_sensor *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor_motion {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_motion_from_repr_vec(struct oic_state_sensor_motion *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_motion */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor_presence {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_presence_from_repr_vec(struct oic_state_sensor_presence *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_presence */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor_radiation_uv {
    double measurement; /* the measured UV Index */
};
static int
oic_state_sensor_radiation_uv_from_repr_vec(struct oic_state_sensor_radiation_uv *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_sensor_radiation_uv fields = {
        .measurement = state->measurement,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "measurement")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.measurement = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.measurement = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->measurement, fields.measurement)) {
        state->measurement = fields.measurement;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}

struct oic_state_sensor_sleep {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_sleep_from_repr_vec(struct oic_state_sensor_sleep *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_sleep */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor_smoke {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_smoke_from_repr_vec(struct oic_state_sensor_smoke *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_smoke */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor_touch {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_touch_from_repr_vec(struct oic_state_sensor_touch *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_touch */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_sensor_water {
    bool value; /* true = sensed, false = not sensed. */
};
static int
oic_state_sensor_water_from_repr_vec(struct oic_state_sensor_water *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_sensor_water */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_signalstrength {
    double lqi; /* current value of Link Quality Indicator */
double rssi; /* current value of Received Signal Strength Indicator */
};
static int
oic_state_signalstrength_from_repr_vec(struct oic_state_signalstrength *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_signalstrength fields = {
        .lqi = state->lqi,
        .rssi = state->rssi,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "lqi")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.lqi = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.lqi = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "rssi")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.rssi = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.rssi = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->lqi, fields.lqi)) {
        state->lqi = fields.lqi;
        updated = true;
    }

    if (check_updated_number(state->rssi, fields.rssi)) {
        state->rssi = fields.rssi;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}
/* Sleep type categorized into awake, light, and deep */
enum oic_state_sleep_Type { OIC_STATE_SLEEP_TYPE_AWAKE, OIC_STATE_SLEEP_TYPE_LIGHT, OIC_STATE_SLEEP_TYPE_DEEP };
static const struct sol_str_table oic_state_sleep_Type_tbl[] = {
    SOL_STR_TABLE_ITEM("awake", OIC_STATE_SLEEP_TYPE_AWAKE),
SOL_STR_TABLE_ITEM("light", OIC_STATE_SLEEP_TYPE_LIGHT),
SOL_STR_TABLE_ITEM("deep", OIC_STATE_SLEEP_TYPE_DEEP),
    { }
};
struct oic_state_sleep {
    double At; /* Total awake time in hours per day */
double Dst; /* Total deep sleep time in hours per day */
double Lst; /* Total light sleep time in hours per day */
double Tst; /* Total sleep time in hours per day */
enum oic_state_sleep_Type Type; /* Sleep type categorized into awake, light, and deep */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_sleep_from_repr_vec(struct oic_state_sleep *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_sleep fields = {
        .At = state->At,
        .Dst = state->Dst,
        .Lst = state->Lst,
        .Tst = state->Tst,
        .Type = state->Type,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "At")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.At = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.At = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "Dst")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.Dst = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.Dst = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "Lst")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.Lst = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.Lst = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "Tst")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.Tst = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.Tst = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<3);
            continue;
        }
        if (decode_mask & (1<<4) && streq(field.key, "Type")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_sleep_Type_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.Type = (enum oic_state_sleep_Type)val;
            decode_mask &= ~(1<<4);
            continue;
        }
        if (decode_mask & (1<<5) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<5);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->At, fields.At)) {
        state->At = fields.At;
        updated = true;
    }

    if (check_updated_number(state->Dst, fields.Dst)) {
        state->Dst = fields.Dst;
        updated = true;
    }

    if (check_updated_number(state->Lst, fields.Lst)) {
        state->Lst = fields.Lst;
        updated = true;
    }

    if (check_updated_number(state->Tst, fields.Tst)) {
        state->Tst = fields.Tst;
        updated = true;
    }

    if (check_updated_int32(state->Type, fields.Type)) {
        state->Type = fields.Type;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}
/* Unit of a soft lean mass of a person */
enum oic_state_slm_units { OIC_STATE_SLM_UNITS_KG, OIC_STATE_SLM_UNITS_LB };
static const struct sol_str_table oic_state_slm_units_tbl[] = {
    SOL_STR_TABLE_ITEM("kg", OIC_STATE_SLM_UNITS_KG),
SOL_STR_TABLE_ITEM("lb", OIC_STATE_SLM_UNITS_LB),
    { }
};
struct oic_state_slm {
    double slm; /* The sum of the body water mass and protein mass */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
enum oic_state_slm_units units; /* Unit of a soft lean mass of a person */
};
static int
oic_state_slm_from_repr_vec(struct oic_state_slm *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_slm fields = {
        .slm = state->slm,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
        .units = state->units,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "slm")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.slm = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.slm = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_slm_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_slm_units)val;
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->slm, fields.slm)) {
        state->slm = fields.slm;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_speech_tts {
    char * supportedLanguages; /* comma separated list of supported language tags */
char * supportedVoices; /* SSML document fragment indicating supported voices */
char * utterance; /* SSML document including the speak body */
};
static int
oic_state_speech_tts_from_repr_vec(struct oic_state_speech_tts *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_speech_tts fields = {
        .supportedLanguages = state->supportedLanguages ? strdup(state->supportedLanguages) : NULL,
        .supportedVoices = state->supportedVoices ? strdup(state->supportedVoices) : NULL,
        .utterance = state->utterance ? strdup(state->utterance) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "supportedLanguages")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.supportedLanguages, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "supportedVoices")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.supportedVoices, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "utterance")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.utterance, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_string(state->supportedLanguages, fields.supportedLanguages)) {
        free(state->supportedLanguages);
        state->supportedLanguages = fields.supportedLanguages;
        fields.supportedLanguages = NULL;
        updated = true;
    }

    if (check_updated_string(state->supportedVoices, fields.supportedVoices)) {
        free(state->supportedVoices);
        state->supportedVoices = fields.supportedVoices;
        fields.supportedVoices = NULL;
        updated = true;
    }

    if (check_updated_string(state->utterance, fields.utterance)) {
        free(state->utterance);
        state->utterance = fields.utterance;
        fields.utterance = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.supportedLanguages);
    free(fields.supportedVoices);
    free(fields.utterance);
    return ret;
}
/* Unit of current speed */
enum oic_state_speed_units { OIC_STATE_SPEED_UNITS_MPH, OIC_STATE_SPEED_UNITS_KM_H, OIC_STATE_SPEED_UNITS_MACH, OIC_STATE_SPEED_UNITS_KNOT, OIC_STATE_SPEED_UNITS_M_MIN, OIC_STATE_SPEED_UNITS_M_S, OIC_STATE_SPEED_UNITS_FT_S, OIC_STATE_SPEED_UNITS_CM_S, OIC_STATE_SPEED_UNITS_YARD_S };
static const struct sol_str_table oic_state_speed_units_tbl[] = {
    SOL_STR_TABLE_ITEM("mph", OIC_STATE_SPEED_UNITS_MPH),
SOL_STR_TABLE_ITEM("km_h", OIC_STATE_SPEED_UNITS_KM_H),
SOL_STR_TABLE_ITEM("Mach", OIC_STATE_SPEED_UNITS_MACH),
SOL_STR_TABLE_ITEM("knot", OIC_STATE_SPEED_UNITS_KNOT),
SOL_STR_TABLE_ITEM("m_min", OIC_STATE_SPEED_UNITS_M_MIN),
SOL_STR_TABLE_ITEM("m_s", OIC_STATE_SPEED_UNITS_M_S),
SOL_STR_TABLE_ITEM("ft_s", OIC_STATE_SPEED_UNITS_FT_S),
SOL_STR_TABLE_ITEM("cm_s", OIC_STATE_SPEED_UNITS_CM_S),
SOL_STR_TABLE_ITEM("yard_s", OIC_STATE_SPEED_UNITS_YARD_S),
    { }
};
struct oic_state_speed {
    int32_t speed; /* current speed */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
enum oic_state_speed_units units; /* Unit of current speed */
};
static int
oic_state_speed_from_repr_vec(struct oic_state_speed *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_speed fields = {
        .speed = state->speed,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
        .units = state->units,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "speed")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.speed = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.speed = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.speed = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "units")) {
            int val;

            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);

            val = sol_str_table_lookup_fallback(oic_state_speed_units_tbl,
                field.v_slice, -1);
            if (val < 0)
                RETURN_ERROR(-EINVAL);
            fields.units = (enum oic_state_speed_units)val;
            decode_mask &= ~(1<<2);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->speed, fields.speed)) {
        state->speed = fields.speed;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }

    if (check_updated_int32(state->units, fields.units)) {
        state->units = fields.units;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_steps {
    int32_t count; /* Current or Target count. */
int32_t distance; /* measured distance based on steps in meters */
double kcal; /* measured calorie based on steps */
char * observedtime; /* An ISO8601 format time indicating when the data is observed (ex: 2016-02-15T09:19Z) */
};
static int
oic_state_steps_from_repr_vec(struct oic_state_steps *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_steps fields = {
        .count = state->count,
        .distance = state->distance,
        .kcal = state->kcal,
        .observedtime = state->observedtime ? strdup(state->observedtime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "count")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.count = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.count = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.count = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "distance")) {
            if (field.type == SOL_OIC_REPR_TYPE_UINT)
                fields.distance = field.v_uint;
            else if (field.type == SOL_OIC_REPR_TYPE_INT)
                fields.distance = field.v_int;
            else if (field.type == SOL_OIC_REPR_TYPE_SIMPLE)
                fields.distance = field.v_simple;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }
        if (decode_mask & (1<<2) && streq(field.key, "kcal")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.kcal = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.kcal = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<2);
            continue;
        }
        if (decode_mask & (1<<3) && streq(field.key, "observedtime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.observedtime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<3);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_int32(state->count, fields.count)) {
        state->count = fields.count;
        updated = true;
    }

    if (check_updated_int32(state->distance, fields.distance)) {
        state->distance = fields.distance;
        updated = true;
    }

    if (check_updated_number(state->kcal, fields.kcal)) {
        state->kcal = fields.kcal;
        updated = true;
    }

    if (check_updated_string(state->observedtime, fields.observedtime)) {
        free(state->observedtime);
        state->observedtime = fields.observedtime;
        fields.observedtime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.observedtime);
    return ret;
}

struct oic_state_switch_binary {
    bool value; /* Status of the switch */
};
static int
oic_state_switch_binary_from_repr_vec(struct oic_state_switch_binary *state,
    const struct sol_oic_map_reader *repr_map, uint32_t decode_mask)
{
    /* oic_state_energy_overload is equivalent to oic_state_switch_binary */
    return oic_state_energy_overload_from_repr_vec((struct oic_state_energy_overload *)state, repr_map, decode_mask);
}

struct oic_state_time_period {
    char * startTime; /* Start time for the time period */
char * stopTime; /* Stop time for the time period */
};
static int
oic_state_time_period_from_repr_vec(struct oic_state_time_period *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_time_period fields = {
        .startTime = state->startTime ? strdup(state->startTime) : NULL,
        .stopTime = state->stopTime ? strdup(state->stopTime) : NULL,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "startTime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.startTime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }
        if (decode_mask & (1<<1) && streq(field.key, "stopTime")) {
            if (field.type != SOL_OIC_REPR_TYPE_TEXT_STRING)
                RETURN_ERROR(-EINVAL);
            if (sol_util_replace_str_from_slice_if_changed(&fields.stopTime, field.v_slice) < 0)
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<1);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_string(state->startTime, fields.startTime)) {
        free(state->startTime);
        state->startTime = fields.startTime;
        fields.startTime = NULL;
        updated = true;
    }

    if (check_updated_string(state->stopTime, fields.stopTime)) {
        free(state->stopTime);
        state->stopTime = fields.stopTime;
        fields.stopTime = NULL;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:
    free(fields.startTime);
    free(fields.stopTime);
    return ret;
}

struct oic_state_weight {
    double weight; /* ReadOnly. Weight of an object */
};
static int
oic_state_weight_from_repr_vec(struct oic_state_weight *state,
    const struct sol_oic_map_reader *repr_vec, uint32_t decode_mask)
{
    struct sol_oic_repr_field field;
    enum sol_oic_map_loop_reason end_reason;
    struct sol_oic_map_reader iterator;
    struct oic_state_weight fields = {
        .weight = state->weight,
    };
    bool updated = false;
    int ret = 0;

    SOL_OIC_MAP_LOOP(repr_vec, &field, &iterator, end_reason) {

        if (decode_mask & (1<<0) && streq(field.key, "weight")) {
            if (field.type == SOL_OIC_REPR_TYPE_DOUBLE)
                fields.weight = field.v_double;
            else if (field.type == SOL_OIC_REPR_TYPE_FLOAT)
                fields.weight = field.v_float;
            else
                RETURN_ERROR(-EINVAL);
            decode_mask &= ~(1<<0);
            continue;
        }

    }
    if (end_reason != SOL_OIC_MAP_LOOP_OK)
        goto out;

    if (check_updated_number(state->weight, fields.weight)) {
        state->weight = fields.weight;
        updated = true;
    }
    ret = updated ? 1 : 0;

out:

    return ret;
}
struct oic_client_light {
    struct client_resource base;
    struct oic_state_light state;
};

static bool
oic_client_light_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_light *state = (struct oic_client_light *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("state", state->state.state));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("power", state->state.power));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("name", state->state.name, state->state.name ? strlen(state->state.name) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_light_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_light *res = (struct oic_client_light *)resource;
    return oic_state_light_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_light_inform_flow(struct client_resource *resource)
{
    struct oic_client_light *state = (struct oic_client_light *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_LIGHT__OUT__STATE, state->state.state);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_LIGHT__OUT__POWER, state->state.power);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_LIGHT__OUT__NAME, state->state.name);
}

static int
oic_client_light_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_light_options *node_opts =
        (const struct sol_flow_node_type_oic_client_light_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_light_to_repr_vec,
        .from_repr_vec = oic_client_light_from_repr_vec,
        .inform_flow = oic_client_light_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_LIGHT__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_LIGHT__OUT__DEVICE_ID
    };
    struct oic_client_light *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "core.light", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.state = false;
resource->state.power = 0;
resource->state.name = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_light_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_light *resource = data;
    free(resource->state.name);
    client_resource_close(&resource->base);
}

static int
oic_client_light_set_state(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_light *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.state, (bool) var)) {
            resource->state.state = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_light_set_power(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_light *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.power, (int32_t) var)) {
            resource->state.power = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_light_set_name(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_light *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.name, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            client_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_client_activity {
    struct client_resource base;
    struct oic_state_activity state;
};

static bool
oic_client_activity_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_activity *state = (struct oic_client_activity *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("Acalory", state->state.Acalory));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("Atime", state->state.Atime));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("Type", oic_state_activity_Type_tbl[state->state.Type].key, oic_state_activity_Type_tbl[state->state.Type].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_activity_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_activity *res = (struct oic_client_activity *)resource;
    return oic_state_activity_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_activity_inform_flow(struct client_resource *resource)
{
    struct oic_client_activity *state = (struct oic_client_activity *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ACTIVITY__OUT__ACALORY, state->state.Acalory);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ACTIVITY__OUT__ATIME, state->state.Atime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ACTIVITY__OUT__TYPE, oic_state_activity_Type_tbl[state->state.Type].key);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ACTIVITY__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_activity_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_activity_options *node_opts =
        (const struct sol_flow_node_type_oic_client_activity_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_activity_to_repr_vec,
        .from_repr_vec = oic_client_activity_from_repr_vec,
        .inform_flow = oic_client_activity_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ACTIVITY__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ACTIVITY__OUT__DEVICE_ID
    };
    struct oic_client_activity *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.activity", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.Acalory = 0;
resource->state.Atime = 0;
resource->state.Type = (enum oic_state_activity_Type)0;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_activity_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_activity *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

static int
oic_client_activity_set_Acalory(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_activity *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.Acalory, (int32_t) var)) {
            resource->state.Acalory = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_activity_set_Atime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_activity *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.Atime, (int32_t) var)) {
            resource->state.Atime = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_activity_set_Type(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_activity *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_activity_Type_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.Type = (enum oic_state_activity_Type)val;
            client_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_client_altimeter {
    struct client_resource base;
    struct oic_state_altimeter state;
};

static bool
oic_client_altimeter_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_altimeter *state = (struct oic_client_altimeter *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("alt", state->state.alt));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_altimeter_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_altimeter *res = (struct oic_client_altimeter *)resource;
    return oic_state_altimeter_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_altimeter_inform_flow(struct client_resource *resource)
{
    struct oic_client_altimeter *state = (struct oic_client_altimeter *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ALTIMETER__OUT__ALT, state->state.alt);
}

static int
oic_client_altimeter_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_altimeter_options *node_opts =
        (const struct sol_flow_node_type_oic_client_altimeter_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_altimeter_to_repr_vec,
        .from_repr_vec = oic_client_altimeter_from_repr_vec,
        .inform_flow = oic_client_altimeter_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ALTIMETER__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ALTIMETER__OUT__DEVICE_ID
    };
    struct oic_client_altimeter *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.altimeter", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.alt = 0.0f;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_altimeter_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_altimeter *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_altimeter_set_alt(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_altimeter *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.alt, (double) var)) {
            resource->state.alt = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_audio {
    struct client_resource base;
    struct oic_state_audio state;
};

static bool
oic_client_audio_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_audio *state = (struct oic_client_audio *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("mute", state->state.mute));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("volume", state->state.volume));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_audio_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_audio *res = (struct oic_client_audio *)resource;
    return oic_state_audio_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_audio_inform_flow(struct client_resource *resource)
{
    struct oic_client_audio *state = (struct oic_client_audio *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUDIO__OUT__MUTE, state->state.mute);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUDIO__OUT__VOLUME, state->state.volume);
}

static int
oic_client_audio_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_audio_options *node_opts =
        (const struct sol_flow_node_type_oic_client_audio_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_audio_to_repr_vec,
        .from_repr_vec = oic_client_audio_from_repr_vec,
        .inform_flow = oic_client_audio_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUDIO__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUDIO__OUT__DEVICE_ID
    };
    struct oic_client_audio *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.audio", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.mute = false;
resource->state.volume = 0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_audio_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_audio *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_audio_set_mute(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_audio *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.mute, (bool) var)) {
            resource->state.mute = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_audio_set_volume(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_audio *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.volume, (int32_t) var)) {
            resource->state.volume = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_autofocus {
    struct client_resource base;
    struct oic_state_autofocus state;
};

static bool
oic_client_autofocus_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_autofocus *state = (struct oic_client_autofocus *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("autoFocus", state->state.autoFocus));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_autofocus_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_autofocus *res = (struct oic_client_autofocus *)resource;
    return oic_state_autofocus_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_autofocus_inform_flow(struct client_resource *resource)
{
    struct oic_client_autofocus *state = (struct oic_client_autofocus *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUTOFOCUS__OUT__AUTO_FOCUS, state->state.autoFocus);
}

static int
oic_client_autofocus_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_autofocus_options *node_opts =
        (const struct sol_flow_node_type_oic_client_autofocus_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_autofocus_to_repr_vec,
        .from_repr_vec = oic_client_autofocus_from_repr_vec,
        .inform_flow = oic_client_autofocus_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUTOFOCUS__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUTOFOCUS__OUT__DEVICE_ID
    };
    struct oic_client_autofocus *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.autofocus", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.autoFocus = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_autofocus_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_autofocus *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_autofocus_set_autoFocus(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_autofocus *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.autoFocus, (bool) var)) {
            resource->state.autoFocus = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_automaticdocumentfeeder {
    struct client_resource base;
    struct oic_state_automaticdocumentfeeder state;
};
static int
oic_client_automaticdocumentfeeder_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_automaticdocumentfeeder *res = (struct oic_client_automaticdocumentfeeder *)resource;
    return oic_state_automaticdocumentfeeder_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_automaticdocumentfeeder_inform_flow(struct client_resource *resource)
{
    struct oic_client_automaticdocumentfeeder *state = (struct oic_client_automaticdocumentfeeder *)resource;
    send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUTOMATICDOCUMENTFEEDER__OUT__ADF_STATES, state->state.adfStates);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUTOMATICDOCUMENTFEEDER__OUT__CURRENT_ADF_STATE, state->state.currentAdfState);
}

static int
oic_client_automaticdocumentfeeder_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_automaticdocumentfeeder_options *node_opts =
        (const struct sol_flow_node_type_oic_client_automaticdocumentfeeder_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_automaticdocumentfeeder_from_repr_vec,
        .inform_flow = oic_client_automaticdocumentfeeder_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUTOMATICDOCUMENTFEEDER__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_AUTOMATICDOCUMENTFEEDER__OUT__DEVICE_ID
    };
    struct oic_client_automaticdocumentfeeder *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.automaticdocumentfeeder", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.adfStates = NULL;
resource->state.currentAdfState = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_automaticdocumentfeeder_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_automaticdocumentfeeder *resource = data;
    free(resource->state.adfStates);
free(resource->state.currentAdfState);
    client_resource_close(&resource->base);
}

struct oic_client_blood_glucose {
    struct client_resource base;
    struct oic_state_blood_glucose state;
};
static int
oic_client_blood_glucose_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_blood_glucose *res = (struct oic_client_blood_glucose *)resource;
    return oic_state_blood_glucose_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_blood_glucose_inform_flow(struct client_resource *resource)
{
    struct oic_client_blood_glucose *state = (struct oic_client_blood_glucose *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_GLUCOSE__OUT__BLOODSUGAR, state->state.bloodsugar);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_GLUCOSE__OUT__OBSERVEDTIME, state->state.observedtime);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_GLUCOSE__OUT__SITE, state->state.site);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_GLUCOSE__OUT__UNITS, oic_state_blood_glucose_units_tbl[state->state.units].key);
}

static int
oic_client_blood_glucose_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_blood_glucose_options *node_opts =
        (const struct sol_flow_node_type_oic_client_blood_glucose_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_blood_glucose_from_repr_vec,
        .inform_flow = oic_client_blood_glucose_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_GLUCOSE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_GLUCOSE__OUT__DEVICE_ID
    };
    struct oic_client_blood_glucose *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.blood.glucose", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.bloodsugar = 0.0f;
resource->state.observedtime = NULL;
resource->state.site = NULL;
resource->state.units = (enum oic_state_blood_glucose_units)0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_blood_glucose_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_blood_glucose *resource = data;
    free(resource->state.observedtime);
free(resource->state.site);
    client_resource_close(&resource->base);
}

struct oic_client_blood_o2_saturation {
    struct client_resource base;
    struct oic_state_blood_o2_saturation state;
};
static int
oic_client_blood_o2_saturation_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_blood_o2_saturation *res = (struct oic_client_blood_o2_saturation *)resource;
    return oic_state_blood_o2_saturation_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_blood_o2_saturation_inform_flow(struct client_resource *resource)
{
    struct oic_client_blood_o2_saturation *state = (struct oic_client_blood_o2_saturation *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_O2_SATURATION__OUT__PERFUSION, state->state.perfusion);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_O2_SATURATION__OUT__SPO2, state->state.spo2);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_O2_SATURATION__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_blood_o2_saturation_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_blood_o2_saturation_options *node_opts =
        (const struct sol_flow_node_type_oic_client_blood_o2_saturation_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_blood_o2_saturation_from_repr_vec,
        .inform_flow = oic_client_blood_o2_saturation_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_O2_SATURATION__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_O2_SATURATION__OUT__DEVICE_ID
    };
    struct oic_client_blood_o2_saturation *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.blood.o2.saturation", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.perfusion = 0.0f;
resource->state.spo2 = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_blood_o2_saturation_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_blood_o2_saturation *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_blood_pressure {
    struct client_resource base;
    struct oic_state_blood_pressure state;
};
static int
oic_client_blood_pressure_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_blood_pressure *res = (struct oic_client_blood_pressure *)resource;
    return oic_state_blood_pressure_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_blood_pressure_inform_flow(struct client_resource *resource)
{
    struct oic_client_blood_pressure *state = (struct oic_client_blood_pressure *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_PRESSURE__OUT__MAP, state->state.map);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_PRESSURE__OUT__BLOODPRESSURE, state->state.bloodpressure);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_PRESSURE__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_blood_pressure_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_blood_pressure_options *node_opts =
        (const struct sol_flow_node_type_oic_client_blood_pressure_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_blood_pressure_from_repr_vec,
        .inform_flow = oic_client_blood_pressure_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_PRESSURE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BLOOD_PRESSURE__OUT__DEVICE_ID
    };
    struct oic_client_blood_pressure *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.blood.pressure", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.map = 0.0f;
resource->state.bloodpressure = NULL;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_blood_pressure_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_blood_pressure *resource = data;
    free(resource->state.bloodpressure);
free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_bmi {
    struct client_resource base;
    struct oic_state_bmi state;
};
static int
oic_client_bmi_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_bmi *res = (struct oic_client_bmi *)resource;
    return oic_state_bmi_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_bmi_inform_flow(struct client_resource *resource)
{
    struct oic_client_bmi *state = (struct oic_client_bmi *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BMI__OUT__BMI, state->state.bmi);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BMI__OUT__BMIP, state->state.bmip);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BMI__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_bmi_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_bmi_options *node_opts =
        (const struct sol_flow_node_type_oic_client_bmi_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_bmi_from_repr_vec,
        .inform_flow = oic_client_bmi_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BMI__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BMI__OUT__DEVICE_ID
    };
    struct oic_client_bmi *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.bmi", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.bmi = 0.0f;
resource->state.bmip = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_bmi_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_bmi *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_bodyfat {
    struct client_resource base;
    struct oic_state_bodyfat state;
};
static int
oic_client_bodyfat_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_bodyfat *res = (struct oic_client_bodyfat *)resource;
    return oic_state_bodyfat_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_bodyfat_inform_flow(struct client_resource *resource)
{
    struct oic_client_bodyfat *state = (struct oic_client_bodyfat *)resource;
    sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYFAT__OUT__UNITS, oic_state_bodyfat_units_tbl[state->state.units].key);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYFAT__OUT__BODYFAT, state->state.bodyfat);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYFAT__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_bodyfat_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_bodyfat_options *node_opts =
        (const struct sol_flow_node_type_oic_client_bodyfat_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_bodyfat_from_repr_vec,
        .inform_flow = oic_client_bodyfat_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYFAT__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYFAT__OUT__DEVICE_ID
    };
    struct oic_client_bodyfat *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.bodyfat", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.units = (enum oic_state_bodyfat_units)0;
resource->state.bodyfat = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_bodyfat_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_bodyfat *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_bodyheight {
    struct client_resource base;
    struct oic_state_bodyheight state;
};

static bool
oic_client_bodyheight_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_bodyheight *state = (struct oic_client_bodyheight *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_bodyheight_units_tbl[state->state.units].key, oic_state_bodyheight_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("height", state->state.height));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_bodyheight_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_bodyheight *res = (struct oic_client_bodyheight *)resource;
    return oic_state_bodyheight_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_bodyheight_inform_flow(struct client_resource *resource)
{
    struct oic_client_bodyheight *state = (struct oic_client_bodyheight *)resource;
    sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYHEIGHT__OUT__UNITS, oic_state_bodyheight_units_tbl[state->state.units].key);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYHEIGHT__OUT__HEIGHT, state->state.height);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYHEIGHT__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_bodyheight_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_bodyheight_options *node_opts =
        (const struct sol_flow_node_type_oic_client_bodyheight_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_bodyheight_to_repr_vec,
        .from_repr_vec = oic_client_bodyheight_from_repr_vec,
        .inform_flow = oic_client_bodyheight_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYHEIGHT__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYHEIGHT__OUT__DEVICE_ID
    };
    struct oic_client_bodyheight *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.bodyheight", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.units = (enum oic_state_bodyheight_units)0;
resource->state.height = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_bodyheight_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_bodyheight *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

static int
oic_client_bodyheight_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_bodyheight *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_bodyheight_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_bodyheight_units)val;
            client_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}

static int
oic_client_bodyheight_set_height(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_bodyheight *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.height, (double) var)) {
            resource->state.height = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_body_position {
    struct client_resource base;
    struct oic_state_body_position state;
};
static int
oic_client_body_position_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_body_position *res = (struct oic_client_body_position *)resource;
    return oic_state_body_position_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_body_position_inform_flow(struct client_resource *resource)
{
    struct oic_client_body_position *state = (struct oic_client_body_position *)resource;
    sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODY_POSITION__OUT__BPOS, oic_state_body_position_bpos_tbl[state->state.bpos].key);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODY_POSITION__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_body_position_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_body_position_options *node_opts =
        (const struct sol_flow_node_type_oic_client_body_position_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_body_position_from_repr_vec,
        .inform_flow = oic_client_body_position_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODY_POSITION__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODY_POSITION__OUT__DEVICE_ID
    };
    struct oic_client_body_position *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.body.position", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.bpos = (enum oic_state_body_position_bpos)0;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_body_position_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_body_position *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_bodysite {
    struct client_resource base;
    struct oic_state_bodysite state;
};
static int
oic_client_bodysite_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_bodysite *res = (struct oic_client_bodysite *)resource;
    return oic_state_bodysite_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_bodysite_inform_flow(struct client_resource *resource)
{
    struct oic_client_bodysite *state = (struct oic_client_bodysite *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYSITE__OUT__BODYSITE, state->state.bodysite);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYSITE__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_bodysite_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_bodysite_options *node_opts =
        (const struct sol_flow_node_type_oic_client_bodysite_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_bodysite_from_repr_vec,
        .inform_flow = oic_client_bodysite_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYSITE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYSITE__OUT__DEVICE_ID
    };
    struct oic_client_bodysite *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.bodysite", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.bodysite = 0;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_bodysite_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_bodysite *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_bodywater {
    struct client_resource base;
    struct oic_state_bodywater state;
};
static int
oic_client_bodywater_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_bodywater *res = (struct oic_client_bodywater *)resource;
    return oic_state_bodywater_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_bodywater_inform_flow(struct client_resource *resource)
{
    struct oic_client_bodywater *state = (struct oic_client_bodywater *)resource;
    sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWATER__OUT__UNITS, oic_state_bodywater_units_tbl[state->state.units].key);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWATER__OUT__BODYWATER, state->state.bodywater);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWATER__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_bodywater_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_bodywater_options *node_opts =
        (const struct sol_flow_node_type_oic_client_bodywater_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_bodywater_from_repr_vec,
        .inform_flow = oic_client_bodywater_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWATER__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWATER__OUT__DEVICE_ID
    };
    struct oic_client_bodywater *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.bodywater", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.units = (enum oic_state_bodywater_units)0;
resource->state.bodywater = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_bodywater_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_bodywater *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_bodyweight {
    struct client_resource base;
    struct oic_state_bodyweight state;
};

static bool
oic_client_bodyweight_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_bodyweight *state = (struct oic_client_bodyweight *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("weight", state->state.weight));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_bodyweight_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_bodyweight *res = (struct oic_client_bodyweight *)resource;
    return oic_state_bodyweight_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_bodyweight_inform_flow(struct client_resource *resource)
{
    struct oic_client_bodyweight *state = (struct oic_client_bodyweight *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWEIGHT__OUT__WEIGHT, state->state.weight);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWEIGHT__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWEIGHT__OUT__UNITS, oic_state_bodyweight_units_tbl[state->state.units].key);
}

static int
oic_client_bodyweight_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_bodyweight_options *node_opts =
        (const struct sol_flow_node_type_oic_client_bodyweight_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_bodyweight_to_repr_vec,
        .from_repr_vec = oic_client_bodyweight_from_repr_vec,
        .inform_flow = oic_client_bodyweight_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWEIGHT__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_BODYWEIGHT__OUT__DEVICE_ID
    };
    struct oic_client_bodyweight *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.bodyweight", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.weight = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_bodyweight_units)0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_bodyweight_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_bodyweight *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

static int
oic_client_bodyweight_set_weight(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_bodyweight *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.weight, (double) var)) {
            resource->state.weight = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_clock {
    struct client_resource base;
    struct oic_state_clock state;
};

static bool
oic_client_clock_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_clock *state = (struct oic_client_clock *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("countdown", state->state.countdown));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("datetime", state->state.datetime, state->state.datetime ? strlen(state->state.datetime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_clock_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_clock *res = (struct oic_client_clock *)resource;
    return oic_state_clock_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_clock_inform_flow(struct client_resource *resource)
{
    struct oic_client_clock *state = (struct oic_client_clock *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CLOCK__OUT__COUNTDOWN, state->state.countdown);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CLOCK__OUT__DATETIME, state->state.datetime);
}

static int
oic_client_clock_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_clock_options *node_opts =
        (const struct sol_flow_node_type_oic_client_clock_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_clock_to_repr_vec,
        .from_repr_vec = oic_client_clock_from_repr_vec,
        .inform_flow = oic_client_clock_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_CLOCK__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_CLOCK__OUT__DEVICE_ID
    };
    struct oic_client_clock *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.clock", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.countdown = 0.0f;
resource->state.datetime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_clock_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_clock *resource = data;
    free(resource->state.datetime);
    client_resource_close(&resource->base);
}

static int
oic_client_clock_set_countdown(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_clock *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.countdown, (double) var)) {
            resource->state.countdown = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_clock_set_datetime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_clock *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.datetime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            client_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_client_colour_autowhitebalance {
    struct client_resource base;
    struct oic_state_colour_autowhitebalance state;
};

static bool
oic_client_colour_autowhitebalance_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_colour_autowhitebalance *state = (struct oic_client_colour_autowhitebalance *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("autoWhiteBalance", state->state.autoWhiteBalance));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_colour_autowhitebalance_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_colour_autowhitebalance *res = (struct oic_client_colour_autowhitebalance *)resource;
    return oic_state_colour_autowhitebalance_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_colour_autowhitebalance_inform_flow(struct client_resource *resource)
{
    struct oic_client_colour_autowhitebalance *state = (struct oic_client_colour_autowhitebalance *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_COLOUR_AUTOWHITEBALANCE__OUT__AUTO_WHITE_BALANCE, state->state.autoWhiteBalance);
}

static int
oic_client_colour_autowhitebalance_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_colour_autowhitebalance_options *node_opts =
        (const struct sol_flow_node_type_oic_client_colour_autowhitebalance_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_colour_autowhitebalance_to_repr_vec,
        .from_repr_vec = oic_client_colour_autowhitebalance_from_repr_vec,
        .inform_flow = oic_client_colour_autowhitebalance_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_COLOUR_AUTOWHITEBALANCE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_COLOUR_AUTOWHITEBALANCE__OUT__DEVICE_ID
    };
    struct oic_client_colour_autowhitebalance *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.colour.autowhitebalance", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.autoWhiteBalance = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_colour_autowhitebalance_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_colour_autowhitebalance *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_colour_autowhitebalance_set_autoWhiteBalance(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_colour_autowhitebalance *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.autoWhiteBalance, (bool) var)) {
            resource->state.autoWhiteBalance = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_colour_saturation {
    struct client_resource base;
    struct oic_state_colour_saturation state;
};

static bool
oic_client_colour_saturation_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_colour_saturation *state = (struct oic_client_colour_saturation *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("colourSaturation", state->state.colourSaturation));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_colour_saturation_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_colour_saturation *res = (struct oic_client_colour_saturation *)resource;
    return oic_state_colour_saturation_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_colour_saturation_inform_flow(struct client_resource *resource)
{
    struct oic_client_colour_saturation *state = (struct oic_client_colour_saturation *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_COLOUR_SATURATION__OUT__COLOUR_SATURATION, state->state.colourSaturation);
}

static int
oic_client_colour_saturation_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_colour_saturation_options *node_opts =
        (const struct sol_flow_node_type_oic_client_colour_saturation_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_colour_saturation_to_repr_vec,
        .from_repr_vec = oic_client_colour_saturation_from_repr_vec,
        .inform_flow = oic_client_colour_saturation_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_COLOUR_SATURATION__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_COLOUR_SATURATION__OUT__DEVICE_ID
    };
    struct oic_client_colour_saturation *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.colour.saturation", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.colourSaturation = 0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_colour_saturation_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_colour_saturation *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_colour_saturation_set_colourSaturation(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_colour_saturation *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.colourSaturation, (int32_t) var)) {
            resource->state.colourSaturation = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_cycling_power {
    struct client_resource base;
    struct oic_state_cycling_power state;
};

static bool
oic_client_cycling_power_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_cycling_power *state = (struct oic_client_cycling_power *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("leftonly", state->state.leftonly));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("al_power", state->state.al_power));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("ar_power", state->state.ar_power));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("ftp", state->state.ftp));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("power", state->state.power));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_cycling_power_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_cycling_power *res = (struct oic_client_cycling_power *)resource;
    return oic_state_cycling_power_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_cycling_power_inform_flow(struct client_resource *resource)
{
    struct oic_client_cycling_power *state = (struct oic_client_cycling_power *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_POWER__OUT__LEFTONLY, state->state.leftonly);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_POWER__OUT__AL_POWER, state->state.al_power);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_POWER__OUT__AR_POWER, state->state.ar_power);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_POWER__OUT__FTP, state->state.ftp);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_POWER__OUT__POWER, state->state.power);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_POWER__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_cycling_power_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_cycling_power_options *node_opts =
        (const struct sol_flow_node_type_oic_client_cycling_power_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_cycling_power_to_repr_vec,
        .from_repr_vec = oic_client_cycling_power_from_repr_vec,
        .inform_flow = oic_client_cycling_power_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_POWER__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_POWER__OUT__DEVICE_ID
    };
    struct oic_client_cycling_power *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.cycling.power", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.leftonly = false;
resource->state.al_power = 0.0f;
resource->state.ar_power = 0.0f;
resource->state.ftp = 0.0f;
resource->state.power = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_cycling_power_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_cycling_power *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

static int
oic_client_cycling_power_set_leftonly(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_cycling_power *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.leftonly, (bool) var)) {
            resource->state.leftonly = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_cycling_power_set_al_power(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_cycling_power *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.al_power, (double) var)) {
            resource->state.al_power = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_cycling_power_set_ar_power(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_cycling_power *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.ar_power, (double) var)) {
            resource->state.ar_power = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_cycling_power_set_ftp(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_cycling_power *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.ftp, (double) var)) {
            resource->state.ftp = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_cycling_power_set_power(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_cycling_power *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.power, (double) var)) {
            resource->state.power = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_cycling_speedcadence {
    struct client_resource base;
    struct oic_state_cycling_speedcadence state;
};

static bool
oic_client_cycling_speedcadence_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_cycling_speedcadence *state = (struct oic_client_cycling_speedcadence *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("speed", state->state.speed));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_cycling_speedcadence_units_tbl[state->state.units].key, oic_state_cycling_speedcadence_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_cycling_speedcadence_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_cycling_speedcadence *res = (struct oic_client_cycling_speedcadence *)resource;
    return oic_state_cycling_speedcadence_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_cycling_speedcadence_inform_flow(struct client_resource *resource)
{
    struct oic_client_cycling_speedcadence *state = (struct oic_client_cycling_speedcadence *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_SPEEDCADENCE__OUT__SPEED, state->state.speed);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_SPEEDCADENCE__OUT__CADENCE, state->state.cadence);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_SPEEDCADENCE__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_SPEEDCADENCE__OUT__UNITS, oic_state_cycling_speedcadence_units_tbl[state->state.units].key);
}

static int
oic_client_cycling_speedcadence_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_cycling_speedcadence_options *node_opts =
        (const struct sol_flow_node_type_oic_client_cycling_speedcadence_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_cycling_speedcadence_to_repr_vec,
        .from_repr_vec = oic_client_cycling_speedcadence_from_repr_vec,
        .inform_flow = oic_client_cycling_speedcadence_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_SPEEDCADENCE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_CYCLING_SPEEDCADENCE__OUT__DEVICE_ID
    };
    struct oic_client_cycling_speedcadence *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.cycling.speedcadence", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.speed = 0;
resource->state.cadence = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_cycling_speedcadence_units)0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_cycling_speedcadence_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_cycling_speedcadence *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

static int
oic_client_cycling_speedcadence_set_speed(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_cycling_speedcadence *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.speed, (int32_t) var)) {
            resource->state.speed = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_cycling_speedcadence_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_cycling_speedcadence *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_cycling_speedcadence_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_cycling_speedcadence_units)val;
            client_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_client_distance {
    struct client_resource base;
    struct oic_state_distance state;
};

static bool
oic_client_distance_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_distance *state = (struct oic_client_distance *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("distance", state->state.distance));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_distance_units_tbl[state->state.units].key, oic_state_distance_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_distance_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_distance *res = (struct oic_client_distance *)resource;
    return oic_state_distance_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_distance_inform_flow(struct client_resource *resource)
{
    struct oic_client_distance *state = (struct oic_client_distance *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_DISTANCE__OUT__DISTANCE, state->state.distance);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_DISTANCE__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_DISTANCE__OUT__UNITS, oic_state_distance_units_tbl[state->state.units].key);
}

static int
oic_client_distance_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_distance_options *node_opts =
        (const struct sol_flow_node_type_oic_client_distance_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_distance_to_repr_vec,
        .from_repr_vec = oic_client_distance_from_repr_vec,
        .inform_flow = oic_client_distance_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_DISTANCE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_DISTANCE__OUT__DEVICE_ID
    };
    struct oic_client_distance *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.distance", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.distance = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_distance_units)0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_distance_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_distance *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

static int
oic_client_distance_set_distance(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_distance *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.distance, (double) var)) {
            resource->state.distance = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_distance_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_distance *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_distance_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_distance_units)val;
            client_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_client_door {
    struct client_resource base;
    struct oic_state_door state;
};

static bool
oic_client_door_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_door *state = (struct oic_client_door *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("openAlarm", state->state.openAlarm));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_door_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_door *res = (struct oic_client_door *)resource;
    return oic_state_door_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_door_inform_flow(struct client_resource *resource)
{
    struct oic_client_door *state = (struct oic_client_door *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_DOOR__OUT__OPEN_ALARM, state->state.openAlarm);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_DOOR__OUT__OPEN_STATE, oic_state_door_openState_tbl[state->state.openState].key);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_DOOR__OUT__OPEN_DURATION, state->state.openDuration);
}

static int
oic_client_door_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_door_options *node_opts =
        (const struct sol_flow_node_type_oic_client_door_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_door_to_repr_vec,
        .from_repr_vec = oic_client_door_from_repr_vec,
        .inform_flow = oic_client_door_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_DOOR__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_DOOR__OUT__DEVICE_ID
    };
    struct oic_client_door *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.door", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.openAlarm = false;
resource->state.openState = (enum oic_state_door_openState)0;
resource->state.openDuration = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_door_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_door *resource = data;
    free(resource->state.openDuration);
    client_resource_close(&resource->base);
}

static int
oic_client_door_set_openAlarm(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_door *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.openAlarm, (bool) var)) {
            resource->state.openAlarm = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_emg {
    struct client_resource base;
    struct oic_state_emg state;
};
static int
oic_client_emg_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_emg *res = (struct oic_client_emg *)resource;
    return oic_state_emg_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_emg_inform_flow(struct client_resource *resource)
{
    struct oic_client_emg *state = (struct oic_client_emg *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_EMG__OUT__MI, state->state.mi);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_EMG__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_emg_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_emg_options *node_opts =
        (const struct sol_flow_node_type_oic_client_emg_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_emg_from_repr_vec,
        .inform_flow = oic_client_emg_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_EMG__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_EMG__OUT__DEVICE_ID
    };
    struct oic_client_emg *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.emg", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.mi = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_emg_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_emg *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_energy_battery {
    struct client_resource base;
    struct oic_state_energy_battery state;
};
static int
oic_client_energy_battery_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_energy_battery *res = (struct oic_client_energy_battery *)resource;
    return oic_state_energy_battery_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_energy_battery_inform_flow(struct client_resource *resource)
{
    struct oic_client_energy_battery *state = (struct oic_client_energy_battery *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_BATTERY__OUT__CHARGE, state->state.charge);
}

static int
oic_client_energy_battery_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_energy_battery_options *node_opts =
        (const struct sol_flow_node_type_oic_client_energy_battery_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_energy_battery_from_repr_vec,
        .inform_flow = oic_client_energy_battery_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_BATTERY__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_BATTERY__OUT__DEVICE_ID
    };
    struct oic_client_energy_battery *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.energy.battery", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.charge = 0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_energy_battery_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_energy_battery *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_energy_consumption {
    struct client_resource base;
    struct oic_state_energy_consumption state;
};
static int
oic_client_energy_consumption_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_energy_consumption *res = (struct oic_client_energy_consumption *)resource;
    return oic_state_energy_consumption_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_energy_consumption_inform_flow(struct client_resource *resource)
{
    struct oic_client_energy_consumption *state = (struct oic_client_energy_consumption *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_CONSUMPTION__OUT__ENERGY, state->state.energy);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_CONSUMPTION__OUT__POWER, state->state.power);
}

static int
oic_client_energy_consumption_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_energy_consumption_options *node_opts =
        (const struct sol_flow_node_type_oic_client_energy_consumption_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_energy_consumption_from_repr_vec,
        .inform_flow = oic_client_energy_consumption_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_CONSUMPTION__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_CONSUMPTION__OUT__DEVICE_ID
    };
    struct oic_client_energy_consumption *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.energy.consumption", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.energy = 0.0f;
resource->state.power = 0.0f;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_energy_consumption_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_energy_consumption *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_energy_drlc {
    struct client_resource base;
    struct oic_state_energy_drlc state;
};

static bool
oic_client_energy_drlc_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_energy_drlc *state = (struct oic_client_energy_drlc *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("override", state->state.override));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("DRType", state->state.DRType));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("duration", state->state.duration));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("start", state->state.start, state->state.start ? strlen(state->state.start) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_energy_drlc_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_energy_drlc *res = (struct oic_client_energy_drlc *)resource;
    return oic_state_energy_drlc_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_energy_drlc_inform_flow(struct client_resource *resource)
{
    struct oic_client_energy_drlc *state = (struct oic_client_energy_drlc *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_DRLC__OUT__OVERRIDE, state->state.override);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_DRLC__OUT__D_RTYPE, state->state.DRType);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_DRLC__OUT__DURATION, state->state.duration);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_DRLC__OUT__START, state->state.start);
}

static int
oic_client_energy_drlc_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_energy_drlc_options *node_opts =
        (const struct sol_flow_node_type_oic_client_energy_drlc_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_energy_drlc_to_repr_vec,
        .from_repr_vec = oic_client_energy_drlc_from_repr_vec,
        .inform_flow = oic_client_energy_drlc_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_DRLC__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_DRLC__OUT__DEVICE_ID
    };
    struct oic_client_energy_drlc *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.energy.drlc", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.override = false;
resource->state.DRType = 0;
resource->state.duration = 0;
resource->state.start = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_energy_drlc_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_energy_drlc *resource = data;
    free(resource->state.start);
    client_resource_close(&resource->base);
}

static int
oic_client_energy_drlc_set_override(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_energy_drlc *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.override, (bool) var)) {
            resource->state.override = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_energy_drlc_set_DRType(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_energy_drlc *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.DRType, (int32_t) var)) {
            resource->state.DRType = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_energy_drlc_set_duration(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_energy_drlc *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.duration, (int32_t) var)) {
            resource->state.duration = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_energy_drlc_set_start(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_energy_drlc *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.start, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            client_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_client_energy_overload {
    struct client_resource base;
    struct oic_state_energy_overload state;
};
static int
oic_client_energy_overload_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_energy_overload *res = (struct oic_client_energy_overload *)resource;
    return oic_state_energy_overload_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_energy_overload_inform_flow(struct client_resource *resource)
{
    struct oic_client_energy_overload *state = (struct oic_client_energy_overload *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_OVERLOAD__OUT__VALUE, state->state.value);
}

static int
oic_client_energy_overload_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_energy_overload_options *node_opts =
        (const struct sol_flow_node_type_oic_client_energy_overload_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_energy_overload_from_repr_vec,
        .inform_flow = oic_client_energy_overload_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_OVERLOAD__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ENERGY_OVERLOAD__OUT__DEVICE_ID
    };
    struct oic_client_energy_overload *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.energy.overload", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_energy_overload_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_energy_overload *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_fev1 {
    struct client_resource base;
    struct oic_state_fev1 state;
};
static int
oic_client_fev1_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_fev1 *res = (struct oic_client_fev1 *)resource;
    return oic_state_fev1_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_fev1_inform_flow(struct client_resource *resource)
{
    struct oic_client_fev1 *state = (struct oic_client_fev1 *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_FEV1__OUT__FEV1, state->state.fev1);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_FEV1__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_fev1_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_fev1_options *node_opts =
        (const struct sol_flow_node_type_oic_client_fev1_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_fev1_from_repr_vec,
        .inform_flow = oic_client_fev1_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_FEV1__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_FEV1__OUT__DEVICE_ID
    };
    struct oic_client_fev1 *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.fev1", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.fev1 = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_fev1_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_fev1 *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_ffm {
    struct client_resource base;
    struct oic_state_ffm state;
};
static int
oic_client_ffm_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_ffm *res = (struct oic_client_ffm *)resource;
    return oic_state_ffm_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_ffm_inform_flow(struct client_resource *resource)
{
    struct oic_client_ffm *state = (struct oic_client_ffm *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_FFM__OUT__FFM, state->state.ffm);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_FFM__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_FFM__OUT__UNITS, oic_state_ffm_units_tbl[state->state.units].key);
}

static int
oic_client_ffm_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_ffm_options *node_opts =
        (const struct sol_flow_node_type_oic_client_ffm_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_ffm_from_repr_vec,
        .inform_flow = oic_client_ffm_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_FFM__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_FFM__OUT__DEVICE_ID
    };
    struct oic_client_ffm *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.ffm", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.ffm = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_ffm_units)0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_ffm_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_ffm *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_gsr {
    struct client_resource base;
    struct oic_state_gsr state;
};
static int
oic_client_gsr_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_gsr *res = (struct oic_client_gsr *)resource;
    return oic_state_gsr_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_gsr_inform_flow(struct client_resource *resource)
{
    struct oic_client_gsr *state = (struct oic_client_gsr *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_GSR__OUT__SKINCON, state->state.skincon);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_GSR__OUT__SKINCV, state->state.skincv);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_GSR__OUT__SKINRES, state->state.skinres);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_GSR__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_gsr_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_gsr_options *node_opts =
        (const struct sol_flow_node_type_oic_client_gsr_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_gsr_from_repr_vec,
        .inform_flow = oic_client_gsr_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_GSR__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_GSR__OUT__DEVICE_ID
    };
    struct oic_client_gsr *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.gsr", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.skincon = 0.0f;
resource->state.skincv = 0.0f;
resource->state.skinres = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_gsr_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_gsr *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_heartrate {
    struct client_resource base;
    struct oic_state_heartrate state;
};
static int
oic_client_heartrate_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_heartrate *res = (struct oic_client_heartrate *)resource;
    return oic_state_heartrate_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_heartrate_inform_flow(struct client_resource *resource)
{
    struct oic_client_heartrate *state = (struct oic_client_heartrate *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_HEARTRATE__OUT__HEARTRATE, state->state.heartrate);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_HEARTRATE__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_heartrate_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_heartrate_options *node_opts =
        (const struct sol_flow_node_type_oic_client_heartrate_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_heartrate_from_repr_vec,
        .inform_flow = oic_client_heartrate_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_HEARTRATE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_HEARTRATE__OUT__DEVICE_ID
    };
    struct oic_client_heartrate *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.heartrate", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.heartrate = 0;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_heartrate_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_heartrate *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_height {
    struct client_resource base;
    struct oic_state_height state;
};

static bool
oic_client_height_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_height *state = (struct oic_client_height *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("height", state->state.height));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_height_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_height *res = (struct oic_client_height *)resource;
    return oic_state_height_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_height_inform_flow(struct client_resource *resource)
{
    struct oic_client_height *state = (struct oic_client_height *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_HEIGHT__OUT__HEIGHT, state->state.height);
}

static int
oic_client_height_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_height_options *node_opts =
        (const struct sol_flow_node_type_oic_client_height_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_height_to_repr_vec,
        .from_repr_vec = oic_client_height_from_repr_vec,
        .inform_flow = oic_client_height_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_HEIGHT__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_HEIGHT__OUT__DEVICE_ID
    };
    struct oic_client_height *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.height", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.height = 0.0f;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_height_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_height *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_height_set_height(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_height *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.height, (double) var)) {
            resource->state.height = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_humidity {
    struct client_resource base;
    struct oic_state_humidity state;
};

static bool
oic_client_humidity_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_humidity *state = (struct oic_client_humidity *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("desiredHumidity", state->state.desiredHumidity));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_humidity_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_humidity *res = (struct oic_client_humidity *)resource;
    return oic_state_humidity_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_humidity_inform_flow(struct client_resource *resource)
{
    struct oic_client_humidity *state = (struct oic_client_humidity *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_HUMIDITY__OUT__DESIRED_HUMIDITY, state->state.desiredHumidity);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_HUMIDITY__OUT__HUMIDITY, state->state.humidity);
}

static int
oic_client_humidity_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_humidity_options *node_opts =
        (const struct sol_flow_node_type_oic_client_humidity_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_humidity_to_repr_vec,
        .from_repr_vec = oic_client_humidity_from_repr_vec,
        .inform_flow = oic_client_humidity_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_HUMIDITY__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_HUMIDITY__OUT__DEVICE_ID
    };
    struct oic_client_humidity *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.humidity", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.desiredHumidity = 0;
resource->state.humidity = 0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_humidity_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_humidity *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_humidity_set_desiredHumidity(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_humidity *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.desiredHumidity, (int32_t) var)) {
            resource->state.desiredHumidity = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_icemaker {
    struct client_resource base;
    struct oic_state_icemaker state;
};

static bool
oic_client_icemaker_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_icemaker *state = (struct oic_client_icemaker *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("status", oic_state_icemaker_status_tbl[state->state.status].key, oic_state_icemaker_status_tbl[state->state.status].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_icemaker_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_icemaker *res = (struct oic_client_icemaker *)resource;
    return oic_state_icemaker_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_icemaker_inform_flow(struct client_resource *resource)
{
    struct oic_client_icemaker *state = (struct oic_client_icemaker *)resource;
    sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_ICEMAKER__OUT__STATUS, oic_state_icemaker_status_tbl[state->state.status].key);
}

static int
oic_client_icemaker_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_icemaker_options *node_opts =
        (const struct sol_flow_node_type_oic_client_icemaker_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_icemaker_to_repr_vec,
        .from_repr_vec = oic_client_icemaker_from_repr_vec,
        .inform_flow = oic_client_icemaker_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ICEMAKER__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_ICEMAKER__OUT__DEVICE_ID
    };
    struct oic_client_icemaker *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.iceMaker", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.status = (enum oic_state_icemaker_status)0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_icemaker_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_icemaker *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_icemaker_set_status(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_icemaker *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_icemaker_status_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.status = (enum oic_state_icemaker_status)val;
            client_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_client_light_brightness {
    struct client_resource base;
    struct oic_state_light_brightness state;
};

static bool
oic_client_light_brightness_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_light_brightness *state = (struct oic_client_light_brightness *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("brightness", state->state.brightness));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_light_brightness_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_light_brightness *res = (struct oic_client_light_brightness *)resource;
    return oic_state_light_brightness_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_light_brightness_inform_flow(struct client_resource *resource)
{
    struct oic_client_light_brightness *state = (struct oic_client_light_brightness *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_LIGHT_BRIGHTNESS__OUT__BRIGHTNESS, state->state.brightness);
}

static int
oic_client_light_brightness_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_light_brightness_options *node_opts =
        (const struct sol_flow_node_type_oic_client_light_brightness_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_light_brightness_to_repr_vec,
        .from_repr_vec = oic_client_light_brightness_from_repr_vec,
        .inform_flow = oic_client_light_brightness_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_LIGHT_BRIGHTNESS__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_LIGHT_BRIGHTNESS__OUT__DEVICE_ID
    };
    struct oic_client_light_brightness *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.light.brightness", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.brightness = 0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_light_brightness_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_light_brightness *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_light_brightness_set_brightness(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_light_brightness *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.brightness, (int32_t) var)) {
            resource->state.brightness = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_lock_status {
    struct client_resource base;
    struct oic_state_lock_status state;
};

static bool
oic_client_lock_status_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_lock_status *state = (struct oic_client_lock_status *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("lockState", oic_state_lock_status_lockState_tbl[state->state.lockState].key, oic_state_lock_status_lockState_tbl[state->state.lockState].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_lock_status_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_lock_status *res = (struct oic_client_lock_status *)resource;
    return oic_state_lock_status_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_lock_status_inform_flow(struct client_resource *resource)
{
    struct oic_client_lock_status *state = (struct oic_client_lock_status *)resource;
    sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_LOCK_STATUS__OUT__LOCK_STATE, oic_state_lock_status_lockState_tbl[state->state.lockState].key);
}

static int
oic_client_lock_status_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_lock_status_options *node_opts =
        (const struct sol_flow_node_type_oic_client_lock_status_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_lock_status_to_repr_vec,
        .from_repr_vec = oic_client_lock_status_from_repr_vec,
        .inform_flow = oic_client_lock_status_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_LOCK_STATUS__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_LOCK_STATUS__OUT__DEVICE_ID
    };
    struct oic_client_lock_status *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.lock.status", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.lockState = (enum oic_state_lock_status_lockState)0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_lock_status_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_lock_status *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_lock_status_set_lockState(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_lock_status *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_lock_status_lockState_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.lockState = (enum oic_state_lock_status_lockState)val;
            client_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_client_mode {
    struct client_resource base;
    struct oic_state_mode state;
};

static bool
oic_client_mode_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_mode *state = (struct oic_client_mode *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("modes", state->state.modes, state->state.modes ? strlen(state->state.modes) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_mode_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_mode *res = (struct oic_client_mode *)resource;
    return oic_state_mode_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_mode_inform_flow(struct client_resource *resource)
{
    struct oic_client_mode *state = (struct oic_client_mode *)resource;
    send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_MODE__OUT__MODES, state->state.modes);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_MODE__OUT__SUPPORTED_MODES, state->state.supportedModes);
}

static int
oic_client_mode_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_mode_options *node_opts =
        (const struct sol_flow_node_type_oic_client_mode_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_mode_to_repr_vec,
        .from_repr_vec = oic_client_mode_from_repr_vec,
        .inform_flow = oic_client_mode_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_MODE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_MODE__OUT__DEVICE_ID
    };
    struct oic_client_mode *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.mode", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.modes = NULL;
resource->state.supportedModes = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_mode_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_mode *resource = data;
    free(resource->state.modes);
free(resource->state.supportedModes);
    client_resource_close(&resource->base);
}

static int
oic_client_mode_set_modes(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_mode *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.modes, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            client_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_client_muscle_o2_saturation {
    struct client_resource base;
    struct oic_state_muscle_o2_saturation state;
};
static int
oic_client_muscle_o2_saturation_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_muscle_o2_saturation *res = (struct oic_client_muscle_o2_saturation *)resource;
    return oic_state_muscle_o2_saturation_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_muscle_o2_saturation_inform_flow(struct client_resource *resource)
{
    struct oic_client_muscle_o2_saturation *state = (struct oic_client_muscle_o2_saturation *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_MUSCLE_O2_SATURATION__OUT__SMO2, state->state.smo2);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_MUSCLE_O2_SATURATION__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_muscle_o2_saturation_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_muscle_o2_saturation_options *node_opts =
        (const struct sol_flow_node_type_oic_client_muscle_o2_saturation_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_muscle_o2_saturation_from_repr_vec,
        .inform_flow = oic_client_muscle_o2_saturation_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_MUSCLE_O2_SATURATION__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_MUSCLE_O2_SATURATION__OUT__DEVICE_ID
    };
    struct oic_client_muscle_o2_saturation *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.muscle.o2.saturation", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.smo2 = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_muscle_o2_saturation_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_muscle_o2_saturation *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_nightmode {
    struct client_resource base;
    struct oic_state_nightmode state;
};

static bool
oic_client_nightmode_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_nightmode *state = (struct oic_client_nightmode *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("nightMode", state->state.nightMode));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_nightmode_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_nightmode *res = (struct oic_client_nightmode *)resource;
    return oic_state_nightmode_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_nightmode_inform_flow(struct client_resource *resource)
{
    struct oic_client_nightmode *state = (struct oic_client_nightmode *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_NIGHTMODE__OUT__NIGHT_MODE, state->state.nightMode);
}

static int
oic_client_nightmode_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_nightmode_options *node_opts =
        (const struct sol_flow_node_type_oic_client_nightmode_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_nightmode_to_repr_vec,
        .from_repr_vec = oic_client_nightmode_from_repr_vec,
        .inform_flow = oic_client_nightmode_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_NIGHTMODE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_NIGHTMODE__OUT__DEVICE_ID
    };
    struct oic_client_nightmode *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.nightmode", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.nightMode = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_nightmode_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_nightmode *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_nightmode_set_nightMode(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_nightmode *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.nightMode, (bool) var)) {
            resource->state.nightMode = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_operational_state {
    struct client_resource base;
    struct oic_state_operational_state state;
};

static bool
oic_client_operational_state_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_operational_state *state = (struct oic_client_operational_state *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("currentJobState", state->state.currentJobState, state->state.currentJobState ? strlen(state->state.currentJobState) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("currentMachineState", state->state.currentMachineState, state->state.currentMachineState ? strlen(state->state.currentMachineState) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_operational_state_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_operational_state *res = (struct oic_client_operational_state *)resource;
    return oic_state_operational_state_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_operational_state_inform_flow(struct client_resource *resource)
{
    struct oic_client_operational_state *state = (struct oic_client_operational_state *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_OPERATIONAL_STATE__OUT__PROGRESS_PERCENTAGE, state->state.progressPercentage);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_OPERATIONAL_STATE__OUT__CURRENT_JOB_STATE, state->state.currentJobState);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_OPERATIONAL_STATE__OUT__CURRENT_MACHINE_STATE, state->state.currentMachineState);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_OPERATIONAL_STATE__OUT__JOB_STATES, state->state.jobStates);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_OPERATIONAL_STATE__OUT__MACHINE_STATES, state->state.machineStates);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_OPERATIONAL_STATE__OUT__REMAINING_TIME, state->state.remainingTime);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_OPERATIONAL_STATE__OUT__RUNNING_TIME, state->state.runningTime);
}

static int
oic_client_operational_state_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_operational_state_options *node_opts =
        (const struct sol_flow_node_type_oic_client_operational_state_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_operational_state_to_repr_vec,
        .from_repr_vec = oic_client_operational_state_from_repr_vec,
        .inform_flow = oic_client_operational_state_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_OPERATIONAL_STATE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_OPERATIONAL_STATE__OUT__DEVICE_ID
    };
    struct oic_client_operational_state *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.operational.state", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.progressPercentage = 0;
resource->state.currentJobState = NULL;
resource->state.currentMachineState = NULL;
resource->state.jobStates = NULL;
resource->state.machineStates = NULL;
resource->state.remainingTime = NULL;
resource->state.runningTime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_operational_state_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_operational_state *resource = data;
    free(resource->state.currentJobState);
free(resource->state.currentMachineState);
free(resource->state.jobStates);
free(resource->state.machineStates);
free(resource->state.remainingTime);
free(resource->state.runningTime);
    client_resource_close(&resource->base);
}

static int
oic_client_operational_state_set_currentJobState(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_operational_state *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.currentJobState, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            client_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_client_operational_state_set_currentMachineState(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_operational_state *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.currentMachineState, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            client_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_client_pef {
    struct client_resource base;
    struct oic_state_pef state;
};
static int
oic_client_pef_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_pef *res = (struct oic_client_pef *)resource;
    return oic_state_pef_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_pef_inform_flow(struct client_resource *resource)
{
    struct oic_client_pef *state = (struct oic_client_pef *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_PEF__OUT__PEF, state->state.pef);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_PEF__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_pef_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_pef_options *node_opts =
        (const struct sol_flow_node_type_oic_client_pef_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_pef_from_repr_vec,
        .inform_flow = oic_client_pef_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_PEF__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_PEF__OUT__DEVICE_ID
    };
    struct oic_client_pef *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.pef", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.pef = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_pef_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_pef *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_refrigeration {
    struct client_resource base;
    struct oic_state_refrigeration state;
};

static bool
oic_client_refrigeration_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_refrigeration *state = (struct oic_client_refrigeration *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("defrost", state->state.defrost));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("rapidCool", state->state.rapidCool));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("rapidFreeze", state->state.rapidFreeze));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_refrigeration_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_refrigeration *res = (struct oic_client_refrigeration *)resource;
    return oic_state_refrigeration_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_refrigeration_inform_flow(struct client_resource *resource)
{
    struct oic_client_refrigeration *state = (struct oic_client_refrigeration *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_REFRIGERATION__OUT__DEFROST, state->state.defrost);
sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_REFRIGERATION__OUT__RAPID_COOL, state->state.rapidCool);
sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_REFRIGERATION__OUT__RAPID_FREEZE, state->state.rapidFreeze);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_REFRIGERATION__OUT__FILTER, state->state.filter);
}

static int
oic_client_refrigeration_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_refrigeration_options *node_opts =
        (const struct sol_flow_node_type_oic_client_refrigeration_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_refrigeration_to_repr_vec,
        .from_repr_vec = oic_client_refrigeration_from_repr_vec,
        .inform_flow = oic_client_refrigeration_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_REFRIGERATION__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_REFRIGERATION__OUT__DEVICE_ID
    };
    struct oic_client_refrigeration *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.refrigeration", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.defrost = false;
resource->state.rapidCool = false;
resource->state.rapidFreeze = false;
resource->state.filter = 0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_refrigeration_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_refrigeration *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_refrigeration_set_defrost(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_refrigeration *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.defrost, (bool) var)) {
            resource->state.defrost = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_refrigeration_set_rapidCool(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_refrigeration *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.rapidCool, (bool) var)) {
            resource->state.rapidCool = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_refrigeration_set_rapidFreeze(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_refrigeration *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.rapidFreeze, (bool) var)) {
            resource->state.rapidFreeze = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_repetition {
    struct client_resource base;
    struct oic_state_repetition state;
};
static int
oic_client_repetition_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_repetition *res = (struct oic_client_repetition *)resource;
    return oic_state_repetition_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_repetition_inform_flow(struct client_resource *resource)
{
    struct oic_client_repetition *state = (struct oic_client_repetition *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_REPETITION__OUT__REPETITION, state->state.repetition);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_REPETITION__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_repetition_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_repetition_options *node_opts =
        (const struct sol_flow_node_type_oic_client_repetition_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_repetition_from_repr_vec,
        .inform_flow = oic_client_repetition_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_REPETITION__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_REPETITION__OUT__DEVICE_ID
    };
    struct oic_client_repetition *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.repetition", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.repetition = 0;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_repetition_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_repetition *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_respirationrate {
    struct client_resource base;
    struct oic_state_respirationrate state;
};
static int
oic_client_respirationrate_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_respirationrate *res = (struct oic_client_respirationrate *)resource;
    return oic_state_respirationrate_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_respirationrate_inform_flow(struct client_resource *resource)
{
    struct oic_client_respirationrate *state = (struct oic_client_respirationrate *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_RESPIRATIONRATE__OUT__RESPIRATE, state->state.respirate);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_RESPIRATIONRATE__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_respirationrate_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_respirationrate_options *node_opts =
        (const struct sol_flow_node_type_oic_client_respirationrate_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_respirationrate_from_repr_vec,
        .inform_flow = oic_client_respirationrate_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_RESPIRATIONRATE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_RESPIRATIONRATE__OUT__DEVICE_ID
    };
    struct oic_client_respirationrate *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.respirationrate", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.respirate = 0;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_respirationrate_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_respirationrate *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_sensor_carbondioxide {
    struct client_resource base;
    struct oic_state_sensor_carbondioxide state;
};
static int
oic_client_sensor_carbondioxide_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_carbondioxide *res = (struct oic_client_sensor_carbondioxide *)resource;
    return oic_state_sensor_carbondioxide_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_carbondioxide_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_carbondioxide *state = (struct oic_client_sensor_carbondioxide *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_CARBONDIOXIDE__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_carbondioxide_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_carbondioxide_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_carbondioxide_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_carbondioxide_from_repr_vec,
        .inform_flow = oic_client_sensor_carbondioxide_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_CARBONDIOXIDE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_CARBONDIOXIDE__OUT__DEVICE_ID
    };
    struct oic_client_sensor_carbondioxide *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.carbondioxide", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_carbondioxide_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_carbondioxide *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_carbonmonoxide {
    struct client_resource base;
    struct oic_state_sensor_carbonmonoxide state;
};
static int
oic_client_sensor_carbonmonoxide_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_carbonmonoxide *res = (struct oic_client_sensor_carbonmonoxide *)resource;
    return oic_state_sensor_carbonmonoxide_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_carbonmonoxide_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_carbonmonoxide *state = (struct oic_client_sensor_carbonmonoxide *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_CARBONMONOXIDE__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_carbonmonoxide_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_carbonmonoxide_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_carbonmonoxide_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_carbonmonoxide_from_repr_vec,
        .inform_flow = oic_client_sensor_carbonmonoxide_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_CARBONMONOXIDE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_CARBONMONOXIDE__OUT__DEVICE_ID
    };
    struct oic_client_sensor_carbonmonoxide *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.carbonmonoxide", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_carbonmonoxide_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_carbonmonoxide *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_contact {
    struct client_resource base;
    struct oic_state_sensor_contact state;
};
static int
oic_client_sensor_contact_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_contact *res = (struct oic_client_sensor_contact *)resource;
    return oic_state_sensor_contact_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_contact_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_contact *state = (struct oic_client_sensor_contact *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_CONTACT__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_contact_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_contact_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_contact_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_contact_from_repr_vec,
        .inform_flow = oic_client_sensor_contact_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_CONTACT__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_CONTACT__OUT__DEVICE_ID
    };
    struct oic_client_sensor_contact *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.contact", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_contact_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_contact *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_geolocation {
    struct client_resource base;
    struct oic_state_sensor_geolocation state;
};

static bool
oic_client_sensor_geolocation_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_sensor_geolocation *state = (struct oic_client_sensor_geolocation *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("accuracy", state->state.accuracy));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("alt", state->state.alt));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("altitudeAccuracy", state->state.altitudeAccuracy));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("heading", state->state.heading));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("latitude", state->state.latitude));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("longitude", state->state.longitude));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("speed", state->state.speed));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_sensor_geolocation_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_geolocation *res = (struct oic_client_sensor_geolocation *)resource;
    return oic_state_sensor_geolocation_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_geolocation_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_geolocation *state = (struct oic_client_sensor_geolocation *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GEOLOCATION__OUT__ACCURACY, state->state.accuracy);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GEOLOCATION__OUT__ALT, state->state.alt);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GEOLOCATION__OUT__ALTITUDE_ACCURACY, state->state.altitudeAccuracy);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GEOLOCATION__OUT__HEADING, state->state.heading);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GEOLOCATION__OUT__LATITUDE, state->state.latitude);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GEOLOCATION__OUT__LONGITUDE, state->state.longitude);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GEOLOCATION__OUT__SPEED, state->state.speed);
}

static int
oic_client_sensor_geolocation_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_geolocation_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_geolocation_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_sensor_geolocation_to_repr_vec,
        .from_repr_vec = oic_client_sensor_geolocation_from_repr_vec,
        .inform_flow = oic_client_sensor_geolocation_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GEOLOCATION__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GEOLOCATION__OUT__DEVICE_ID
    };
    struct oic_client_sensor_geolocation *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.geolocation", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.accuracy = 0.0f;
resource->state.alt = 0.0f;
resource->state.altitudeAccuracy = 0.0f;
resource->state.heading = 0.0f;
resource->state.latitude = 0.0f;
resource->state.longitude = 0.0f;
resource->state.speed = 0.0f;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_geolocation_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_geolocation *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_sensor_geolocation_set_accuracy(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.accuracy, (double) var)) {
            resource->state.accuracy = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_sensor_geolocation_set_alt(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.alt, (double) var)) {
            resource->state.alt = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_sensor_geolocation_set_altitudeAccuracy(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.altitudeAccuracy, (double) var)) {
            resource->state.altitudeAccuracy = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_sensor_geolocation_set_heading(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.heading, (double) var)) {
            resource->state.heading = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_sensor_geolocation_set_latitude(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.latitude, (double) var)) {
            resource->state.latitude = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_sensor_geolocation_set_longitude(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.longitude, (double) var)) {
            resource->state.longitude = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_sensor_geolocation_set_speed(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.speed, (double) var)) {
            resource->state.speed = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_sensor_glassbreak {
    struct client_resource base;
    struct oic_state_sensor_glassbreak state;
};
static int
oic_client_sensor_glassbreak_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_glassbreak *res = (struct oic_client_sensor_glassbreak *)resource;
    return oic_state_sensor_glassbreak_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_glassbreak_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_glassbreak *state = (struct oic_client_sensor_glassbreak *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GLASSBREAK__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_glassbreak_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_glassbreak_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_glassbreak_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_glassbreak_from_repr_vec,
        .inform_flow = oic_client_sensor_glassbreak_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GLASSBREAK__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_GLASSBREAK__OUT__DEVICE_ID
    };
    struct oic_client_sensor_glassbreak *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.glassbreak", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_glassbreak_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_glassbreak *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor {
    struct client_resource base;
    struct oic_state_sensor state;
};
static int
oic_client_sensor_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor *res = (struct oic_client_sensor *)resource;
    return oic_state_sensor_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor *state = (struct oic_client_sensor *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_from_repr_vec,
        .inform_flow = oic_client_sensor_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR__OUT__DEVICE_ID
    };
    struct oic_client_sensor *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_motion {
    struct client_resource base;
    struct oic_state_sensor_motion state;
};
static int
oic_client_sensor_motion_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_motion *res = (struct oic_client_sensor_motion *)resource;
    return oic_state_sensor_motion_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_motion_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_motion *state = (struct oic_client_sensor_motion *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_MOTION__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_motion_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_motion_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_motion_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_motion_from_repr_vec,
        .inform_flow = oic_client_sensor_motion_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_MOTION__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_MOTION__OUT__DEVICE_ID
    };
    struct oic_client_sensor_motion *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.motion", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_motion_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_motion *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_presence {
    struct client_resource base;
    struct oic_state_sensor_presence state;
};
static int
oic_client_sensor_presence_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_presence *res = (struct oic_client_sensor_presence *)resource;
    return oic_state_sensor_presence_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_presence_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_presence *state = (struct oic_client_sensor_presence *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_PRESENCE__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_presence_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_presence_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_presence_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_presence_from_repr_vec,
        .inform_flow = oic_client_sensor_presence_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_PRESENCE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_PRESENCE__OUT__DEVICE_ID
    };
    struct oic_client_sensor_presence *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.presence", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_presence_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_presence *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_radiation_uv {
    struct client_resource base;
    struct oic_state_sensor_radiation_uv state;
};
static int
oic_client_sensor_radiation_uv_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_radiation_uv *res = (struct oic_client_sensor_radiation_uv *)resource;
    return oic_state_sensor_radiation_uv_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_radiation_uv_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_radiation_uv *state = (struct oic_client_sensor_radiation_uv *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_RADIATION_UV__OUT__MEASUREMENT, state->state.measurement);
}

static int
oic_client_sensor_radiation_uv_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_radiation_uv_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_radiation_uv_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_radiation_uv_from_repr_vec,
        .inform_flow = oic_client_sensor_radiation_uv_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_RADIATION_UV__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_RADIATION_UV__OUT__DEVICE_ID
    };
    struct oic_client_sensor_radiation_uv *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.radiation.uv", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.measurement = 0.0f;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_radiation_uv_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_radiation_uv *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_sleep {
    struct client_resource base;
    struct oic_state_sensor_sleep state;
};
static int
oic_client_sensor_sleep_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_sleep *res = (struct oic_client_sensor_sleep *)resource;
    return oic_state_sensor_sleep_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_sleep_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_sleep *state = (struct oic_client_sensor_sleep *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_SLEEP__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_sleep_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_sleep_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_sleep_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_sleep_from_repr_vec,
        .inform_flow = oic_client_sensor_sleep_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_SLEEP__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_SLEEP__OUT__DEVICE_ID
    };
    struct oic_client_sensor_sleep *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.sleep", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_sleep_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_sleep *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_smoke {
    struct client_resource base;
    struct oic_state_sensor_smoke state;
};
static int
oic_client_sensor_smoke_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_smoke *res = (struct oic_client_sensor_smoke *)resource;
    return oic_state_sensor_smoke_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_smoke_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_smoke *state = (struct oic_client_sensor_smoke *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_SMOKE__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_smoke_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_smoke_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_smoke_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_smoke_from_repr_vec,
        .inform_flow = oic_client_sensor_smoke_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_SMOKE__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_SMOKE__OUT__DEVICE_ID
    };
    struct oic_client_sensor_smoke *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.smoke", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_smoke_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_smoke *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_touch {
    struct client_resource base;
    struct oic_state_sensor_touch state;
};
static int
oic_client_sensor_touch_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_touch *res = (struct oic_client_sensor_touch *)resource;
    return oic_state_sensor_touch_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_touch_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_touch *state = (struct oic_client_sensor_touch *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_TOUCH__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_touch_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_touch_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_touch_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_touch_from_repr_vec,
        .inform_flow = oic_client_sensor_touch_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_TOUCH__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_TOUCH__OUT__DEVICE_ID
    };
    struct oic_client_sensor_touch *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.touch", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_touch_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_touch *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sensor_water {
    struct client_resource base;
    struct oic_state_sensor_water state;
};
static int
oic_client_sensor_water_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sensor_water *res = (struct oic_client_sensor_water *)resource;
    return oic_state_sensor_water_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sensor_water_inform_flow(struct client_resource *resource)
{
    struct oic_client_sensor_water *state = (struct oic_client_sensor_water *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_WATER__OUT__VALUE, state->state.value);
}

static int
oic_client_sensor_water_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sensor_water_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sensor_water_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sensor_water_from_repr_vec,
        .inform_flow = oic_client_sensor_water_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_WATER__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SENSOR_WATER__OUT__DEVICE_ID
    };
    struct oic_client_sensor_water *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sensor.water", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sensor_water_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sensor_water *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_signalstrength {
    struct client_resource base;
    struct oic_state_signalstrength state;
};
static int
oic_client_signalstrength_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_signalstrength *res = (struct oic_client_signalstrength *)resource;
    return oic_state_signalstrength_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_signalstrength_inform_flow(struct client_resource *resource)
{
    struct oic_client_signalstrength *state = (struct oic_client_signalstrength *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SIGNALSTRENGTH__OUT__LQI, state->state.lqi);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SIGNALSTRENGTH__OUT__RSSI, state->state.rssi);
}

static int
oic_client_signalstrength_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_signalstrength_options *node_opts =
        (const struct sol_flow_node_type_oic_client_signalstrength_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_signalstrength_from_repr_vec,
        .inform_flow = oic_client_signalstrength_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SIGNALSTRENGTH__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SIGNALSTRENGTH__OUT__DEVICE_ID
    };
    struct oic_client_signalstrength *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.signalstrength", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.lqi = 0.0f;
resource->state.rssi = 0.0f;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_signalstrength_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_signalstrength *resource = data;
    
    client_resource_close(&resource->base);
}

struct oic_client_sleep {
    struct client_resource base;
    struct oic_state_sleep state;
};
static int
oic_client_sleep_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_sleep *res = (struct oic_client_sleep *)resource;
    return oic_state_sleep_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_sleep_inform_flow(struct client_resource *resource)
{
    struct oic_client_sleep *state = (struct oic_client_sleep *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLEEP__OUT__AT, state->state.At);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLEEP__OUT__DST, state->state.Dst);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLEEP__OUT__LST, state->state.Lst);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLEEP__OUT__TST, state->state.Tst);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLEEP__OUT__TYPE, oic_state_sleep_Type_tbl[state->state.Type].key);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLEEP__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_sleep_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_sleep_options *node_opts =
        (const struct sol_flow_node_type_oic_client_sleep_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_sleep_from_repr_vec,
        .inform_flow = oic_client_sleep_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLEEP__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLEEP__OUT__DEVICE_ID
    };
    struct oic_client_sleep *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.sleep", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.At = 0.0f;
resource->state.Dst = 0.0f;
resource->state.Lst = 0.0f;
resource->state.Tst = 0.0f;
resource->state.Type = (enum oic_state_sleep_Type)0;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_sleep_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_sleep *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_slm {
    struct client_resource base;
    struct oic_state_slm state;
};
static int
oic_client_slm_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_slm *res = (struct oic_client_slm *)resource;
    return oic_state_slm_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_slm_inform_flow(struct client_resource *resource)
{
    struct oic_client_slm *state = (struct oic_client_slm *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLM__OUT__SLM, state->state.slm);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLM__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLM__OUT__UNITS, oic_state_slm_units_tbl[state->state.units].key);
}

static int
oic_client_slm_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_slm_options *node_opts =
        (const struct sol_flow_node_type_oic_client_slm_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = NULL,
        .from_repr_vec = oic_client_slm_from_repr_vec,
        .inform_flow = oic_client_slm_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLM__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SLM__OUT__DEVICE_ID
    };
    struct oic_client_slm *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.slm", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.slm = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_slm_units)0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_slm_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_slm *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

struct oic_client_speech_tts {
    struct client_resource base;
    struct oic_state_speech_tts state;
};

static bool
oic_client_speech_tts_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_speech_tts *state = (struct oic_client_speech_tts *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("utterance", state->state.utterance, state->state.utterance ? strlen(state->state.utterance) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_speech_tts_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_speech_tts *res = (struct oic_client_speech_tts *)resource;
    return oic_state_speech_tts_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_speech_tts_inform_flow(struct client_resource *resource)
{
    struct oic_client_speech_tts *state = (struct oic_client_speech_tts *)resource;
    send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEECH_TTS__OUT__SUPPORTED_LANGUAGES, state->state.supportedLanguages);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEECH_TTS__OUT__SUPPORTED_VOICES, state->state.supportedVoices);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEECH_TTS__OUT__UTTERANCE, state->state.utterance);
}

static int
oic_client_speech_tts_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_speech_tts_options *node_opts =
        (const struct sol_flow_node_type_oic_client_speech_tts_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_speech_tts_to_repr_vec,
        .from_repr_vec = oic_client_speech_tts_from_repr_vec,
        .inform_flow = oic_client_speech_tts_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEECH_TTS__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEECH_TTS__OUT__DEVICE_ID
    };
    struct oic_client_speech_tts *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.speech.tts", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.supportedLanguages = NULL;
resource->state.supportedVoices = NULL;
resource->state.utterance = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_speech_tts_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_speech_tts *resource = data;
    free(resource->state.supportedLanguages);
free(resource->state.supportedVoices);
free(resource->state.utterance);
    client_resource_close(&resource->base);
}

static int
oic_client_speech_tts_set_utterance(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_speech_tts *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.utterance, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            client_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_client_speed {
    struct client_resource base;
    struct oic_state_speed state;
};

static bool
oic_client_speed_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_speed *state = (struct oic_client_speed *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("speed", state->state.speed));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_speed_units_tbl[state->state.units].key, oic_state_speed_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_speed_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_speed *res = (struct oic_client_speed *)resource;
    return oic_state_speed_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_speed_inform_flow(struct client_resource *resource)
{
    struct oic_client_speed *state = (struct oic_client_speed *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEED__OUT__SPEED, state->state.speed);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEED__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEED__OUT__UNITS, oic_state_speed_units_tbl[state->state.units].key);
}

static int
oic_client_speed_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_speed_options *node_opts =
        (const struct sol_flow_node_type_oic_client_speed_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_speed_to_repr_vec,
        .from_repr_vec = oic_client_speed_from_repr_vec,
        .inform_flow = oic_client_speed_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEED__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SPEED__OUT__DEVICE_ID
    };
    struct oic_client_speed *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.speed", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.speed = 0;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_speed_units)0;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_speed_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_speed *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

static int
oic_client_speed_set_speed(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_speed *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.speed, (int32_t) var)) {
            resource->state.speed = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_client_speed_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_speed *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_speed_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_speed_units)val;
            client_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_client_steps {
    struct client_resource base;
    struct oic_state_steps state;
};

static bool
oic_client_steps_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_steps *state = (struct oic_client_steps *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("count", state->state.count));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_steps_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_steps *res = (struct oic_client_steps *)resource;
    return oic_state_steps_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_steps_inform_flow(struct client_resource *resource)
{
    struct oic_client_steps *state = (struct oic_client_steps *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_STEPS__OUT__COUNT, state->state.count);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_STEPS__OUT__DISTANCE, state->state.distance);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_STEPS__OUT__KCAL, state->state.kcal);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_STEPS__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_client_steps_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_steps_options *node_opts =
        (const struct sol_flow_node_type_oic_client_steps_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_steps_to_repr_vec,
        .from_repr_vec = oic_client_steps_from_repr_vec,
        .inform_flow = oic_client_steps_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_STEPS__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_STEPS__OUT__DEVICE_ID
    };
    struct oic_client_steps *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.steps", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.count = 0;
resource->state.distance = 0;
resource->state.kcal = 0.0f;
resource->state.observedtime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_steps_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_steps *resource = data;
    free(resource->state.observedtime);
    client_resource_close(&resource->base);
}

static int
oic_client_steps_set_count(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_steps *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.count, (int32_t) var)) {
            resource->state.count = (int32_t) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_switch_binary {
    struct client_resource base;
    struct oic_state_switch_binary state;
};

static bool
oic_client_switch_binary_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_switch_binary *state = (struct oic_client_switch_binary *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("value", state->state.value));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_switch_binary_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_switch_binary *res = (struct oic_client_switch_binary *)resource;
    return oic_state_switch_binary_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_switch_binary_inform_flow(struct client_resource *resource)
{
    struct oic_client_switch_binary *state = (struct oic_client_switch_binary *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_SWITCH_BINARY__OUT__VALUE, state->state.value);
}

static int
oic_client_switch_binary_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_switch_binary_options *node_opts =
        (const struct sol_flow_node_type_oic_client_switch_binary_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_switch_binary_to_repr_vec,
        .from_repr_vec = oic_client_switch_binary_from_repr_vec,
        .inform_flow = oic_client_switch_binary_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SWITCH_BINARY__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_SWITCH_BINARY__OUT__DEVICE_ID
    };
    struct oic_client_switch_binary *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.switch.binary", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.value = false;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_switch_binary_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_switch_binary *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_switch_binary_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_switch_binary *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_client_time_period {
    struct client_resource base;
    struct oic_state_time_period state;
};

static bool
oic_client_time_period_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_time_period *state = (struct oic_client_time_period *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("startTime", state->state.startTime, state->state.startTime ? strlen(state->state.startTime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("stopTime", state->state.stopTime, state->state.stopTime ? strlen(state->state.stopTime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_time_period_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_time_period *res = (struct oic_client_time_period *)resource;
    return oic_state_time_period_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_time_period_inform_flow(struct client_resource *resource)
{
    struct oic_client_time_period *state = (struct oic_client_time_period *)resource;
    send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_TIME_PERIOD__OUT__START_TIME, state->state.startTime);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_TIME_PERIOD__OUT__STOP_TIME, state->state.stopTime);
}

static int
oic_client_time_period_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_time_period_options *node_opts =
        (const struct sol_flow_node_type_oic_client_time_period_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_time_period_to_repr_vec,
        .from_repr_vec = oic_client_time_period_from_repr_vec,
        .inform_flow = oic_client_time_period_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_TIME_PERIOD__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_TIME_PERIOD__OUT__DEVICE_ID
    };
    struct oic_client_time_period *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.time.period", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.startTime = NULL;
resource->state.stopTime = NULL;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_time_period_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_time_period *resource = data;
    free(resource->state.startTime);
free(resource->state.stopTime);
    client_resource_close(&resource->base);
}

static int
oic_client_time_period_set_startTime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_time_period *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.startTime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            client_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_client_time_period_set_stopTime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_time_period *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.stopTime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            client_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_client_weight {
    struct client_resource base;
    struct oic_state_weight state;
};

static bool
oic_client_weight_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_client_weight *state = (struct oic_client_weight *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("weight", state->state.weight));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_client_weight_from_repr_vec(struct client_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_client_weight *res = (struct oic_client_weight *)resource;
    return oic_state_weight_from_repr_vec(&res->state, repr_vec, ~0);
}

static void oic_client_weight_inform_flow(struct client_resource *resource)
{
    struct oic_client_weight *state = (struct oic_client_weight *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_CLIENT_WEIGHT__OUT__WEIGHT, state->state.weight);
}

static int
oic_client_weight_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    const struct sol_flow_node_type_oic_client_weight_options *node_opts =
        (const struct sol_flow_node_type_oic_client_weight_options *)options;
    static const struct client_resource_funcs funcs = {
        .to_repr_vec = oic_client_weight_to_repr_vec,
        .from_repr_vec = oic_client_weight_from_repr_vec,
        .inform_flow = oic_client_weight_inform_flow,
        .found_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_WEIGHT__OUT__FOUND,
        .device_id_port = SOL_FLOW_NODE_TYPE_OIC_CLIENT_WEIGHT__OUT__DEVICE_ID
    };
    struct oic_client_weight *resource = data;
    int r;

    r = client_resource_init(node, &resource->base, "oic.r.weight", &funcs);
    SOL_INT_CHECK(r, < 0, r);
    resource->state.weight = 0.0f;

    return client_connect(&resource->base, node_opts->device_id);
}

static void oic_client_weight_close(struct sol_flow_node *node, void *data)
{
    struct oic_client_weight *resource = data;
    
    client_resource_close(&resource->base);
}

static int
oic_client_weight_set_weight(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_client_weight *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.weight, (double) var)) {
            resource->state.weight = (double) var;
            client_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_light {
    struct server_resource base;
    struct oic_state_light state;
};

static bool
oic_server_light_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_light *state = (struct oic_server_light *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("state", state->state.state));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("power", state->state.power));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("name", state->state.name, state->state.name ? strlen(state->state.name) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_light_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_light *res = (struct oic_server_light *)resource;
    return oic_state_light_from_repr_vec(&res->state, repr_vec, 0x7);
}

static void oic_server_light_inform_flow(struct server_resource *resource)
{
    struct oic_server_light *state = (struct oic_server_light *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_LIGHT__OUT__STATE, state->state.state);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_LIGHT__OUT__POWER, state->state.power);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_LIGHT__OUT__NAME, state->state.name);
}

static int
oic_server_light_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("core.light");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_light_to_repr_vec,
        .from_repr_vec = oic_server_light_from_repr_vec,
        .inform_flow = oic_server_light_inform_flow
    };
    struct oic_server_light *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.state = false;
resource->state.power = 0;
resource->state.name = NULL;
    }

    return r;
}

static void oic_server_light_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_light *resource = data;
    free(resource->state.name);
    server_resource_close(&resource->base);
}

static int
oic_server_light_set_state(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_light *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.state, (bool) var)) {
            resource->state.state = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_light_set_power(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_light *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.power, (int32_t) var)) {
            resource->state.power = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_light_set_name(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_light *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.name, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_activity {
    struct server_resource base;
    struct oic_state_activity state;
};

static bool
oic_server_activity_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_activity *state = (struct oic_server_activity *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("Acalory", state->state.Acalory));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("Atime", state->state.Atime));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("Type", oic_state_activity_Type_tbl[state->state.Type].key, oic_state_activity_Type_tbl[state->state.Type].len));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_activity_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_activity *res = (struct oic_server_activity *)resource;
    return oic_state_activity_from_repr_vec(&res->state, repr_vec, 0x7);
}

static void oic_server_activity_inform_flow(struct server_resource *resource)
{
    struct oic_server_activity *state = (struct oic_server_activity *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ACTIVITY__OUT__ACALORY, state->state.Acalory);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ACTIVITY__OUT__ATIME, state->state.Atime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ACTIVITY__OUT__TYPE, oic_state_activity_Type_tbl[state->state.Type].key);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ACTIVITY__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_server_activity_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.activity");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_activity_to_repr_vec,
        .from_repr_vec = oic_server_activity_from_repr_vec,
        .inform_flow = oic_server_activity_inform_flow
    };
    struct oic_server_activity *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.Acalory = 0;
resource->state.Atime = 0;
resource->state.Type = (enum oic_state_activity_Type)0;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_activity_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_activity *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_activity_set_Acalory(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_activity *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.Acalory, (int32_t) var)) {
            resource->state.Acalory = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_activity_set_Atime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_activity *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.Atime, (int32_t) var)) {
            resource->state.Atime = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_activity_set_Type(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_activity *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_activity_Type_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.Type = (enum oic_state_activity_Type)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}

static int
oic_server_activity_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_activity *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_altimeter {
    struct server_resource base;
    struct oic_state_altimeter state;
};

static bool
oic_server_altimeter_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_altimeter *state = (struct oic_server_altimeter *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("alt", state->state.alt));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_altimeter_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_altimeter *res = (struct oic_server_altimeter *)resource;
    return oic_state_altimeter_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_altimeter_inform_flow(struct server_resource *resource)
{
    struct oic_server_altimeter *state = (struct oic_server_altimeter *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ALTIMETER__OUT__ALT, state->state.alt);
}

static int
oic_server_altimeter_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.altimeter");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_altimeter_to_repr_vec,
        .from_repr_vec = oic_server_altimeter_from_repr_vec,
        .inform_flow = oic_server_altimeter_inform_flow
    };
    struct oic_server_altimeter *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.alt = 0.0f;
    }

    return r;
}

static void oic_server_altimeter_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_altimeter *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_altimeter_set_alt(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_altimeter *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.alt, (double) var)) {
            resource->state.alt = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_audio {
    struct server_resource base;
    struct oic_state_audio state;
};

static bool
oic_server_audio_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_audio *state = (struct oic_server_audio *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("mute", state->state.mute));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("volume", state->state.volume));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_audio_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_audio *res = (struct oic_server_audio *)resource;
    return oic_state_audio_from_repr_vec(&res->state, repr_vec, 0x3);
}

static void oic_server_audio_inform_flow(struct server_resource *resource)
{
    struct oic_server_audio *state = (struct oic_server_audio *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_AUDIO__OUT__MUTE, state->state.mute);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_AUDIO__OUT__VOLUME, state->state.volume);
}

static int
oic_server_audio_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.audio");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_audio_to_repr_vec,
        .from_repr_vec = oic_server_audio_from_repr_vec,
        .inform_flow = oic_server_audio_inform_flow
    };
    struct oic_server_audio *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.mute = false;
resource->state.volume = 0;
    }

    return r;
}

static void oic_server_audio_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_audio *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_audio_set_mute(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_audio *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.mute, (bool) var)) {
            resource->state.mute = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_audio_set_volume(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_audio *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.volume, (int32_t) var)) {
            resource->state.volume = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_autofocus {
    struct server_resource base;
    struct oic_state_autofocus state;
};

static bool
oic_server_autofocus_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_autofocus *state = (struct oic_server_autofocus *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("autoFocus", state->state.autoFocus));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_autofocus_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_autofocus *res = (struct oic_server_autofocus *)resource;
    return oic_state_autofocus_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_autofocus_inform_flow(struct server_resource *resource)
{
    struct oic_server_autofocus *state = (struct oic_server_autofocus *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_AUTOFOCUS__OUT__AUTO_FOCUS, state->state.autoFocus);
}

static int
oic_server_autofocus_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.autofocus");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_autofocus_to_repr_vec,
        .from_repr_vec = oic_server_autofocus_from_repr_vec,
        .inform_flow = oic_server_autofocus_inform_flow
    };
    struct oic_server_autofocus *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.autoFocus = false;
    }

    return r;
}

static void oic_server_autofocus_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_autofocus *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_autofocus_set_autoFocus(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_autofocus *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.autoFocus, (bool) var)) {
            resource->state.autoFocus = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_automaticdocumentfeeder {
    struct server_resource base;
    struct oic_state_automaticdocumentfeeder state;
};

static bool
oic_server_automaticdocumentfeeder_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_automaticdocumentfeeder *state = (struct oic_server_automaticdocumentfeeder *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("adfStates", state->state.adfStates, state->state.adfStates ? strlen(state->state.adfStates) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("currentAdfState", state->state.currentAdfState, state->state.currentAdfState ? strlen(state->state.currentAdfState) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_automaticdocumentfeeder_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.automaticdocumentfeeder");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_automaticdocumentfeeder_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_automaticdocumentfeeder *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.adfStates = NULL;
resource->state.currentAdfState = NULL;
    }

    return r;
}

static void oic_server_automaticdocumentfeeder_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_automaticdocumentfeeder *resource = data;
    free(resource->state.adfStates);
free(resource->state.currentAdfState);
    server_resource_close(&resource->base);
}

static int
oic_server_automaticdocumentfeeder_set_adfStates(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_automaticdocumentfeeder *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.adfStates, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_automaticdocumentfeeder_set_currentAdfState(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_automaticdocumentfeeder *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.currentAdfState, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_blood_glucose {
    struct server_resource base;
    struct oic_state_blood_glucose state;
};

static bool
oic_server_blood_glucose_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_blood_glucose *state = (struct oic_server_blood_glucose *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("bloodsugar", state->state.bloodsugar));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("site", state->state.site, state->state.site ? strlen(state->state.site) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_blood_glucose_units_tbl[state->state.units].key, oic_state_blood_glucose_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_blood_glucose_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.blood.glucose");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_blood_glucose_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_blood_glucose *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.bloodsugar = 0.0f;
resource->state.observedtime = NULL;
resource->state.site = NULL;
resource->state.units = (enum oic_state_blood_glucose_units)0;
    }

    return r;
}

static void oic_server_blood_glucose_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_blood_glucose *resource = data;
    free(resource->state.observedtime);
free(resource->state.site);
    server_resource_close(&resource->base);
}

static int
oic_server_blood_glucose_set_bloodsugar(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_glucose *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.bloodsugar, (double) var)) {
            resource->state.bloodsugar = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_blood_glucose_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_glucose *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_blood_glucose_set_site(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_glucose *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.site, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_blood_glucose_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_glucose *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_blood_glucose_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_blood_glucose_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_server_blood_o2_saturation {
    struct server_resource base;
    struct oic_state_blood_o2_saturation state;
};

static bool
oic_server_blood_o2_saturation_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_blood_o2_saturation *state = (struct oic_server_blood_o2_saturation *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("perfusion", state->state.perfusion));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("spo2", state->state.spo2));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_blood_o2_saturation_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.blood.o2.saturation");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_blood_o2_saturation_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_blood_o2_saturation *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.perfusion = 0.0f;
resource->state.spo2 = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_blood_o2_saturation_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_blood_o2_saturation *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_blood_o2_saturation_set_perfusion(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_o2_saturation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.perfusion, (double) var)) {
            resource->state.perfusion = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_blood_o2_saturation_set_spo2(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_o2_saturation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.spo2, (double) var)) {
            resource->state.spo2 = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_blood_o2_saturation_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_o2_saturation *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_blood_pressure {
    struct server_resource base;
    struct oic_state_blood_pressure state;
};

static bool
oic_server_blood_pressure_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_blood_pressure *state = (struct oic_server_blood_pressure *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("map", state->state.map));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("bloodpressure", state->state.bloodpressure, state->state.bloodpressure ? strlen(state->state.bloodpressure) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_blood_pressure_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.blood.pressure");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_blood_pressure_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_blood_pressure *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.map = 0.0f;
resource->state.bloodpressure = NULL;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_blood_pressure_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_blood_pressure *resource = data;
    free(resource->state.bloodpressure);
free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_blood_pressure_set_map(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_pressure *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.map, (double) var)) {
            resource->state.map = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_blood_pressure_set_bloodpressure(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_pressure *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.bloodpressure, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_blood_pressure_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_blood_pressure *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_bmi {
    struct server_resource base;
    struct oic_state_bmi state;
};

static bool
oic_server_bmi_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_bmi *state = (struct oic_server_bmi *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("bmi", state->state.bmi));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("bmip", state->state.bmip));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_bmi_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.bmi");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_bmi_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_bmi *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.bmi = 0.0f;
resource->state.bmip = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_bmi_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_bmi *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_bmi_set_bmi(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bmi *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.bmi, (double) var)) {
            resource->state.bmi = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_bmi_set_bmip(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bmi *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.bmip, (double) var)) {
            resource->state.bmip = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_bmi_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bmi *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_bodyfat {
    struct server_resource base;
    struct oic_state_bodyfat state;
};

static bool
oic_server_bodyfat_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_bodyfat *state = (struct oic_server_bodyfat *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_bodyfat_units_tbl[state->state.units].key, oic_state_bodyfat_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("bodyfat", state->state.bodyfat));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_bodyfat_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.bodyfat");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_bodyfat_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_bodyfat *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.units = (enum oic_state_bodyfat_units)0;
resource->state.bodyfat = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_bodyfat_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_bodyfat *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_bodyfat_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodyfat *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_bodyfat_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_bodyfat_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}

static int
oic_server_bodyfat_set_bodyfat(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodyfat *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.bodyfat, (double) var)) {
            resource->state.bodyfat = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_bodyfat_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodyfat *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_bodyheight {
    struct server_resource base;
    struct oic_state_bodyheight state;
};

static bool
oic_server_bodyheight_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_bodyheight *state = (struct oic_server_bodyheight *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_bodyheight_units_tbl[state->state.units].key, oic_state_bodyheight_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("height", state->state.height));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_bodyheight_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_bodyheight *res = (struct oic_server_bodyheight *)resource;
    return oic_state_bodyheight_from_repr_vec(&res->state, repr_vec, 0x3);
}

static void oic_server_bodyheight_inform_flow(struct server_resource *resource)
{
    struct oic_server_bodyheight *state = (struct oic_server_bodyheight *)resource;
    sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_BODYHEIGHT__OUT__UNITS, oic_state_bodyheight_units_tbl[state->state.units].key);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_BODYHEIGHT__OUT__HEIGHT, state->state.height);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_BODYHEIGHT__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_server_bodyheight_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.bodyheight");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_bodyheight_to_repr_vec,
        .from_repr_vec = oic_server_bodyheight_from_repr_vec,
        .inform_flow = oic_server_bodyheight_inform_flow
    };
    struct oic_server_bodyheight *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.units = (enum oic_state_bodyheight_units)0;
resource->state.height = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_bodyheight_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_bodyheight *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_bodyheight_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodyheight *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_bodyheight_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_bodyheight_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}

static int
oic_server_bodyheight_set_height(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodyheight *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.height, (double) var)) {
            resource->state.height = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_bodyheight_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodyheight *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_body_position {
    struct server_resource base;
    struct oic_state_body_position state;
};

static bool
oic_server_body_position_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_body_position *state = (struct oic_server_body_position *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("bpos", oic_state_body_position_bpos_tbl[state->state.bpos].key, oic_state_body_position_bpos_tbl[state->state.bpos].len));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_body_position_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.body.position");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_body_position_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_body_position *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.bpos = (enum oic_state_body_position_bpos)0;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_body_position_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_body_position *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_body_position_set_bpos(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_body_position *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_body_position_bpos_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.bpos = (enum oic_state_body_position_bpos)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}

static int
oic_server_body_position_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_body_position *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_bodysite {
    struct server_resource base;
    struct oic_state_bodysite state;
};

static bool
oic_server_bodysite_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_bodysite *state = (struct oic_server_bodysite *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("bodysite", state->state.bodysite));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_bodysite_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.bodysite");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_bodysite_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_bodysite *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.bodysite = 0;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_bodysite_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_bodysite *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_bodysite_set_bodysite(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodysite *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.bodysite, (int32_t) var)) {
            resource->state.bodysite = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_bodysite_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodysite *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_bodywater {
    struct server_resource base;
    struct oic_state_bodywater state;
};

static bool
oic_server_bodywater_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_bodywater *state = (struct oic_server_bodywater *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_bodywater_units_tbl[state->state.units].key, oic_state_bodywater_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("bodywater", state->state.bodywater));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_bodywater_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.bodywater");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_bodywater_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_bodywater *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.units = (enum oic_state_bodywater_units)0;
resource->state.bodywater = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_bodywater_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_bodywater *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_bodywater_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodywater *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_bodywater_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_bodywater_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}

static int
oic_server_bodywater_set_bodywater(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodywater *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.bodywater, (double) var)) {
            resource->state.bodywater = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_bodywater_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodywater *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_bodyweight {
    struct server_resource base;
    struct oic_state_bodyweight state;
};

static bool
oic_server_bodyweight_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_bodyweight *state = (struct oic_server_bodyweight *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("weight", state->state.weight));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_bodyweight_units_tbl[state->state.units].key, oic_state_bodyweight_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_bodyweight_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_bodyweight *res = (struct oic_server_bodyweight *)resource;
    return oic_state_bodyweight_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_bodyweight_inform_flow(struct server_resource *resource)
{
    struct oic_server_bodyweight *state = (struct oic_server_bodyweight *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_BODYWEIGHT__OUT__WEIGHT, state->state.weight);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_BODYWEIGHT__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_BODYWEIGHT__OUT__UNITS, oic_state_bodyweight_units_tbl[state->state.units].key);
}

static int
oic_server_bodyweight_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.bodyweight");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_bodyweight_to_repr_vec,
        .from_repr_vec = oic_server_bodyweight_from_repr_vec,
        .inform_flow = oic_server_bodyweight_inform_flow
    };
    struct oic_server_bodyweight *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.weight = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_bodyweight_units)0;
    }

    return r;
}

static void oic_server_bodyweight_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_bodyweight *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_bodyweight_set_weight(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodyweight *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.weight, (double) var)) {
            resource->state.weight = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_bodyweight_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodyweight *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_bodyweight_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_bodyweight *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_bodyweight_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_bodyweight_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_server_clock {
    struct server_resource base;
    struct oic_state_clock state;
};

static bool
oic_server_clock_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_clock *state = (struct oic_server_clock *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("countdown", state->state.countdown));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("datetime", state->state.datetime, state->state.datetime ? strlen(state->state.datetime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_clock_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_clock *res = (struct oic_server_clock *)resource;
    return oic_state_clock_from_repr_vec(&res->state, repr_vec, 0x3);
}

static void oic_server_clock_inform_flow(struct server_resource *resource)
{
    struct oic_server_clock *state = (struct oic_server_clock *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CLOCK__OUT__COUNTDOWN, state->state.countdown);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CLOCK__OUT__DATETIME, state->state.datetime);
}

static int
oic_server_clock_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.clock");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_clock_to_repr_vec,
        .from_repr_vec = oic_server_clock_from_repr_vec,
        .inform_flow = oic_server_clock_inform_flow
    };
    struct oic_server_clock *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.countdown = 0.0f;
resource->state.datetime = NULL;
    }

    return r;
}

static void oic_server_clock_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_clock *resource = data;
    free(resource->state.datetime);
    server_resource_close(&resource->base);
}

static int
oic_server_clock_set_countdown(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_clock *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.countdown, (double) var)) {
            resource->state.countdown = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_clock_set_datetime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_clock *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.datetime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_colour_autowhitebalance {
    struct server_resource base;
    struct oic_state_colour_autowhitebalance state;
};

static bool
oic_server_colour_autowhitebalance_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_colour_autowhitebalance *state = (struct oic_server_colour_autowhitebalance *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("autoWhiteBalance", state->state.autoWhiteBalance));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_colour_autowhitebalance_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_colour_autowhitebalance *res = (struct oic_server_colour_autowhitebalance *)resource;
    return oic_state_colour_autowhitebalance_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_colour_autowhitebalance_inform_flow(struct server_resource *resource)
{
    struct oic_server_colour_autowhitebalance *state = (struct oic_server_colour_autowhitebalance *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_COLOUR_AUTOWHITEBALANCE__OUT__AUTO_WHITE_BALANCE, state->state.autoWhiteBalance);
}

static int
oic_server_colour_autowhitebalance_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.colour.autowhitebalance");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_colour_autowhitebalance_to_repr_vec,
        .from_repr_vec = oic_server_colour_autowhitebalance_from_repr_vec,
        .inform_flow = oic_server_colour_autowhitebalance_inform_flow
    };
    struct oic_server_colour_autowhitebalance *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.autoWhiteBalance = false;
    }

    return r;
}

static void oic_server_colour_autowhitebalance_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_colour_autowhitebalance *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_colour_autowhitebalance_set_autoWhiteBalance(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_colour_autowhitebalance *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.autoWhiteBalance, (bool) var)) {
            resource->state.autoWhiteBalance = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_colour_saturation {
    struct server_resource base;
    struct oic_state_colour_saturation state;
};

static bool
oic_server_colour_saturation_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_colour_saturation *state = (struct oic_server_colour_saturation *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("colourSaturation", state->state.colourSaturation));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_colour_saturation_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_colour_saturation *res = (struct oic_server_colour_saturation *)resource;
    return oic_state_colour_saturation_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_colour_saturation_inform_flow(struct server_resource *resource)
{
    struct oic_server_colour_saturation *state = (struct oic_server_colour_saturation *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_COLOUR_SATURATION__OUT__COLOUR_SATURATION, state->state.colourSaturation);
}

static int
oic_server_colour_saturation_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.colour.saturation");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_colour_saturation_to_repr_vec,
        .from_repr_vec = oic_server_colour_saturation_from_repr_vec,
        .inform_flow = oic_server_colour_saturation_inform_flow
    };
    struct oic_server_colour_saturation *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.colourSaturation = 0;
    }

    return r;
}

static void oic_server_colour_saturation_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_colour_saturation *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_colour_saturation_set_colourSaturation(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_colour_saturation *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.colourSaturation, (int32_t) var)) {
            resource->state.colourSaturation = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_cycling_power {
    struct server_resource base;
    struct oic_state_cycling_power state;
};

static bool
oic_server_cycling_power_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_cycling_power *state = (struct oic_server_cycling_power *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("leftonly", state->state.leftonly));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("al_power", state->state.al_power));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("ar_power", state->state.ar_power));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("ftp", state->state.ftp));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("power", state->state.power));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_cycling_power_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_cycling_power *res = (struct oic_server_cycling_power *)resource;
    return oic_state_cycling_power_from_repr_vec(&res->state, repr_vec, 0x1f);
}

static void oic_server_cycling_power_inform_flow(struct server_resource *resource)
{
    struct oic_server_cycling_power *state = (struct oic_server_cycling_power *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_POWER__OUT__LEFTONLY, state->state.leftonly);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_POWER__OUT__AL_POWER, state->state.al_power);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_POWER__OUT__AR_POWER, state->state.ar_power);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_POWER__OUT__FTP, state->state.ftp);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_POWER__OUT__POWER, state->state.power);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_POWER__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_server_cycling_power_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.cycling.power");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_cycling_power_to_repr_vec,
        .from_repr_vec = oic_server_cycling_power_from_repr_vec,
        .inform_flow = oic_server_cycling_power_inform_flow
    };
    struct oic_server_cycling_power *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.leftonly = false;
resource->state.al_power = 0.0f;
resource->state.ar_power = 0.0f;
resource->state.ftp = 0.0f;
resource->state.power = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_cycling_power_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_cycling_power *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_cycling_power_set_leftonly(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_power *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.leftonly, (bool) var)) {
            resource->state.leftonly = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_cycling_power_set_al_power(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_power *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.al_power, (double) var)) {
            resource->state.al_power = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_cycling_power_set_ar_power(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_power *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.ar_power, (double) var)) {
            resource->state.ar_power = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_cycling_power_set_ftp(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_power *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.ftp, (double) var)) {
            resource->state.ftp = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_cycling_power_set_power(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_power *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.power, (double) var)) {
            resource->state.power = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_cycling_power_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_power *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_cycling_speedcadence {
    struct server_resource base;
    struct oic_state_cycling_speedcadence state;
};

static bool
oic_server_cycling_speedcadence_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_cycling_speedcadence *state = (struct oic_server_cycling_speedcadence *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("speed", state->state.speed));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("cadence", state->state.cadence));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_cycling_speedcadence_units_tbl[state->state.units].key, oic_state_cycling_speedcadence_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_cycling_speedcadence_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_cycling_speedcadence *res = (struct oic_server_cycling_speedcadence *)resource;
    return oic_state_cycling_speedcadence_from_repr_vec(&res->state, repr_vec, 0x9);
}

static void oic_server_cycling_speedcadence_inform_flow(struct server_resource *resource)
{
    struct oic_server_cycling_speedcadence *state = (struct oic_server_cycling_speedcadence *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_SPEEDCADENCE__OUT__SPEED, state->state.speed);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_SPEEDCADENCE__OUT__CADENCE, state->state.cadence);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_SPEEDCADENCE__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_CYCLING_SPEEDCADENCE__OUT__UNITS, oic_state_cycling_speedcadence_units_tbl[state->state.units].key);
}

static int
oic_server_cycling_speedcadence_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.cycling.speedcadence");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_cycling_speedcadence_to_repr_vec,
        .from_repr_vec = oic_server_cycling_speedcadence_from_repr_vec,
        .inform_flow = oic_server_cycling_speedcadence_inform_flow
    };
    struct oic_server_cycling_speedcadence *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.speed = 0;
resource->state.cadence = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_cycling_speedcadence_units)0;
    }

    return r;
}

static void oic_server_cycling_speedcadence_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_cycling_speedcadence *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_cycling_speedcadence_set_speed(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_speedcadence *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.speed, (int32_t) var)) {
            resource->state.speed = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_cycling_speedcadence_set_cadence(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_speedcadence *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.cadence, (double) var)) {
            resource->state.cadence = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_cycling_speedcadence_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_speedcadence *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_cycling_speedcadence_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_cycling_speedcadence *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_cycling_speedcadence_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_cycling_speedcadence_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_server_distance {
    struct server_resource base;
    struct oic_state_distance state;
};

static bool
oic_server_distance_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_distance *state = (struct oic_server_distance *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("distance", state->state.distance));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_distance_units_tbl[state->state.units].key, oic_state_distance_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_distance_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_distance *res = (struct oic_server_distance *)resource;
    return oic_state_distance_from_repr_vec(&res->state, repr_vec, 0x5);
}

static void oic_server_distance_inform_flow(struct server_resource *resource)
{
    struct oic_server_distance *state = (struct oic_server_distance *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_DISTANCE__OUT__DISTANCE, state->state.distance);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_DISTANCE__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_DISTANCE__OUT__UNITS, oic_state_distance_units_tbl[state->state.units].key);
}

static int
oic_server_distance_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.distance");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_distance_to_repr_vec,
        .from_repr_vec = oic_server_distance_from_repr_vec,
        .inform_flow = oic_server_distance_inform_flow
    };
    struct oic_server_distance *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.distance = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_distance_units)0;
    }

    return r;
}

static void oic_server_distance_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_distance *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_distance_set_distance(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_distance *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.distance, (double) var)) {
            resource->state.distance = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_distance_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_distance *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_distance_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_distance *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_distance_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_distance_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_server_door {
    struct server_resource base;
    struct oic_state_door state;
};

static bool
oic_server_door_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_door *state = (struct oic_server_door *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("openAlarm", state->state.openAlarm));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("openState", oic_state_door_openState_tbl[state->state.openState].key, oic_state_door_openState_tbl[state->state.openState].len));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("openDuration", state->state.openDuration, state->state.openDuration ? strlen(state->state.openDuration) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_door_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_door *res = (struct oic_server_door *)resource;
    return oic_state_door_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_door_inform_flow(struct server_resource *resource)
{
    struct oic_server_door *state = (struct oic_server_door *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_DOOR__OUT__OPEN_ALARM, state->state.openAlarm);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_DOOR__OUT__OPEN_STATE, oic_state_door_openState_tbl[state->state.openState].key);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_DOOR__OUT__OPEN_DURATION, state->state.openDuration);
}

static int
oic_server_door_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.door");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_door_to_repr_vec,
        .from_repr_vec = oic_server_door_from_repr_vec,
        .inform_flow = oic_server_door_inform_flow
    };
    struct oic_server_door *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.openAlarm = false;
resource->state.openState = (enum oic_state_door_openState)0;
resource->state.openDuration = NULL;
    }

    return r;
}

static void oic_server_door_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_door *resource = data;
    free(resource->state.openDuration);
    server_resource_close(&resource->base);
}

static int
oic_server_door_set_openAlarm(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_door *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.openAlarm, (bool) var)) {
            resource->state.openAlarm = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_door_set_openState(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_door *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_door_openState_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.openState = (enum oic_state_door_openState)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}

static int
oic_server_door_set_openDuration(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_door *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.openDuration, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_emg {
    struct server_resource base;
    struct oic_state_emg state;
};

static bool
oic_server_emg_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_emg *state = (struct oic_server_emg *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("mi", state->state.mi));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_emg_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.emg");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_emg_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_emg *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.mi = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_emg_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_emg *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_emg_set_mi(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_emg *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.mi, (double) var)) {
            resource->state.mi = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_emg_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_emg *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_energy_battery {
    struct server_resource base;
    struct oic_state_energy_battery state;
};

static bool
oic_server_energy_battery_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_energy_battery *state = (struct oic_server_energy_battery *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("charge", state->state.charge));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_energy_battery_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.energy.battery");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_energy_battery_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_energy_battery *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.charge = 0;
    }

    return r;
}

static void oic_server_energy_battery_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_energy_battery *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_energy_battery_set_charge(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_energy_battery *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.charge, (int32_t) var)) {
            resource->state.charge = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_energy_consumption {
    struct server_resource base;
    struct oic_state_energy_consumption state;
};

static bool
oic_server_energy_consumption_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_energy_consumption *state = (struct oic_server_energy_consumption *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("energy", state->state.energy));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("power", state->state.power));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_energy_consumption_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.energy.consumption");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_energy_consumption_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_energy_consumption *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.energy = 0.0f;
resource->state.power = 0.0f;
    }

    return r;
}

static void oic_server_energy_consumption_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_energy_consumption *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_energy_consumption_set_energy(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_energy_consumption *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.energy, (double) var)) {
            resource->state.energy = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_energy_consumption_set_power(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_energy_consumption *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.power, (double) var)) {
            resource->state.power = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_energy_drlc {
    struct server_resource base;
    struct oic_state_energy_drlc state;
};

static bool
oic_server_energy_drlc_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_energy_drlc *state = (struct oic_server_energy_drlc *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("override", state->state.override));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("DRType", state->state.DRType));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("duration", state->state.duration));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("start", state->state.start, state->state.start ? strlen(state->state.start) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_energy_drlc_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_energy_drlc *res = (struct oic_server_energy_drlc *)resource;
    return oic_state_energy_drlc_from_repr_vec(&res->state, repr_vec, 0xf);
}

static void oic_server_energy_drlc_inform_flow(struct server_resource *resource)
{
    struct oic_server_energy_drlc *state = (struct oic_server_energy_drlc *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ENERGY_DRLC__OUT__OVERRIDE, state->state.override);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ENERGY_DRLC__OUT__D_RTYPE, state->state.DRType);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ENERGY_DRLC__OUT__DURATION, state->state.duration);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ENERGY_DRLC__OUT__START, state->state.start);
}

static int
oic_server_energy_drlc_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.energy.drlc");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_energy_drlc_to_repr_vec,
        .from_repr_vec = oic_server_energy_drlc_from_repr_vec,
        .inform_flow = oic_server_energy_drlc_inform_flow
    };
    struct oic_server_energy_drlc *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.override = false;
resource->state.DRType = 0;
resource->state.duration = 0;
resource->state.start = NULL;
    }

    return r;
}

static void oic_server_energy_drlc_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_energy_drlc *resource = data;
    free(resource->state.start);
    server_resource_close(&resource->base);
}

static int
oic_server_energy_drlc_set_override(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_energy_drlc *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.override, (bool) var)) {
            resource->state.override = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_energy_drlc_set_DRType(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_energy_drlc *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.DRType, (int32_t) var)) {
            resource->state.DRType = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_energy_drlc_set_duration(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_energy_drlc *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.duration, (int32_t) var)) {
            resource->state.duration = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_energy_drlc_set_start(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_energy_drlc *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.start, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_energy_overload {
    struct server_resource base;
    struct oic_state_energy_overload state;
};

static bool
oic_server_energy_overload_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_energy_overload *state = (struct oic_server_energy_overload *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("value", state->state.value));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_energy_overload_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.energy.overload");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_energy_overload_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_energy_overload *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_energy_overload_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_energy_overload *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_energy_overload_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_energy_overload *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_fev1 {
    struct server_resource base;
    struct oic_state_fev1 state;
};

static bool
oic_server_fev1_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_fev1 *state = (struct oic_server_fev1 *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("fev1", state->state.fev1));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_fev1_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.fev1");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_fev1_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_fev1 *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.fev1 = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_fev1_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_fev1 *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_fev1_set_fev1(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_fev1 *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.fev1, (double) var)) {
            resource->state.fev1 = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_fev1_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_fev1 *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_ffm {
    struct server_resource base;
    struct oic_state_ffm state;
};

static bool
oic_server_ffm_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_ffm *state = (struct oic_server_ffm *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("ffm", state->state.ffm));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_ffm_units_tbl[state->state.units].key, oic_state_ffm_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_ffm_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.ffm");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_ffm_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_ffm *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.ffm = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_ffm_units)0;
    }

    return r;
}

static void oic_server_ffm_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_ffm *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_ffm_set_ffm(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_ffm *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.ffm, (double) var)) {
            resource->state.ffm = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_ffm_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_ffm *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_ffm_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_ffm *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_ffm_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_ffm_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_server_gsr {
    struct server_resource base;
    struct oic_state_gsr state;
};

static bool
oic_server_gsr_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_gsr *state = (struct oic_server_gsr *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("skincon", state->state.skincon));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("skincv", state->state.skincv));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("skinres", state->state.skinres));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_gsr_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.gsr");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_gsr_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_gsr *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.skincon = 0.0f;
resource->state.skincv = 0.0f;
resource->state.skinres = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_gsr_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_gsr *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_gsr_set_skincon(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_gsr *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.skincon, (double) var)) {
            resource->state.skincon = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_gsr_set_skincv(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_gsr *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.skincv, (double) var)) {
            resource->state.skincv = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_gsr_set_skinres(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_gsr *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.skinres, (double) var)) {
            resource->state.skinres = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_gsr_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_gsr *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_heartrate {
    struct server_resource base;
    struct oic_state_heartrate state;
};

static bool
oic_server_heartrate_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_heartrate *state = (struct oic_server_heartrate *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("heartrate", state->state.heartrate));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_heartrate_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.heartrate");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_heartrate_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_heartrate *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.heartrate = 0;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_heartrate_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_heartrate *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_heartrate_set_heartrate(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_heartrate *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.heartrate, (int32_t) var)) {
            resource->state.heartrate = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_heartrate_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_heartrate *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_height {
    struct server_resource base;
    struct oic_state_height state;
};

static bool
oic_server_height_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_height *state = (struct oic_server_height *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("height", state->state.height));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_height_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_height *res = (struct oic_server_height *)resource;
    return oic_state_height_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_height_inform_flow(struct server_resource *resource)
{
    struct oic_server_height *state = (struct oic_server_height *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_HEIGHT__OUT__HEIGHT, state->state.height);
}

static int
oic_server_height_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.height");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_height_to_repr_vec,
        .from_repr_vec = oic_server_height_from_repr_vec,
        .inform_flow = oic_server_height_inform_flow
    };
    struct oic_server_height *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.height = 0.0f;
    }

    return r;
}

static void oic_server_height_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_height *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_height_set_height(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_height *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.height, (double) var)) {
            resource->state.height = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_humidity {
    struct server_resource base;
    struct oic_state_humidity state;
};

static bool
oic_server_humidity_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_humidity *state = (struct oic_server_humidity *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("desiredHumidity", state->state.desiredHumidity));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("humidity", state->state.humidity));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_humidity_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_humidity *res = (struct oic_server_humidity *)resource;
    return oic_state_humidity_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_humidity_inform_flow(struct server_resource *resource)
{
    struct oic_server_humidity *state = (struct oic_server_humidity *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_HUMIDITY__OUT__DESIRED_HUMIDITY, state->state.desiredHumidity);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_HUMIDITY__OUT__HUMIDITY, state->state.humidity);
}

static int
oic_server_humidity_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.humidity");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_humidity_to_repr_vec,
        .from_repr_vec = oic_server_humidity_from_repr_vec,
        .inform_flow = oic_server_humidity_inform_flow
    };
    struct oic_server_humidity *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.desiredHumidity = 0;
resource->state.humidity = 0;
    }

    return r;
}

static void oic_server_humidity_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_humidity *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_humidity_set_desiredHumidity(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_humidity *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.desiredHumidity, (int32_t) var)) {
            resource->state.desiredHumidity = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_humidity_set_humidity(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_humidity *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.humidity, (int32_t) var)) {
            resource->state.humidity = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_icemaker {
    struct server_resource base;
    struct oic_state_icemaker state;
};

static bool
oic_server_icemaker_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_icemaker *state = (struct oic_server_icemaker *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("status", oic_state_icemaker_status_tbl[state->state.status].key, oic_state_icemaker_status_tbl[state->state.status].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_icemaker_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_icemaker *res = (struct oic_server_icemaker *)resource;
    return oic_state_icemaker_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_icemaker_inform_flow(struct server_resource *resource)
{
    struct oic_server_icemaker *state = (struct oic_server_icemaker *)resource;
    sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_ICEMAKER__OUT__STATUS, oic_state_icemaker_status_tbl[state->state.status].key);
}

static int
oic_server_icemaker_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.iceMaker");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_icemaker_to_repr_vec,
        .from_repr_vec = oic_server_icemaker_from_repr_vec,
        .inform_flow = oic_server_icemaker_inform_flow
    };
    struct oic_server_icemaker *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.status = (enum oic_state_icemaker_status)0;
    }

    return r;
}

static void oic_server_icemaker_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_icemaker *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_icemaker_set_status(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_icemaker *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_icemaker_status_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.status = (enum oic_state_icemaker_status)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_server_light_brightness {
    struct server_resource base;
    struct oic_state_light_brightness state;
};

static bool
oic_server_light_brightness_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_light_brightness *state = (struct oic_server_light_brightness *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("brightness", state->state.brightness));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_light_brightness_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_light_brightness *res = (struct oic_server_light_brightness *)resource;
    return oic_state_light_brightness_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_light_brightness_inform_flow(struct server_resource *resource)
{
    struct oic_server_light_brightness *state = (struct oic_server_light_brightness *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_LIGHT_BRIGHTNESS__OUT__BRIGHTNESS, state->state.brightness);
}

static int
oic_server_light_brightness_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.light.brightness");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_light_brightness_to_repr_vec,
        .from_repr_vec = oic_server_light_brightness_from_repr_vec,
        .inform_flow = oic_server_light_brightness_inform_flow
    };
    struct oic_server_light_brightness *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.brightness = 0;
    }

    return r;
}

static void oic_server_light_brightness_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_light_brightness *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_light_brightness_set_brightness(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_light_brightness *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.brightness, (int32_t) var)) {
            resource->state.brightness = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_lock_status {
    struct server_resource base;
    struct oic_state_lock_status state;
};

static bool
oic_server_lock_status_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_lock_status *state = (struct oic_server_lock_status *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("lockState", oic_state_lock_status_lockState_tbl[state->state.lockState].key, oic_state_lock_status_lockState_tbl[state->state.lockState].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_lock_status_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_lock_status *res = (struct oic_server_lock_status *)resource;
    return oic_state_lock_status_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_lock_status_inform_flow(struct server_resource *resource)
{
    struct oic_server_lock_status *state = (struct oic_server_lock_status *)resource;
    sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_LOCK_STATUS__OUT__LOCK_STATE, oic_state_lock_status_lockState_tbl[state->state.lockState].key);
}

static int
oic_server_lock_status_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.lock.status");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_lock_status_to_repr_vec,
        .from_repr_vec = oic_server_lock_status_from_repr_vec,
        .inform_flow = oic_server_lock_status_inform_flow
    };
    struct oic_server_lock_status *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.lockState = (enum oic_state_lock_status_lockState)0;
    }

    return r;
}

static void oic_server_lock_status_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_lock_status *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_lock_status_set_lockState(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_lock_status *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_lock_status_lockState_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.lockState = (enum oic_state_lock_status_lockState)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_server_mode {
    struct server_resource base;
    struct oic_state_mode state;
};

static bool
oic_server_mode_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_mode *state = (struct oic_server_mode *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("modes", state->state.modes, state->state.modes ? strlen(state->state.modes) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("supportedModes", state->state.supportedModes, state->state.supportedModes ? strlen(state->state.supportedModes) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_mode_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_mode *res = (struct oic_server_mode *)resource;
    return oic_state_mode_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_mode_inform_flow(struct server_resource *resource)
{
    struct oic_server_mode *state = (struct oic_server_mode *)resource;
    send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_MODE__OUT__MODES, state->state.modes);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_MODE__OUT__SUPPORTED_MODES, state->state.supportedModes);
}

static int
oic_server_mode_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.mode");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_mode_to_repr_vec,
        .from_repr_vec = oic_server_mode_from_repr_vec,
        .inform_flow = oic_server_mode_inform_flow
    };
    struct oic_server_mode *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.modes = NULL;
resource->state.supportedModes = NULL;
    }

    return r;
}

static void oic_server_mode_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_mode *resource = data;
    free(resource->state.modes);
free(resource->state.supportedModes);
    server_resource_close(&resource->base);
}

static int
oic_server_mode_set_modes(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_mode *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.modes, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_mode_set_supportedModes(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_mode *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.supportedModes, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_muscle_o2_saturation {
    struct server_resource base;
    struct oic_state_muscle_o2_saturation state;
};

static bool
oic_server_muscle_o2_saturation_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_muscle_o2_saturation *state = (struct oic_server_muscle_o2_saturation *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("smo2", state->state.smo2));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_muscle_o2_saturation_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.muscle.o2.saturation");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_muscle_o2_saturation_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_muscle_o2_saturation *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.smo2 = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_muscle_o2_saturation_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_muscle_o2_saturation *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_muscle_o2_saturation_set_smo2(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_muscle_o2_saturation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.smo2, (double) var)) {
            resource->state.smo2 = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_muscle_o2_saturation_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_muscle_o2_saturation *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_nightmode {
    struct server_resource base;
    struct oic_state_nightmode state;
};

static bool
oic_server_nightmode_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_nightmode *state = (struct oic_server_nightmode *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("nightMode", state->state.nightMode));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_nightmode_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_nightmode *res = (struct oic_server_nightmode *)resource;
    return oic_state_nightmode_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_nightmode_inform_flow(struct server_resource *resource)
{
    struct oic_server_nightmode *state = (struct oic_server_nightmode *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_NIGHTMODE__OUT__NIGHT_MODE, state->state.nightMode);
}

static int
oic_server_nightmode_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.nightmode");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_nightmode_to_repr_vec,
        .from_repr_vec = oic_server_nightmode_from_repr_vec,
        .inform_flow = oic_server_nightmode_inform_flow
    };
    struct oic_server_nightmode *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.nightMode = false;
    }

    return r;
}

static void oic_server_nightmode_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_nightmode *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_nightmode_set_nightMode(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_nightmode *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.nightMode, (bool) var)) {
            resource->state.nightMode = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_operational_state {
    struct server_resource base;
    struct oic_state_operational_state state;
};

static bool
oic_server_operational_state_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_operational_state *state = (struct oic_server_operational_state *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("progressPercentage", state->state.progressPercentage));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("currentJobState", state->state.currentJobState, state->state.currentJobState ? strlen(state->state.currentJobState) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("currentMachineState", state->state.currentMachineState, state->state.currentMachineState ? strlen(state->state.currentMachineState) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("jobStates", state->state.jobStates, state->state.jobStates ? strlen(state->state.jobStates) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("machineStates", state->state.machineStates, state->state.machineStates ? strlen(state->state.machineStates) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("remainingTime", state->state.remainingTime, state->state.remainingTime ? strlen(state->state.remainingTime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("runningTime", state->state.runningTime, state->state.runningTime ? strlen(state->state.runningTime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_operational_state_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_operational_state *res = (struct oic_server_operational_state *)resource;
    return oic_state_operational_state_from_repr_vec(&res->state, repr_vec, 0x6);
}

static void oic_server_operational_state_inform_flow(struct server_resource *resource)
{
    struct oic_server_operational_state *state = (struct oic_server_operational_state *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_OPERATIONAL_STATE__OUT__PROGRESS_PERCENTAGE, state->state.progressPercentage);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_OPERATIONAL_STATE__OUT__CURRENT_JOB_STATE, state->state.currentJobState);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_OPERATIONAL_STATE__OUT__CURRENT_MACHINE_STATE, state->state.currentMachineState);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_OPERATIONAL_STATE__OUT__JOB_STATES, state->state.jobStates);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_OPERATIONAL_STATE__OUT__MACHINE_STATES, state->state.machineStates);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_OPERATIONAL_STATE__OUT__REMAINING_TIME, state->state.remainingTime);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_OPERATIONAL_STATE__OUT__RUNNING_TIME, state->state.runningTime);
}

static int
oic_server_operational_state_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.operational.state");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_operational_state_to_repr_vec,
        .from_repr_vec = oic_server_operational_state_from_repr_vec,
        .inform_flow = oic_server_operational_state_inform_flow
    };
    struct oic_server_operational_state *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.progressPercentage = 0;
resource->state.currentJobState = NULL;
resource->state.currentMachineState = NULL;
resource->state.jobStates = NULL;
resource->state.machineStates = NULL;
resource->state.remainingTime = NULL;
resource->state.runningTime = NULL;
    }

    return r;
}

static void oic_server_operational_state_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_operational_state *resource = data;
    free(resource->state.currentJobState);
free(resource->state.currentMachineState);
free(resource->state.jobStates);
free(resource->state.machineStates);
free(resource->state.remainingTime);
free(resource->state.runningTime);
    server_resource_close(&resource->base);
}

static int
oic_server_operational_state_set_progressPercentage(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_operational_state *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.progressPercentage, (int32_t) var)) {
            resource->state.progressPercentage = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_operational_state_set_currentJobState(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_operational_state *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.currentJobState, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_operational_state_set_currentMachineState(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_operational_state *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.currentMachineState, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_operational_state_set_jobStates(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_operational_state *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.jobStates, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_operational_state_set_machineStates(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_operational_state *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.machineStates, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_operational_state_set_remainingTime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_operational_state *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.remainingTime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_operational_state_set_runningTime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_operational_state *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.runningTime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_pef {
    struct server_resource base;
    struct oic_state_pef state;
};

static bool
oic_server_pef_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_pef *state = (struct oic_server_pef *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("pef", state->state.pef));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_pef_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.pef");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_pef_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_pef *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.pef = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_pef_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_pef *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_pef_set_pef(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_pef *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.pef, (double) var)) {
            resource->state.pef = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_pef_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_pef *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_refrigeration {
    struct server_resource base;
    struct oic_state_refrigeration state;
};

static bool
oic_server_refrigeration_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_refrigeration *state = (struct oic_server_refrigeration *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("defrost", state->state.defrost));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("rapidCool", state->state.rapidCool));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_BOOL("rapidFreeze", state->state.rapidFreeze));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("filter", state->state.filter));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_refrigeration_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_refrigeration *res = (struct oic_server_refrigeration *)resource;
    return oic_state_refrigeration_from_repr_vec(&res->state, repr_vec, 0x7);
}

static void oic_server_refrigeration_inform_flow(struct server_resource *resource)
{
    struct oic_server_refrigeration *state = (struct oic_server_refrigeration *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_REFRIGERATION__OUT__DEFROST, state->state.defrost);
sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_REFRIGERATION__OUT__RAPID_COOL, state->state.rapidCool);
sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_REFRIGERATION__OUT__RAPID_FREEZE, state->state.rapidFreeze);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_REFRIGERATION__OUT__FILTER, state->state.filter);
}

static int
oic_server_refrigeration_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.refrigeration");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_refrigeration_to_repr_vec,
        .from_repr_vec = oic_server_refrigeration_from_repr_vec,
        .inform_flow = oic_server_refrigeration_inform_flow
    };
    struct oic_server_refrigeration *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.defrost = false;
resource->state.rapidCool = false;
resource->state.rapidFreeze = false;
resource->state.filter = 0;
    }

    return r;
}

static void oic_server_refrigeration_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_refrigeration *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_refrigeration_set_defrost(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_refrigeration *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.defrost, (bool) var)) {
            resource->state.defrost = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_refrigeration_set_rapidCool(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_refrigeration *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.rapidCool, (bool) var)) {
            resource->state.rapidCool = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_refrigeration_set_rapidFreeze(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_refrigeration *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.rapidFreeze, (bool) var)) {
            resource->state.rapidFreeze = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_refrigeration_set_filter(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_refrigeration *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.filter, (int32_t) var)) {
            resource->state.filter = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_repetition {
    struct server_resource base;
    struct oic_state_repetition state;
};

static bool
oic_server_repetition_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_repetition *state = (struct oic_server_repetition *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("repetition", state->state.repetition));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_repetition_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.repetition");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_repetition_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_repetition *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.repetition = 0;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_repetition_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_repetition *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_repetition_set_repetition(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_repetition *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.repetition, (int32_t) var)) {
            resource->state.repetition = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_repetition_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_repetition *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_respirationrate {
    struct server_resource base;
    struct oic_state_respirationrate state;
};

static bool
oic_server_respirationrate_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_respirationrate *state = (struct oic_server_respirationrate *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("respirate", state->state.respirate));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_respirationrate_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.respirationrate");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_respirationrate_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_respirationrate *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.respirate = 0;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_respirationrate_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_respirationrate *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_respirationrate_set_respirate(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_respirationrate *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.respirate, (int32_t) var)) {
            resource->state.respirate = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_respirationrate_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_respirationrate *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_sensor_carbondioxide {
    struct server_resource base;
    struct oic_state_sensor_carbondioxide state;
};

static bool
oic_server_sensor_carbondioxide_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_carbondioxide */
}

static int
oic_server_sensor_carbondioxide_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.carbondioxide");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_carbondioxide_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_carbondioxide *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_carbondioxide_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_carbondioxide *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_carbondioxide_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_carbondioxide *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_carbonmonoxide {
    struct server_resource base;
    struct oic_state_sensor_carbonmonoxide state;
};

static bool
oic_server_sensor_carbonmonoxide_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_carbonmonoxide */
}

static int
oic_server_sensor_carbonmonoxide_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.carbonmonoxide");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_carbonmonoxide_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_carbonmonoxide *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_carbonmonoxide_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_carbonmonoxide *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_carbonmonoxide_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_carbonmonoxide *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_contact {
    struct server_resource base;
    struct oic_state_sensor_contact state;
};

static bool
oic_server_sensor_contact_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_contact */
}

static int
oic_server_sensor_contact_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.contact");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_contact_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_contact *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_contact_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_contact *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_contact_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_contact *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_geolocation {
    struct server_resource base;
    struct oic_state_sensor_geolocation state;
};

static bool
oic_server_sensor_geolocation_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_sensor_geolocation *state = (struct oic_server_sensor_geolocation *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("accuracy", state->state.accuracy));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("alt", state->state.alt));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("altitudeAccuracy", state->state.altitudeAccuracy));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("heading", state->state.heading));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("latitude", state->state.latitude));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("longitude", state->state.longitude));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("speed", state->state.speed));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_sensor_geolocation_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_sensor_geolocation *res = (struct oic_server_sensor_geolocation *)resource;
    return oic_state_sensor_geolocation_from_repr_vec(&res->state, repr_vec, 0x7f);
}

static void oic_server_sensor_geolocation_inform_flow(struct server_resource *resource)
{
    struct oic_server_sensor_geolocation *state = (struct oic_server_sensor_geolocation *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SENSOR_GEOLOCATION__OUT__ACCURACY, state->state.accuracy);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SENSOR_GEOLOCATION__OUT__ALT, state->state.alt);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SENSOR_GEOLOCATION__OUT__ALTITUDE_ACCURACY, state->state.altitudeAccuracy);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SENSOR_GEOLOCATION__OUT__HEADING, state->state.heading);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SENSOR_GEOLOCATION__OUT__LATITUDE, state->state.latitude);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SENSOR_GEOLOCATION__OUT__LONGITUDE, state->state.longitude);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SENSOR_GEOLOCATION__OUT__SPEED, state->state.speed);
}

static int
oic_server_sensor_geolocation_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.geolocation");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_geolocation_to_repr_vec,
        .from_repr_vec = oic_server_sensor_geolocation_from_repr_vec,
        .inform_flow = oic_server_sensor_geolocation_inform_flow
    };
    struct oic_server_sensor_geolocation *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.accuracy = 0.0f;
resource->state.alt = 0.0f;
resource->state.altitudeAccuracy = 0.0f;
resource->state.heading = 0.0f;
resource->state.latitude = 0.0f;
resource->state.longitude = 0.0f;
resource->state.speed = 0.0f;
    }

    return r;
}

static void oic_server_sensor_geolocation_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_geolocation *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_geolocation_set_accuracy(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.accuracy, (double) var)) {
            resource->state.accuracy = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sensor_geolocation_set_alt(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.alt, (double) var)) {
            resource->state.alt = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sensor_geolocation_set_altitudeAccuracy(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.altitudeAccuracy, (double) var)) {
            resource->state.altitudeAccuracy = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sensor_geolocation_set_heading(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.heading, (double) var)) {
            resource->state.heading = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sensor_geolocation_set_latitude(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.latitude, (double) var)) {
            resource->state.latitude = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sensor_geolocation_set_longitude(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.longitude, (double) var)) {
            resource->state.longitude = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sensor_geolocation_set_speed(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_geolocation *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.speed, (double) var)) {
            resource->state.speed = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_glassbreak {
    struct server_resource base;
    struct oic_state_sensor_glassbreak state;
};

static bool
oic_server_sensor_glassbreak_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_glassbreak */
}

static int
oic_server_sensor_glassbreak_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.glassbreak");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_glassbreak_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_glassbreak *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_glassbreak_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_glassbreak *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_glassbreak_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_glassbreak *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor {
    struct server_resource base;
    struct oic_state_sensor state;
};

static bool
oic_server_sensor_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor */
}

static int
oic_server_sensor_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_motion {
    struct server_resource base;
    struct oic_state_sensor_motion state;
};

static bool
oic_server_sensor_motion_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_motion */
}

static int
oic_server_sensor_motion_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.motion");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_motion_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_motion *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_motion_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_motion *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_motion_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_motion *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_presence {
    struct server_resource base;
    struct oic_state_sensor_presence state;
};

static bool
oic_server_sensor_presence_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_presence */
}

static int
oic_server_sensor_presence_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.presence");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_presence_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_presence *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_presence_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_presence *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_presence_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_presence *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_radiation_uv {
    struct server_resource base;
    struct oic_state_sensor_radiation_uv state;
};

static bool
oic_server_sensor_radiation_uv_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_sensor_radiation_uv *state = (struct oic_server_sensor_radiation_uv *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("measurement", state->state.measurement));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_sensor_radiation_uv_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.radiation.uv");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_radiation_uv_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_radiation_uv *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.measurement = 0.0f;
    }

    return r;
}

static void oic_server_sensor_radiation_uv_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_radiation_uv *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_radiation_uv_set_measurement(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_radiation_uv *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.measurement, (double) var)) {
            resource->state.measurement = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_sleep {
    struct server_resource base;
    struct oic_state_sensor_sleep state;
};

static bool
oic_server_sensor_sleep_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_sleep */
}

static int
oic_server_sensor_sleep_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.sleep");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_sleep_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_sleep *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_sleep_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_sleep *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_sleep_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_sleep *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_smoke {
    struct server_resource base;
    struct oic_state_sensor_smoke state;
};

static bool
oic_server_sensor_smoke_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_smoke */
}

static int
oic_server_sensor_smoke_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.smoke");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_smoke_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_smoke *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_smoke_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_smoke *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_smoke_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_smoke *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_touch {
    struct server_resource base;
    struct oic_state_sensor_touch state;
};

static bool
oic_server_sensor_touch_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_touch */
}

static int
oic_server_sensor_touch_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.touch");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_touch_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_touch *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_touch_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_touch *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_touch_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_touch *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sensor_water {
    struct server_resource base;
    struct oic_state_sensor_water state;
};

static bool
oic_server_sensor_water_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_sensor_water */
}

static int
oic_server_sensor_water_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sensor.water");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sensor_water_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sensor_water *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_sensor_water_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sensor_water *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_sensor_water_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sensor_water *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_signalstrength {
    struct server_resource base;
    struct oic_state_signalstrength state;
};

static bool
oic_server_signalstrength_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_signalstrength *state = (struct oic_server_signalstrength *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("lqi", state->state.lqi));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("rssi", state->state.rssi));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_signalstrength_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.signalstrength");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_signalstrength_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_signalstrength *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.lqi = 0.0f;
resource->state.rssi = 0.0f;
    }

    return r;
}

static void oic_server_signalstrength_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_signalstrength *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_signalstrength_set_lqi(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_signalstrength *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.lqi, (double) var)) {
            resource->state.lqi = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_signalstrength_set_rssi(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_signalstrength *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.rssi, (double) var)) {
            resource->state.rssi = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_sleep {
    struct server_resource base;
    struct oic_state_sleep state;
};

static bool
oic_server_sleep_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_sleep *state = (struct oic_server_sleep *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("At", state->state.At));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("Dst", state->state.Dst));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("Lst", state->state.Lst));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("Tst", state->state.Tst));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("Type", oic_state_sleep_Type_tbl[state->state.Type].key, oic_state_sleep_Type_tbl[state->state.Type].len));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_sleep_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.sleep");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_sleep_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_sleep *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.At = 0.0f;
resource->state.Dst = 0.0f;
resource->state.Lst = 0.0f;
resource->state.Tst = 0.0f;
resource->state.Type = (enum oic_state_sleep_Type)0;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_sleep_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_sleep *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_sleep_set_At(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sleep *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.At, (double) var)) {
            resource->state.At = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sleep_set_Dst(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sleep *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.Dst, (double) var)) {
            resource->state.Dst = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sleep_set_Lst(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sleep *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.Lst, (double) var)) {
            resource->state.Lst = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sleep_set_Tst(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sleep *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.Tst, (double) var)) {
            resource->state.Tst = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_sleep_set_Type(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sleep *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_sleep_Type_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.Type = (enum oic_state_sleep_Type)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}

static int
oic_server_sleep_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_sleep *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_slm {
    struct server_resource base;
    struct oic_state_slm state;
};

static bool
oic_server_slm_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_slm *state = (struct oic_server_slm *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("slm", state->state.slm));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_slm_units_tbl[state->state.units].key, oic_state_slm_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_slm_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.slm");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_slm_to_repr_vec,
        .from_repr_vec = NULL,
        .inform_flow = NULL
    };
    struct oic_server_slm *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.slm = 0.0f;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_slm_units)0;
    }

    return r;
}

static void oic_server_slm_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_slm *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_slm_set_slm(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_slm *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.slm, (double) var)) {
            resource->state.slm = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_slm_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_slm *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_slm_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_slm *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_slm_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_slm_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_server_speech_tts {
    struct server_resource base;
    struct oic_state_speech_tts state;
};

static bool
oic_server_speech_tts_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_speech_tts *state = (struct oic_server_speech_tts *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("supportedLanguages", state->state.supportedLanguages, state->state.supportedLanguages ? strlen(state->state.supportedLanguages) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("supportedVoices", state->state.supportedVoices, state->state.supportedVoices ? strlen(state->state.supportedVoices) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("utterance", state->state.utterance, state->state.utterance ? strlen(state->state.utterance) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_speech_tts_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_speech_tts *res = (struct oic_server_speech_tts *)resource;
    return oic_state_speech_tts_from_repr_vec(&res->state, repr_vec, 0x4);
}

static void oic_server_speech_tts_inform_flow(struct server_resource *resource)
{
    struct oic_server_speech_tts *state = (struct oic_server_speech_tts *)resource;
    send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SPEECH_TTS__OUT__SUPPORTED_LANGUAGES, state->state.supportedLanguages);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SPEECH_TTS__OUT__SUPPORTED_VOICES, state->state.supportedVoices);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SPEECH_TTS__OUT__UTTERANCE, state->state.utterance);
}

static int
oic_server_speech_tts_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.speech.tts");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_speech_tts_to_repr_vec,
        .from_repr_vec = oic_server_speech_tts_from_repr_vec,
        .inform_flow = oic_server_speech_tts_inform_flow
    };
    struct oic_server_speech_tts *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.supportedLanguages = NULL;
resource->state.supportedVoices = NULL;
resource->state.utterance = NULL;
    }

    return r;
}

static void oic_server_speech_tts_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_speech_tts *resource = data;
    free(resource->state.supportedLanguages);
free(resource->state.supportedVoices);
free(resource->state.utterance);
    server_resource_close(&resource->base);
}

static int
oic_server_speech_tts_set_supportedLanguages(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_speech_tts *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.supportedLanguages, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_speech_tts_set_supportedVoices(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_speech_tts *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.supportedVoices, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_speech_tts_set_utterance(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_speech_tts *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.utterance, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_speed {
    struct server_resource base;
    struct oic_state_speed state;
};

static bool
oic_server_speed_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_speed *state = (struct oic_server_speed *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("speed", state->state.speed));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("units", oic_state_speed_units_tbl[state->state.units].key, oic_state_speed_units_tbl[state->state.units].len));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_speed_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_speed *res = (struct oic_server_speed *)resource;
    return oic_state_speed_from_repr_vec(&res->state, repr_vec, 0x5);
}

static void oic_server_speed_inform_flow(struct server_resource *resource)
{
    struct oic_server_speed *state = (struct oic_server_speed *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SPEED__OUT__SPEED, state->state.speed);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SPEED__OUT__OBSERVEDTIME, state->state.observedtime);
sol_flow_send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SPEED__OUT__UNITS, oic_state_speed_units_tbl[state->state.units].key);
}

static int
oic_server_speed_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.speed");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_speed_to_repr_vec,
        .from_repr_vec = oic_server_speed_from_repr_vec,
        .inform_flow = oic_server_speed_inform_flow
    };
    struct oic_server_speed *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.speed = 0;
resource->state.observedtime = NULL;
resource->state.units = (enum oic_state_speed_units)0;
    }

    return r;
}

static void oic_server_speed_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_speed *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_speed_set_speed(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_speed *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.speed, (int32_t) var)) {
            resource->state.speed = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_speed_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_speed *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_speed_set_units(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_speed *resource = data;
    const char *var;

    if (!sol_flow_packet_get_string(packet, &var)) {
        int16_t val = sol_str_table_lookup_fallback(oic_state_speed_units_tbl,
            sol_str_slice_from_str(var), -1);
        if (val >= 0) {
            resource->state.units = (enum oic_state_speed_units)val;
            server_resource_schedule_update(&resource->base);
            return 0;
        }
        return -ENOENT;
    }
    return -EINVAL;
}
struct oic_server_steps {
    struct server_resource base;
    struct oic_state_steps state;
};

static bool
oic_server_steps_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_steps *state = (struct oic_server_steps *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("count", state->state.count));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_INT("distance", state->state.distance));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("kcal", state->state.kcal));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("observedtime", state->state.observedtime, state->state.observedtime ? strlen(state->state.observedtime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_steps_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_steps *res = (struct oic_server_steps *)resource;
    return oic_state_steps_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_steps_inform_flow(struct server_resource *resource)
{
    struct oic_server_steps *state = (struct oic_server_steps *)resource;
    sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_STEPS__OUT__COUNT, state->state.count);
sol_flow_send_irange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_STEPS__OUT__DISTANCE, state->state.distance);
sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_STEPS__OUT__KCAL, state->state.kcal);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_STEPS__OUT__OBSERVEDTIME, state->state.observedtime);
}

static int
oic_server_steps_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.steps");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_steps_to_repr_vec,
        .from_repr_vec = oic_server_steps_from_repr_vec,
        .inform_flow = oic_server_steps_inform_flow
    };
    struct oic_server_steps *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.count = 0;
resource->state.distance = 0;
resource->state.kcal = 0.0f;
resource->state.observedtime = NULL;
    }

    return r;
}

static void oic_server_steps_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_steps *resource = data;
    free(resource->state.observedtime);
    server_resource_close(&resource->base);
}

static int
oic_server_steps_set_count(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_steps *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.count, (int32_t) var)) {
            resource->state.count = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_steps_set_distance(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_steps *resource = data;
    int32_t var;
    int r;

    r = sol_flow_packet_get_irange_value(packet, &var);
    if (!r) {
        if (check_updated_int32(resource->state.distance, (int32_t) var)) {
            resource->state.distance = (int32_t) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_steps_set_kcal(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_steps *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.kcal, (double) var)) {
            resource->state.kcal = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

static int
oic_server_steps_set_observedtime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_steps *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.observedtime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_switch_binary {
    struct server_resource base;
    struct oic_state_switch_binary state;
};

static bool
oic_server_switch_binary_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map_encoder)
{
    return oic_server_energy_overload_to_repr_vec(data, repr_map_encoder); /* oic_server_energy_overload is equivalent to oic_server_switch_binary */
}

static int
oic_server_switch_binary_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_switch_binary *res = (struct oic_server_switch_binary *)resource;
    return oic_state_switch_binary_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_switch_binary_inform_flow(struct server_resource *resource)
{
    struct oic_server_switch_binary *state = (struct oic_server_switch_binary *)resource;
    sol_flow_send_bool_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_SWITCH_BINARY__OUT__VALUE, state->state.value);
}

static int
oic_server_switch_binary_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.switch.binary");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_switch_binary_to_repr_vec,
        .from_repr_vec = oic_server_switch_binary_from_repr_vec,
        .inform_flow = oic_server_switch_binary_inform_flow
    };
    struct oic_server_switch_binary *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.value = false;
    }

    return r;
}

static void oic_server_switch_binary_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_switch_binary *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_switch_binary_set_value(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_switch_binary *resource = data;
    bool var;
    int r;

    r = sol_flow_packet_get_bool(packet, &var);
    if (!r) {
        if (check_updated_boolean(resource->state.value, (bool) var)) {
            resource->state.value = (bool) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}
struct oic_server_time_period {
    struct server_resource base;
    struct oic_state_time_period state;
};

static bool
oic_server_time_period_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_time_period *state = (struct oic_server_time_period *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("startTime", state->state.startTime, state->state.startTime ? strlen(state->state.startTime) : 0));
        SOL_INT_CHECK(r, < 0, false);

r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_TEXT_STRING("stopTime", state->state.stopTime, state->state.stopTime ? strlen(state->state.stopTime) : 0));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_time_period_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_time_period *res = (struct oic_server_time_period *)resource;
    return oic_state_time_period_from_repr_vec(&res->state, repr_vec, 0x3);
}

static void oic_server_time_period_inform_flow(struct server_resource *resource)
{
    struct oic_server_time_period *state = (struct oic_server_time_period *)resource;
    send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_TIME_PERIOD__OUT__START_TIME, state->state.startTime);
send_string_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_TIME_PERIOD__OUT__STOP_TIME, state->state.stopTime);
}

static int
oic_server_time_period_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.time.period");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_time_period_to_repr_vec,
        .from_repr_vec = oic_server_time_period_from_repr_vec,
        .inform_flow = oic_server_time_period_inform_flow
    };
    struct oic_server_time_period *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.startTime = NULL;
resource->state.stopTime = NULL;
    }

    return r;
}

static void oic_server_time_period_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_time_period *resource = data;
    free(resource->state.startTime);
free(resource->state.stopTime);
    server_resource_close(&resource->base);
}

static int
oic_server_time_period_set_startTime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_time_period *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.startTime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}

static int
oic_server_time_period_set_stopTime(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_time_period *resource = data;
    const char *var;
    int r;

    r = sol_flow_packet_get_string(packet, &var);
    if (!r) {
        r = sol_util_replace_str_if_changed(&resource->state.stopTime, var);
        SOL_INT_CHECK(r, < 0, r);
        if (r > 0) {
            server_resource_schedule_update(&resource->base);
            r = 0;
        }
    }
    return r;
}
struct oic_server_weight {
    struct server_resource base;
    struct oic_state_weight state;
};

static bool
oic_server_weight_to_repr_vec(void *data, struct sol_oic_map_writer *repr_map)
{
    struct oic_server_weight *state = (struct oic_server_weight *)data;
    int r;

    r = sol_oic_map_append(repr_map, &SOL_OIC_REPR_DOUBLE("weight", state->state.weight));
        SOL_INT_CHECK(r, < 0, false);
    return true;
}

static int
oic_server_weight_from_repr_vec(struct server_resource *resource, const struct sol_oic_map_reader *repr_vec)
{
    struct oic_server_weight *res = (struct oic_server_weight *)resource;
    return oic_state_weight_from_repr_vec(&res->state, repr_vec, 0x1);
}

static void oic_server_weight_inform_flow(struct server_resource *resource)
{
    struct oic_server_weight *state = (struct oic_server_weight *)resource;
    sol_flow_send_drange_value_packet(resource->node, SOL_FLOW_NODE_TYPE_OIC_SERVER_WEIGHT__OUT__WEIGHT, state->state.weight);
}

static int
oic_server_weight_open(struct sol_flow_node *node, void *data, const struct sol_flow_node_options *options)
{
    static const struct sol_str_slice rt_slice = SOL_STR_SLICE_LITERAL("oic.r.weight");
    static const struct server_resource_funcs funcs = {
        .to_repr_vec = oic_server_weight_to_repr_vec,
        .from_repr_vec = oic_server_weight_from_repr_vec,
        .inform_flow = oic_server_weight_inform_flow
    };
    struct oic_server_weight *resource = data;
    int r;

    r = server_resource_init(&resource->base, node, rt_slice, &funcs);
    if (!r) {
        resource->state.weight = 0.0f;
    }

    return r;
}

static void oic_server_weight_close(struct sol_flow_node *node, void *data)
{
    struct oic_server_weight *resource = data;
    
    server_resource_close(&resource->base);
}

static int
oic_server_weight_set_weight(struct sol_flow_node *node, void *data, uint16_t port,
    uint16_t conn_id, const struct sol_flow_packet *packet)
{
    struct oic_server_weight *resource = data;
    double var;
    int r;

    r = sol_flow_packet_get_drange_value(packet, &var);
    if (!r) {
        if (check_updated_number(resource->state.weight, (double) var)) {
            resource->state.weight = (double) var;
            server_resource_schedule_update(&resource->base);
        }
    }
    return r;
}

#undef RETURN_ERROR

#include "oic-gen.c"
